/*
 * rc-deps.c: Dependency resolution biits
 *
 * Copyright (C) 2000, 2001   Ximian, Inc.
 *
 * Authors: Vladimir Vukicevic <vladimir@ximian.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>

#include <glib.h>

#include "rc-deps.h"
#include "rc-debug.h"
#include "rc-debug-misc.h"
#include "rc-channel.h"
#include "rc-packman.h"

#include "rc-package-info.h"

#ifndef DEBUG
#define DEBUG
#endif

RCPackman *das_global_packman;

static void populate_dr_with_system (RCDepResolution *rcdr,
                                     RCPackageSList *system_packages);
static void deps_add_package (RCDepResolution *rcdr,
                              RCPackage *pkg,
                              gboolean force);
static void deps_remove_package (RCDepResolution *rcdr,
                                 RCPackage *pkg,
                                 gboolean as_conflict);
static RCDepProvide *deps_provide_new (gchar *name, RCPackageDepItem *provision, RCPackage *provided_by);
static RCDepRequire *deps_require_new (GSList *names, RCPackageDep *dep);
static void package_push_provide_list (RCDepResolution *rcdr,
                                       RCPackage *pkg,
                                       RCPackageDepSList *provides,
                                       gboolean force);
static void package_push_require_list (RCDepResolution *rcdr,
                                       RCPackage *pkg,
                                       RCPackageDepSList *requires,
                                       gboolean force);
static void package_push_conflict_list (RCDepResolution *rcdr,
                                        RCPackage *pkg,
                                        RCPackageDepSList *conflicts,
                                        gboolean force);
static void package_push_provide (RCDepResolution *rcdr,
                                  RCPackage *pkg,
                                  RCPackageDep *pdep,
                                  gboolean force);
static gchar *deps_join_names (GSList *name_list);
static gchar *deps_make_names (RCPackageDep *dep, GSList **list_out);
static gboolean deps_require_merge_requirement (RCDepRequire *rcdr,
                                               RCPackageDep *rcd);
static gboolean deps_require_add_package (RCDepRequire *rcdr,
                                          RCPackage *pkg,
                                          RCPackageDep *rcd,
                                          gboolean force);
static void deps_require_unprovide (RCDepResolution *rcdr, RCDepRequire *req,
                                    gboolean as_conflict);
static void package_push_require (RCDepResolution *rcdr,
                                  RCPackage *pkg,
                                  RCPackageDep *pdep,
                                  gboolean forced);
static RCPackage *deps_find_newer_package_for_dap (RCDepResolution *rcdr,
                                                   RCDepAndPackage *dap);
static RCPackage * deps_find_newer_package_for_dep (RCDepResolution *rcdr,
                                                    RCPackageDep *dep);

#if NOTUSED
static RCPackage *deps_find_newer_package_for_pdi (RCDepResolution *rcdr,
                                                   RCPackageDepItem *pdi);
#endif
static RCPackage *deps_find_newer_package_by_name (RCDepResolution *rcdr,
                                                   gchar *name);
static RCPackage *deps_find_newer_package (RCDepResolution *rcdr,
                                           RCPackage *pkg);

static void package_mark_provides_as_removed (RCDepResolution *rcdr,
                                              RCPackage *pkg,
                                              RCPackageDepSList *pkg_provs,
                                              gboolean as_conflict);

static RCPackageDepSList *deps_dep_to_or_slist (RCPackageDep *dep);

static void deps_provide_require_iterator (gpointer key, gpointer value, gpointer data);
static int deps_provide_require (RCDepResolution *rcdr, RCDepRequire *req, gchar *name,
                                 gboolean invalidate_provide);
static gboolean deps_provide_require_check (RCDepResolution *rcdr, RCDepRequire *req, RCDepProvide *prov);

static void deps_replace_provide (RCDepResolution *rcdr, RCDepProvide *oldp, RCDepProvide *newp);
static void deps_provide_unprovide (RCDepResolution *rcdr, RCDepProvide *prov);
static int deps_require_remove_deps (RCDepResolution *rcdr, RCDepRequire *req,
                                     RCPackageDepItemSList *depis,
                                     gboolean force);
#ifdef NOTUSED
static gboolean deps_remove_require (RCDepResolution *rcdr, RCDepRequire *req, gboolean hash_remove);
#endif
static void deps_require_free (RCDepRequire *req);

static RCPackage *deps_find_best_package_for_dep_slist (RCPackageDepSList *deps,
                                                       RCChannelSList *chs,
                                                       gint user_pref);
static RCPackage *deps_find_best_package_for_dep (RCPackageDep *dep,
                                                 RCChannelSList *chs,
                                                 gint user_pref);

static gboolean deps_verify_and_dep_slist (RCPackageDepSList *dep_list, RCPackageDepItem *item,
                                           GSList **failtures);
static int deps_find_best_provide (RCDepResolution *rcdr, RCDepRequire *req, RCDepProvide **prov_out);
static RCPackageDepSList *deps_dep_to_or_slist (RCPackageDep *dep);
static RCDepAndPackageSList *find_daps_for_pdi (RCDepResolution *rcdrr, RCDepRequire *req,
                                                RCPackageDepItem *pdi);

static gboolean process_failed_list (RCDepResolution *rcdr);
static void process_missing_list (RCDepResolution *rcdr);
static void process_provides_list (RCDepResolution *rcdr);

static void deps_add_conflict (RCDepResolution *rcdr,
                               RCDepConflictType type,
                               gchar *what_str,
                               RCPackage *what,
                               RCPackageDepItem *what_conflicts,
                               RCPackage *with_other_package);

static void deps_untangle_package (RCDepResolution *rcdr, RCPackage *pkg);
static void deps_update_remove_list (RCDepResolution *rcdr, RCPackage *pkg);
static void deps_update_install_list (RCDepResolution *rcdr, RCPackage *pkg);

static void unprovided_hash_to_list (gpointer key, gpointer value, gpointer data);
RCPackageSList *deps_require_list_packages (RCDepRequire *req);
static gboolean deps_provide_check_must_provide (RCDepResolution *rcdr, RCDepRequire *req, RCDepProvide *maybe_provide);
static void check_daps_for_removal (RCDepResolution *rcdr, RCDepRequire *req);
static void deps_upgrade_or_remove_package (RCDepResolution *rcdr, RCPackage *pkg, gboolean force);

static gint package_find_by_name (gconstpointer a, gconstpointer b);

static void print_provide (int dlev, RCDepProvide *rcdr);
static void print_provide_hash (gpointer key, gpointer value, gpointer data);
static void print_missing_hash (gpointer key, gpointer value, gpointer data);
static void print_require (int dlev, RCDepRequire *rcdr);
static void print_require_hash (gpointer key, gpointer value, gpointer data);
static void print_conflict (int dlev, RCDepConflict *conf);

void rc_dep_set_packman (RCPackman *packman);


/** debugging control **/
#if 1
int rc_deps_debug_level = 1;
#else
static const int rc_deps_debug_level = 0;
#endif

/* debugging fun for the whole family! */
gchar *deps_pkg_watch_1 = NULL;
gchar *deps_pkg_watch_2 = NULL;
gchar *deps_pkg_watch_3 = NULL;

static void deps_bt (gchar *s) {
    if (rc_deps_debug_level) {
//        rc_debug (RC_DEBUG_LEVEL_DEBUG, "Backtrace:\n");
        if (s) {
            if (deps_pkg_watch_1 && !strcmp (deps_pkg_watch_1, s))
                rc_debug (RC_DEBUG_LEVEL_DEBUG, "BREAK PKG WATCH 1\n");
            if (deps_pkg_watch_2 && !strcmp (deps_pkg_watch_2, s))
                rc_debug (RC_DEBUG_LEVEL_DEBUG, "BREAK PKG WATCH 2\n");
            if (deps_pkg_watch_3 && !strcmp (deps_pkg_watch_3, s))
                rc_debug (RC_DEBUG_LEVEL_DEBUG, "BREAK PKG WATCH 3\n");
        }
//        rc_debug (RC_DEBUG_LEVEL_DEBUG, "----------\n");
    }
}


static int global_indent_level = 0;

#define DEPS_ENTER \
    do { int i; \
        if (rc_deps_debug_level > 20) { \
            for (i=0;i<global_indent_level;i++) fputc(' ', stderr); \
            global_indent_level++; \
            rc_debug (RC_DEBUG_LEVEL_DEBUG, ">>> %s[%d]\n", __FUNCTION__, __LINE__); \
        } \
    } while (0)

#define DEPS_EXIT \
    do { int i; \
        if (rc_deps_debug_level > 20) { \
            global_indent_level--; \
            for (i=0;i<global_indent_level;i++) fputc(' ', stderr); \
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "<<< %s[%d]\n", __FUNCTION__, __LINE__); \
        } \
    } while (0)

#define DEPS_DEBUG_PROV(name,prov) do { if (rc_deps_debug_level > 20) { \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] RCDepProvide %s:\n", __LINE__, name); \
    print_provide (RC_DEBUG_LEVEL_DEBUG, prov); \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "------\n"); \
} } while (0)

#define DEPS_DEBUG_REQ(name,req) do { if (rc_deps_debug_level > 20) { \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] RCDepRequire %s:\n", __LINE__, name); \
    print_require (RC_DEBUG_LEVEL_DEBUG, req); \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "------\n"); \
} } while (0)

#define DEPS_DEBUG_PKG(name,pkg) do { if (rc_deps_debug_level > 20) { \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] RCPackage %s: %s\n", __LINE__, name, rc_package_spec_to_str (&(pkg)->spec)); \
} } while (0)

#define DEPS_DEBUG_PDI(name,pdi) do { if (rc_deps_debug_level > 20) { \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] RCPackageDepItem %s: %s\n", __LINE__, name, debug_rc_package_dep_item_str (pdi)); \
} } while (0)

#define DEPS_DEBUG_PD(name,pd) do { if (rc_deps_debug_level > 20) { \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] RCPackageDepItem %s: ", __LINE__, name); debug_rc_package_dep_dump (RC_DEBUG_LEVEL_DEBUG, pd); \
} } while (0)

#define DEPS_DEBUG_PDSL(name,pdsl) do { if (rc_deps_debug_level > 20) { GSList *iter = pdsl; \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] RCPackageDepSList %s: ", __LINE__, name); while (iter) { debug_rc_package_dep_dump (RC_DEBUG_LEVEL_DEBUG, (RCPackageDep *) iter->data); rc_debug (RC_DEBUG_LEVEL_DEBUG, "\n"); iter = iter->next; } rc_debug (RC_DEBUG_LEVEL_DEBUG, "------\n"); \
} } while (0)

#define DEPS_DEBUG_STRING(name,str) do { if (rc_deps_debug_level > 20) { \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] string %s: '%s'\n", __LINE__, name, str); \
} } while (0)

#define DEPS_DEBUG_INT(name,num) do { if (rc_deps_debug_level > 20) { \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] int %s: '%d'\n", __LINE__, name, num); \
} } while (0)

#define DEPS_DEBUG_BOOL(name,bool) do { if (rc_deps_debug_level > 20) { \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] gboolean %s: '%s'\n", __LINE__, name, (bool) ? "TRUE" : "FALSE"); \
} } while (0)

#define DEPS_DEBUG_DAP(name,dap) do { if (rc_deps_debug_level > 20) { \
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "---[%d] RCPackageDap %s: pkg: %s dep:", __LINE__, name, rc_package_spec_to_str (&((dap)->package->spec))); debug_rc_package_dep_dump (RC_DEBUG_LEVEL_DEBUG, (dap)->dep); \
} } while (0)


/**************/
static gboolean conflicts_use_virtual = TRUE;

void deps_conflicts_use_virtual_packages (gboolean value)
{
    conflicts_use_virtual = value;
}
/**************/


RCDepResolution *
rc_dep_resolution_new (RCPackageSList *system_packages,
                       RCChannel *current_ch,
                       RCChannelSList *preferred_chs,
                       RCChannelSList *other_chs,
                       gint user_pref)
{
    RCDepResolution *rcdr;

#ifdef DEBUG
    if (getenv ("RC_DEPS_DEBUG_LEVEL")) {
        rc_deps_debug_level = strtol (getenv ("RC_DEPS_DEBUG_LEVEL"), NULL, 0);
    }
#endif

    rcdr = g_new0 (RCDepResolution, 1);

    rcdr->provide_hash = g_hash_table_new (g_str_hash, g_str_equal);
    rcdr->require_hash = g_hash_table_new (g_str_hash, g_str_equal);

    rcdr->current_ch = current_ch;
    rcdr->current_chs = current_ch ? g_slist_append (NULL, current_ch) : NULL;
    rcdr->preferred_chs = g_slist_copy (preferred_chs);
    rcdr->other_chs = g_slist_copy (other_chs);

    rcdr->user_pref = user_pref;

    if (system_packages) {
        /* We need to make sure that the system is in a consistent state,
         * and also populate the necessary hashes
         */
        populate_dr_with_system (rcdr, system_packages);
        rcdr->system_packages = g_slist_copy (system_packages);
    } else {
        rcdr->system_packages = NULL;
    }

    return rcdr;
}

void
rc_dep_resolution_add_package (RCDepResolution *rcdr,
                               RCPackage *pkg)
{
    DEPS_ENTER;
    DEPS_DEBUG_PKG("pkg", pkg);

    g_assert (rcdr);
    g_assert (pkg);

    deps_add_package (rcdr, pkg, TRUE);

    DEPS_EXIT;
}

void
rc_dep_resolution_add_package_by_name (RCDepResolution *rcdr,
                                       gchar *pkg_name)
{
    RCPackage *pkg;
    DEPS_ENTER;
    DEPS_DEBUG_STRING("pkg_name", pkg_name);

    if (rcdr->current_ch == NULL) {
        g_warning ("add_package_by_name: current_ch == NULL!");
        return;
    }

    pkg = pkginfo_find_package (rcdr->current_ch, pkg_name, rcdr->user_pref);
    if (!pkg) {
        g_warning ("add_package_by_name: package %s doesn't exist!", pkg_name);
        return;
    }

    rc_dep_resolution_add_package (rcdr, pkg);
    DEPS_EXIT;
}

void
rc_dep_resolution_remove_package (RCDepResolution *rcdr,
                                  RCPackage *pkg)
{
    DEPS_ENTER;
    DEPS_DEBUG_PKG("pkg", pkg);

    deps_bt(pkg->spec.name);

    g_assert (rcdr);
    g_assert (pkg);

    deps_remove_package (rcdr, pkg, FALSE);
    DEPS_EXIT;
}

void
rc_dep_resolution_remove_package_by_name (RCDepResolution *rcdr,
                                          gchar *pkg_name)
{
    RCPackage *pkg = NULL;
    GSList *iter;

    DEPS_ENTER;
    DEPS_DEBUG_STRING("pkg_name", pkg_name);

    /* First we check if it's something we've been trying to install */
    iter = rcdr->packages_added;
    while (iter) {
        RCPackage *p = (RCPackage *) iter->data;
        if (strcmp (p->spec.name, pkg_name) == 0) {
            pkg = p;
            break;
        }
        iter = iter->next;
    }

    if (!pkg && rcdr->provide_hash) {
        RCDepProvideSList *provs = g_hash_table_lookup (rcdr->provide_hash, pkg_name);
        while (provs) {
            RCDepProvide *prov = (RCDepProvide *) provs->data;
            if (strcmp (prov->provided_by->spec.name, pkg_name) == 0) {
                pkg = prov->provided_by;
                break;
            }
            provs = provs->next;
        }
    }

    if (pkg)
        rc_dep_resolution_remove_package (rcdr, pkg);
    else
        g_warning ("rc_dep_resolution_remove_package_by_name: No package named '%s' found\n",
                   pkg_name);

    DEPS_EXIT;
}

static gint
package_find_by_name (gconstpointer a, gconstpointer b)
{
    const RCPackageSpec *s = (const RCPackageSpec *) a;
    const gchar *name = (const gchar *) b;

    if (s->name) {
        return strcmp (s->name, name);
    } else {
        return -1;
    }
}

static gint
dap_find_by_name (gconstpointer a, gconstpointer b)
{
    const RCDepAndPackage *dap = (const RCDepAndPackage *) a;
    const gchar *name = (const gchar *) b;

    return strcmp (((RCPackageDepItem *)(dap->dep->data))->spec.name, name);
}


static void
deps_fix_missing_iter (gpointer key, gpointer value, gpointer data)
{
    RCDepRequireSList *reqs = (RCDepRequireSList *) value;
    RCDepResolution *rcdr = (RCDepResolution *) data;

    while (reqs) {
        RCDepRequire *req = (RCDepRequire *) reqs->data;
        if (req->top_provide == NULL) {
            RCDepAndPackageSList *iter;
            RCPackageSList *already_removed = NULL;

            if (rc_package_dep_slist_is_fully_weak (req->deps)) {
                reqs = reqs->next;
                continue;
            }

            iter = req->required_list;
            while (iter) {
                RCDepAndPackage *dap = (RCDepAndPackage *) iter->data;
                if (g_slist_find (rcdr->packages_added, dap->package)) {
                    deps_add_conflict (rcdr,
                                       RC_CONFLICT_REQUIREMENT,
                                       g_strconcat (deps_join_names (req->names), " cannot be installed", NULL),
                                       dap->package,
                                       NULL,
                                       NULL);
                    iter = iter->next;
                    continue;
                }

                if (!g_slist_find (already_removed, dap->package)) {
                    RCPackage *pkg = dap->package;
                    deps_remove_package (rcdr, pkg, FALSE);
                    already_removed = g_slist_append (already_removed, pkg);
                    iter = req->required_list;
                    continue;
                }
                iter = iter->next;
            }
            /* FIXME fill leak. We should also handle forced daps here, as these shouldn't
             * be this easy to get rid of.
             */
            g_slist_free (req->deps);
            req->deps = NULL;
            g_slist_free (already_removed);
        }
        reqs = reqs->next;
    }
}

static void
deps_fix_missing (RCDepResolution *rcdr)
{
    g_hash_table_foreach (rcdr->require_hash, deps_fix_missing_iter, rcdr);
}

static void
deps_fix_removed (RCDepResolution *rcdr)
{
    RCPackageSList *iter;

    iter = rcdr->packages_added;

    while (iter) {
        RCPackage *pkg = (RCPackage *) iter->data;
        GSList *lnk;

        if ((lnk = g_slist_find_custom (rcdr->packages_removed,
                                        pkg->spec.name,
                                        package_find_by_name))
            != NULL)
        {
            /* This package was upgraded (so the old one was removed) */
            rcdr->packages_removed = g_slist_remove (rcdr->packages_removed, lnk->data);
        }
        iter = iter->next;
    }
}

static void
deps_count_missing_iterator (gpointer key, gpointer value, gpointer data)
{
    RCDepRequireSList *reqs = (RCDepRequireSList *) value;
    gint **missing = (gint **) data;

    while (reqs) {
        RCDepRequire *req = (RCDepRequire *) reqs->data;
        if (req->top_provide == NULL ||
            req->top_provide == REQUIRE_ALL_REMOVED ||
            req->top_provide == REQUIRE_HAS_CONFLICT)
        {
            if (rc_package_dep_slist_is_fully_weak (req->deps)) {
                reqs = reqs->next;
                continue;
            }

            /* Increment the missing counter */
            if (req->top_provide == NULL)
                missing[0]++;
            else if (req->top_provide == REQUIRE_ALL_REMOVED)
                missing[1]++;
            else if (req->top_provide == REQUIRE_HAS_CONFLICT)
                missing[2]++;
        }
        reqs = reqs->next;
    }
}


static gint
deps_count_missing (RCDepResolution *rcdr, gint *unprovs, gint *removes, gint *confs)
{
    gint missing[3] = {0};
    g_hash_table_foreach (rcdr->require_hash, deps_count_missing_iterator, &missing);
    if (unprovs) *unprovs = missing[0];
    if (removes) *removes = missing[1];
    if (confs) *confs = missing[2];
    return missing[0] + missing[1] + missing[2];
}

RCDepResolutionResult *
rc_dep_resolution_get_result (RCDepResolution *rcdr)
{
    RCDepResolutionResult *res;
    gint last_miss = 0, last_confs = 0, last_rems = 0, last_rpkgs = 0;
    gint this_miss = 0, this_confs = 0, this_rems = 0, this_rpkgs = 0;
    gint passno = 0;
    gboolean force_again = FALSE;
    gboolean last_time = FALSE;
    gboolean dep_continue_iter = TRUE;

    /* Do the provide/require matching */
    do {
        last_miss = this_miss;
        last_rems = this_rems;
        last_confs = this_confs;
        last_rpkgs = this_rpkgs;

        rcdr->reqs_to_remove = NULL;

        rc_debug (RC_DEBUG_LEVEL_DEBUG, "-------------- PASS %d: %d [START] ------------\n", passno, last_miss);

        g_hash_table_foreach (rcdr->require_hash, deps_provide_require_iterator, rcdr);

        rc_debug (RC_DEBUG_LEVEL_DEBUG, "----- PASS %d: process_provides\n", passno);
        process_provides_list (rcdr);

        rc_debug (RC_DEBUG_LEVEL_DEBUG, "----- PASS %d: process_missing\n", passno);
        process_missing_list (rcdr);

        rc_debug (RC_DEBUG_LEVEL_DEBUG, "----- PASS %d: process_failed\n", passno);
        force_again = process_failed_list (rcdr);

        if (force_again) {
            passno++;
            continue;
        }

        deps_count_missing (rcdr, &this_miss, &this_rems, &this_confs);
        this_rpkgs = g_slist_length (rcdr->packages_removed);

        rcdr->unprovided_packages = this_miss;

        if (last_miss == this_miss && 
            last_rems == this_rems &&
            last_confs == this_confs &&
            last_rpkgs == this_rpkgs)
        {
            if (last_time)
                break;

            last_time = TRUE;
        } else {
            last_time = FALSE;
        }

        passno++;
    } while (passno != 200);

    /* post-process fix thingys */

    deps_fix_missing (rcdr);

    deps_fix_removed (rcdr);

    if (passno == 200) {
        rc_debug (RC_DEBUG_LEVEL_INFO, "**************** FAILED AFTER 200 PASSES *****************\n");
    }

    /* Then, for debugging's sake, we print some goodies */
    if (rc_deps_debug_level >= 10) {
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "--- PROVIDE ---\n");
        g_hash_table_foreach (rcdr->provide_hash, print_provide_hash, NULL);
        rc_debug (RC_DEBUG_LEVEL_INFO, "--- REQUIRE ---\n");
        g_hash_table_foreach (rcdr->require_hash, print_require_hash, NULL);
    }

    {
        /* do some conflict loving */
        GSList *new_list = NULL;
        GSList *iter = rcdr->packages_conflicting;
        while (iter) {
            RCDepConflict *conf = iter->data;
            if (conf->what) {
                GSList *lnk;
                if ((lnk = g_slist_find_custom (rcdr->packages_removed, conf->what->spec.name,
                                                package_find_by_name)) != NULL)
                {
                    RCPackage *pkg = lnk->data;
                    if (pkg && rc_package_spec_equal (&pkg->spec, &conf->what->spec)) {
                        iter = iter->next;
                        continue;
                    }
                } else if (g_slist_find (rcdr->packages_removed, conf->what)) {
                    iter = iter->next;
                    continue;
                }
            }
            new_list = g_slist_append (new_list, conf);
            iter = iter->next;
        }
        g_slist_free (rcdr->packages_conflicting);
        rcdr->packages_conflicting = new_list;
    }

    if (rc_deps_debug_level) {
        rc_debug (RC_DEBUG_LEVEL_INFO, "--- ADDED ---\n");
        {
            GSList *ps = rcdr->packages_added;
            int i = 0;
            while (ps) {
                RCPackage *pkg = (RCPackage *) ps->data;
                rc_debug (RC_DEBUG_LEVEL_INFO, "% 3d: %s\n", i++, rc_package_spec_to_str (&pkg->spec));
                ps = ps->next;
            }
        }
        rc_debug (RC_DEBUG_LEVEL_INFO, "--- REMOVED ---\n");
        {
            GSList *ps = rcdr->packages_removed;
            int i = 0;
            while (ps) {
                RCPackage *pkg = (RCPackage *) ps->data;
                rc_debug (RC_DEBUG_LEVEL_INFO, "% 3d: %s\n", i++, rc_package_spec_to_str (&pkg->spec));
                ps = ps->next;
            }
        }

        rc_debug (RC_DEBUG_LEVEL_INFO, "--- MISSING ---\n");
        {
            g_hash_table_foreach (rcdr->require_hash, print_missing_hash, NULL);
            rc_debug (RC_DEBUG_LEVEL_INFO, "Total missing: %d\n", rcdr->unprovided_packages);
        }

        rc_debug (RC_DEBUG_LEVEL_INFO, "--- FAILED ---\n");
        {
            GSList *ps = rcdr->failed_daps;
            int i = 0;
            while (ps) {
                RCDepAndPackage *dap = (RCDepAndPackage *) ps->data;
                rc_debug (RC_DEBUG_LEVEL_INFO, "% 3d: ", i++);
                debug_deps_dump_dap (dap);
                ps = ps->next;
            }
            rc_debug (RC_DEBUG_LEVEL_INFO, "Total failed: %d\n", i);
        }

        rc_debug (RC_DEBUG_LEVEL_INFO, "--- CONFLICTS ---\n");
        {
            GSList *cs = rcdr->packages_conflicting;
            while (cs) {
                RCDepConflict *conf = (RCDepConflict *) cs->data;
                print_conflict (RC_DEBUG_LEVEL_INFO, conf);
                cs = cs->next;
            }
        }
    }

    res = g_new0 (RCDepResolutionResult, 1);

    if (passno == 200) {
        res->abort = TRUE;
    }

    res->packages_to_install = g_slist_copy (rcdr->packages_added);
    res->packages_to_remove = g_slist_copy (rcdr->packages_removed);
    res->packages_conflicting = g_slist_copy (rcdr->packages_conflicting);

    g_hash_table_foreach (rcdr->require_hash, unprovided_hash_to_list, &res->unresolved_reqs);
    
    /* need to put some stuff in the output, NULL for now */
    return res;
}

static void
unprovided_hash_to_list (gpointer key, gpointer value, gpointer data)
{
    //    gchar *name = (gchar *) key;
    RCDepRequireSList *ll = (GSList *) value;
    RCDepRequireSList **undep_reqs = (RCDepRequireSList **) data;

    while (ll) {
        RCDepRequire *req = (RCDepRequire *) ll->data;
        if (req->top_provide != NULL) {
            ll = ll->next;
            continue;
        }

        if (rc_package_dep_slist_is_fully_weak (req->deps)) {
            ll = ll->next;
            continue;
        }

	if (req->top_provide == REQUIRE_ALL_REMOVED) {
            ll = ll->next;
            continue;
        }

        *undep_reqs = g_slist_append (*undep_reqs, req);
        ll = ll->next;
    }
}


void
rc_dep_resolution_result_free (RCDepResolutionResult *res)
{
    g_slist_free (res->packages_to_install);
    g_slist_free (res->packages_to_remove);
    g_free (res);
}

/*** private functions ***/

static void
populate_dr_with_system (RCDepResolution *rcdr,
                         RCPackageSList *system_packages)
{
    GSList *pkgs;

    DEPS_ENTER;

    if (!system_packages) {
        DEPS_EXIT;
        return;
    }

    /* First push the provides */
    pkgs = system_packages;
    while (pkgs) {
        RCPackage *pkg = (RCPackage *) pkgs->data;
        package_push_provide_list (rcdr, pkg, pkg->provides, FALSE);
        pkgs = pkgs->next;
    }

    /* Then form the requires */
    pkgs = system_packages;
    while (pkgs) {
        RCPackage *pkg = (RCPackage *) pkgs->data;
        package_push_require_list (rcdr, pkg, pkg->requires, FALSE);
        pkgs = pkgs->next;
    }

    pkgs = system_packages;
    while (pkgs) {
        RCPackage *pkg = (RCPackage *) pkgs->data;
        package_push_conflict_list (rcdr, pkg, pkg->conflicts, FALSE);
        pkgs = pkgs->next;
    }

    /* Then tie up the requires with the provides */
    g_hash_table_foreach (rcdr->require_hash, deps_provide_require_iterator, rcdr);

    DEPS_EXIT;
}

static void
deps_add_package (RCDepResolution *rcdr,
                  RCPackage *pkg,
                  gboolean force)
{
    RCDepProvideSList *provs;

    DEPS_ENTER;
    DEPS_DEBUG_PKG("pkg", pkg);

    deps_bt (pkg->spec.name);

    /* Pull older packages out, if any */
    provs = g_hash_table_lookup (rcdr->provide_hash,
                                 pkg->spec.name);
    while (provs) {
        RCDepProvide *prov = (RCDepProvide *) provs->data;

        if (pkg == prov->provided_by) {
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "Trying to add already installed/inserted package '%s'\n",
                      pkg->spec.name);
            return;
        }

        if (pkg != prov->provided_by &&
            !strcmp (pkg->spec.name, prov->provided_by->spec.name) &&
            (rc_packman_version_compare (das_global_packman, &pkg->spec, &prov->provided_by->spec) > 0))
        {
            /* pkg is the same as what this provides, and is newer,
             * so we get rid of the older package
             */
            deps_remove_package (rcdr, prov->provided_by, FALSE);
//            deps_untangle_package (rcdr, prov->provided_by);
        }

        provs = provs->next;
    }

    package_push_provide_list (rcdr, pkg, pkg->provides, force);
    package_push_require_list (rcdr, pkg, pkg->requires, force);
    package_push_conflict_list (rcdr, pkg, pkg->conflicts, force);

    deps_update_install_list (rcdr, pkg);

    DEPS_EXIT;
}

static void
deps_remove_package (RCDepResolution *rcdr,
                     RCPackage *pkg,
                     gboolean as_conflict)
{
    RCPackage *real_pkg = NULL;
    GSList *iter;

    DEPS_ENTER;
    DEPS_DEBUG_PKG ("pkg", pkg);

    if (g_slist_find (rcdr->packages_removed, pkg)) {
        /* already removed */
        DEPS_EXIT;
        return;
    }

    /* We need to lookup the ptr to what we know this package as */
    iter = rcdr->packages_added;
    while (iter) {
        RCPackage *p = (RCPackage *) iter->data;
        if (rc_package_spec_equal (&p->spec, &pkg->spec)) {
            real_pkg = p;
            break;
        }
        iter = iter->next;
    }

    iter = g_hash_table_lookup (rcdr->provide_hash, pkg->spec.name);
    if (!real_pkg) {
        RCDepProvideSList *provs = g_hash_table_lookup (rcdr->provide_hash, pkg->spec.name);
        while (provs) {
            RCDepProvide *prov = (RCDepProvide *) provs->data;
            if (rc_package_spec_equal (&prov->provided_by->spec, &pkg->spec)) {
                real_pkg = prov->provided_by;
                break;
            }
            provs = provs->next;
        }
    }

    if (real_pkg) {
        /* Find this in the must_provides */
        iter = rcdr->must_provide_list;
        while (iter) {
            RCDepAndPackage *dap = (RCDepAndPackage *) iter->data;
            if (dap->package == real_pkg) {
                GSList *nextp = iter->next;
                g_free (iter->data); /* dap */
                rcdr->must_provide_list = g_slist_remove_link (rcdr->must_provide_list, iter);
                iter = nextp;
                continue;
            }
            iter = iter->next;
        }

        deps_update_remove_list (rcdr, real_pkg);
        package_mark_provides_as_removed (rcdr, real_pkg, real_pkg->provides, as_conflict);

//        deps_untangle_package (rcdr, real_pkg);
//
//        if (g_slist_find (rcdr->packages_added, real_pkg)) {
//            rcdr->packages_added = g_slist_remove (rcdr->packages_added, real_pkg);
//        } else {
//            rcdr->packages_removed = g_slist_append (rcdr->packages_removed, real_pkg);
//        }
//    } else {
//        g_warning ("Couldn't find equivalent local deps package for '%s' while removing!\n",
//                   rc_package_spec_to_str (&pkg->spec));
    }

    DEPS_EXIT;
}

static void
package_mark_provides_as_removed (RCDepResolution *rcdr,
                                  RCPackage *pkg,
                                  RCPackageDepSList *pkg_provs,
                                  gboolean as_conflict)
{
    DEPS_ENTER;
    DEPS_DEBUG_PKG("pkg", pkg);
    DEPS_DEBUG_PDSL("pkg_provs", pkg_provs);

    while (pkg_provs) {
        RCPackageDep *dep = (RCPackageDep *) pkg_provs->data;
        while (dep) {
            RCPackageDepItem *depi = (RCPackageDepItem *) dep->data;
            RCDepProvideSList *provs;

            provs = g_hash_table_lookup (rcdr->provide_hash,
                                         depi->spec.name);
            while (provs) {
                RCDepProvide *prov = (RCDepProvide *) provs->data;
                if (prov->provided_by == pkg) {
                    RCDepRequireSList *reqs;
                    rc_debug (RC_DEBUG_LEVEL_DEBUG, "PROV removing: %s\n", prov->name);
                    prov->force = FALSE;
                    prov->removed = TRUE;
                    reqs = prov->required_by;
                    while (reqs) {
                        RCDepRequire *req = (RCDepRequire *) reqs->data;
                        /* HACK HACK HACK */
                        if (req->top_provide == NULL) {
                            prov->required_by = g_slist_remove (prov->required_by, req);
                            reqs = prov->required_by;
                            continue;
                        }

                        deps_require_unprovide (rcdr, req, as_conflict);

                        /* NOT reqs->next, since that will get us a bogus pointer,
                         * since unprovide will remove it from the list
                         */
                        reqs = prov->required_by;
                    }
                }
                provs = provs->next;
            }
            dep = dep->next;
        }
        pkg_provs = pkg_provs->next;
    }

    DEPS_EXIT;
}

#if NOTUSED
static gboolean
deps_remove_require (RCDepResolution *rcdr, RCDepRequire *req, gboolean hash_remove)
{
    gchar *orig_name;
    RCDepRequireSList *reqs;
    gboolean ret = FALSE;

    gchar *the_name;

    DEPS_ENTER;
    DEPS_DEBUG_REQ ("req", req);
    DEPS_DEBUG_BOOL ("hash_remove", hash_remove);
    
    the_name = deps_join_names (req->names);
    g_hash_table_lookup_extended (rcdr->require_hash, the_name,
                                  (gpointer *) &orig_name, (gpointer *) &reqs);
    g_free (the_name);
    reqs = g_slist_remove (reqs, req);

    if (reqs) {
        g_hash_table_insert (rcdr->require_hash, orig_name, reqs);
        deps_require_free (req);
    } else {
        if (hash_remove) {
            g_hash_table_remove (rcdr->require_hash, orig_name);
            g_free (orig_name);
            deps_require_free (req);
        } else {
            ret = TRUE;
        }
    }

    DEPS_DEBUG_BOOL("ret", ret);
    DEPS_EXIT;
    return ret;
}
#endif

static void
deps_require_free (RCDepRequire *req)
{
    GSList *iter;

    iter = req->names;
    while (iter) {
        g_free (iter->data);
        iter = iter->next;
    }
    g_slist_free (req->names);

    iter = req->required_list;
    while (iter) {
        g_free (iter->data);    /* DAP */
        iter = iter->next;
    }
    g_slist_free (req->required_list);

    iter = req->deps;
    while (iter) {
        rc_package_dep_free (iter->data); /* PD */
        iter = iter->next;
    }
    g_slist_free (req->deps);

    rc_package_dep_slist_free (req->deps);
}

static void
print_provide (int dlev, RCDepProvide *prov)
{
    rc_debug (dlev,  "    +--: [%s] by %s\n", debug_rc_package_dep_item_str (prov->provision),
             rc_package_spec_to_str (&prov->provided_by->spec));
    if (prov->required_by) {
        GSList *l = prov->required_by;
        rc_debug (dlev, "  req'd: ");
        while (l) {
            RCDepRequire *req = (RCDepRequire *) l->data;
            GSList *nl = req->names;
            while (nl) {
                rc_debug (dlev, "%s", (gchar *) nl->data);
                if (nl->next) rc_debug (dlev, ":");
                nl = nl->next;
            }
            rc_debug (dlev, " ");
            l = l->next;
        }
        rc_debug (dlev, "\n");
    }
}

static void
print_provide_hash (gpointer key, gpointer value, gpointer data)
{
    gchar *name = (gchar *) key;
    GSList *prov_list = (GSList *) value;

    rc_debug (RC_DEBUG_LEVEL_DEBUG, "Provide: %s\n", name);
    while (prov_list) {
        RCDepProvide *prov = (RCDepProvide *) prov_list->data;
        print_provide (RC_DEBUG_LEVEL_DEBUG, prov);
        prov_list = prov_list->next;
    }
}

static void
print_require (int dlev, RCDepRequire *req)
{
    GSList *l;

    rc_debug (dlev,  "--names: ");
    l = req->names;
    while (l) {
        rc_debug (dlev,  "%s ", (gchar *) l->data);
        l = l->next;
    }
    rc_debug (dlev,  "\n");

    rc_debug (dlev, "   deps: ");
    l = req->deps;
    while (l) {
        RCPackageDep *rcpd = (RCPackageDep *) l->data;
        debug_rc_package_dep_dump (dlev, rcpd);
        if (l->next) {
            rc_debug (dlev, "      || ");
        }
        l = l->next;
    }
        
    rc_debug (dlev, "   list: ");
    l = req->required_list;
    while (l) {
        RCDepAndPackage *dap = (RCDepAndPackage *) l->data;
            
        rc_debug (dlev, "%s [%d]-- ", rc_package_spec_to_str (&dap->package->spec), dap->force);
        debug_rc_package_dep_dump (dlev, dap->dep);
//        rc_debug (dlev, "\n");  // printed by the debug dump
        if (l->next) {
            rc_debug (dlev, "         ");
        }
        l = l->next;
    }

    rc_debug (dlev, "  top_p: ");
    if (req->top_provide) {
        if (req->top_provide == REQUIRE_HAS_CONFLICT)
            rc_debug (dlev, "HAS CONFLICT\n");
        else if (req->top_provide == REQUIRE_ALL_REMOVED)
            rc_debug (dlev, "ALL REMOVED\n");
        else
            rc_debug (dlev, "%s [%s]\n", req->top_provide->name,
                     rc_package_spec_to_str (&req->top_provide->provided_by->spec));
    } else {
        rc_debug (dlev, "NULL\n");
    }
}

static void
print_require_hash (gpointer key, gpointer value, gpointer data)
{
    gchar *name = (gchar *) key;
    RCDepRequireSList *ll = (GSList *) value;

    rc_debug (RC_DEBUG_LEVEL_DEBUG, "Require: %s [%d in list]\n", name, g_slist_length (ll));
    while (ll) {
        RCDepRequire *req = (RCDepRequire *) ll->data;
        print_require (RC_DEBUG_LEVEL_DEBUG, req);
        ll = ll->next;
    }
}

static void
print_conflict (int dlev, RCDepConflict *conf)
{
    rc_debug (dlev, "CONF: %s [%s] conflicts with %s\n",
             conf->what ? rc_package_spec_to_str (&conf->what->spec) : "???",
             conf->what_str,
             conf->with_other_package ? rc_package_spec_to_str (&conf->with_other_package->spec) : "???");
    rc_debug (dlev, "type: %d\n", conf->type);
    rc_debug (dlev, "what: %s\n", conf->what_conflicts ?
             debug_rc_package_dep_item_str (conf->what_conflicts) :
             "???");
}

static void
print_missing_hash (gpointer key, gpointer value, gpointer data)
{
//    gchar *name = (gchar *) key;
    RCDepRequireSList *ll = (GSList *) value;

    while (ll) {
        RCDepRequire *req = (RCDepRequire *) ll->data;
        if (req->top_provide != NULL) {
            ll = ll->next;
            continue;
        }

        if (rc_package_dep_slist_is_fully_weak (req->deps)) {
            ll = ll->next;
            continue;
        }

	if (req->top_provide == REQUIRE_ALL_REMOVED) {
            ll = ll->next;
            continue;
        }

        print_require (RC_DEBUG_LEVEL_INFO, req);
        ll = ll->next;
    }
}



static RCDepProvide *
deps_provide_new (gchar *name, RCPackageDepItem *provision, RCPackage *provided_by)
{
    RCDepProvide *prov;

    DEPS_ENTER;

    prov = g_new0 (RCDepProvide, 1);
    prov->name = g_strdup (name);
    prov->provision = provision;
    prov->provided_by = provided_by;
    prov->required_by = NULL;

    if (rc_deps_debug_level > 20) {
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "+PROV: %s (%p)\n", name, provided_by);
    }

    DEPS_EXIT;
    return prov;
}

static RCDepRequire *
deps_require_new (GSList *names, RCPackageDep *dep)
{
    RCDepRequire *req;

    req = g_new0 (RCDepRequire, 1);
    req->names = names;
    req->deps = deps_dep_to_or_slist (dep);

    return req;
}

/* This could be sped up by a LOT */
static RCPackageDepItemSList *
find_debian_self_conflicts (RCPackage *pkg)
{
    RCPackageDepItemSList *out_list = NULL;

    RCPackageDepSList *provs = pkg->provides;

    while (provs) {
        RCPackageDep *prov = (RCPackageDep *) provs->data;
        while (prov) {
            RCPackageDepItem *prov_item = (RCPackageDepItem *) prov->data;
            RCPackageDepSList *confs = pkg->conflicts;

            /* Now loop through the conflicts */
            while (confs) {
                RCPackageDep *conf = (RCPackageDep *) confs->data;
                while (conf) {
                    RCPackageDepItem *conf_item = (RCPackageDepItem *) conf->data;
                    if (rc_package_dep_item_equal (conf_item, prov_item)) {
                        out_list = g_slist_append (out_list, prov_item);
                    }
                    conf = conf->next;
                }
                confs = confs->next;
            }
            prov = prov->next;
        }
        provs = provs->next;
    }

    return out_list;
}

static void
package_push_provide_list (RCDepResolution *rcdr,
                           RCPackage *pkg,
                           RCPackageDepSList *provides,
                           gboolean force)
{
    RCPackageDepItemSList *onlymes;

    onlymes = find_debian_self_conflicts (pkg);

    while (provides) {
        RCPackageDep *pdep = (RCPackageDep *) provides->data;

        if (onlymes && g_slist_find_custom (onlymes, ((RCPackageDepItem *) pdep->data)->spec.name,
                                            package_find_by_name))
        {
            RCDepAndPackage *new_dap;
            /* and set something else? */

            new_dap = g_new0 (RCDepAndPackage, 1);
            new_dap->package = pkg;
            new_dap->dep = pdep;
            new_dap->force = force;

            rcdr->must_provide_list = g_slist_append (rcdr->must_provide_list, new_dap);
            force = TRUE;
        }

        package_push_provide (rcdr, pkg, pdep, force);
        provides = provides->next;
    }

    g_slist_free (onlymes);
}

static void
package_push_require_list (RCDepResolution *rcdr,
                           RCPackage *pkg,
                           RCPackageDepSList *requires,
                           gboolean force)
{
    /* Each package needs to require itself first */
    RCPackageDep *self_req = 
        rc_package_dep_new_with_item
        (rc_package_dep_item_new_from_spec (&pkg->spec, RC_RELATION_EQUAL));
    package_push_require (rcdr, pkg, self_req, force);

    deps_bt (pkg->spec.name);

    while (requires) {
        RCPackageDep *pdep = (RCPackageDep *) requires->data;
        package_push_require (rcdr, pkg, pdep, force);
        requires = requires->next;
    }
}

static void
package_push_conflict_list (RCDepResolution *rcdr,
                            RCPackage *pkg,
                            RCPackageDepSList *conflicts,
                            gboolean force)
{
    DEPS_ENTER;
    DEPS_DEBUG_PKG("pkg", pkg);
    DEPS_DEBUG_PDSL("conflicts", conflicts);

    deps_bt (pkg->spec.name);

    while (conflicts) {
        RCPackageDep *pdep = (RCPackageDep *) conflicts->data;
        RCPackageDep *orig_pdep = g_slist_copy (pdep);
        RCPackageDep *orig_pdep_it = orig_pdep;
        RCPackageDep *inv_pdep;


        /* debian self conflicts */
        /* This could be done better, but this code should go away anyway
         * once these are handled correctly. We just nuke them.
         */

        /* don't get rid of them. do get rid of them. what to do, what to do... */
#if 1
//        rc_debug (RC_DEBUG_LEVEL_DEBUG, "^^^^ PKG: %s\n", pkg->spec.name);
        while (orig_pdep_it) {
            RCPackageDepItem *di = (RCPackageDepItem *) orig_pdep_it->data;
            gboolean nukeit = FALSE;

//            rc_debug (RC_DEBUG_LEVEL_DEBUG, "^^^^ %s\n", di->spec.name);
            if (((di->relation == RC_RELATION_EQUAL) || (di->relation == RC_RELATION_ANY)) &&
                (strcmp (di->spec.name, pkg->spec.name) == 0))
//                rc_package_spec_equal (&di->spec, &pkg->spec))
            {
                nukeit = TRUE;
            } else {
                RCPackageDepSList *provs = pkg->provides;
                while (provs) {
                    RCPackageDep *pd = (RCPackageDep *) provs->data;
                    while (pd) {
                        RCPackageDepItem *pdi = (RCPackageDepItem *) pd->data;
                        if (((pdi->relation == di->relation) || (pdi->relation == RC_RELATION_ANY)) &&
                            (strcmp (di->spec.name, pdi->spec.name) == 0))
//                            rc_package_spec_equal (&di->spec, &pdi->spec))
                        {
                            nukeit = TRUE;
                            break;
                        }
                        pd = pd->next;
                    }
                    if (nukeit) break;
                    provs = provs->next;
                }
            }

            if (nukeit) {
                GSList *link = orig_pdep_it;
                orig_pdep = g_slist_remove_link (orig_pdep, orig_pdep_it);
                g_slist_free_1 (link);
                orig_pdep_it = orig_pdep;
                continue;
            }
                
            orig_pdep_it = orig_pdep_it->next;
        }

        if (!orig_pdep) {
            conflicts = conflicts->next;
            continue;
        }

#endif

        if (rc_deps_debug_level > 30) {
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "Original pdep: ");
            debug_rc_package_dep_dump (RC_DEBUG_LEVEL_DEBUG, orig_pdep);
        }

        inv_pdep = rc_package_dep_invert (orig_pdep);
        g_slist_free (orig_pdep);
        rc_package_dep_weaken (inv_pdep);

        if (rc_deps_debug_level > 30) {
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "Inverted pdep: ");
            debug_rc_package_dep_dump (RC_DEBUG_LEVEL_DEBUG, inv_pdep);
        }

        package_push_require (rcdr, pkg, inv_pdep, force);

        /* keep this around so we can free it in the destory */
        rcdr->conflict_deps = g_slist_append (rcdr->conflict_deps, inv_pdep);

        conflicts = conflicts->next;
    }

    DEPS_EXIT;
}

static void
package_push_provide (RCDepResolution *rcdr,
                      RCPackage *pkg,
                      RCPackageDep *pdep,
                      gboolean force)
{
    DEPS_ENTER;
    DEPS_DEBUG_PKG ("pkg", pkg);
    DEPS_DEBUG_PD ("pdep", pdep);
    DEPS_DEBUG_BOOL ("force", force);

    while (pdep) {
        RCPackageDepItem *pdepi = (RCPackageDepItem *) pdep->data;
        RCDepProvide *prov;
        gboolean found;

        gchar *name = NULL;
        RCDepProvideSList *dpsl = NULL;

        prov = deps_provide_new (pdepi->spec.name, pdepi, pkg);
        prov->force = force;

        if (prov->virtual != VIRTUAL_UNSET) {
            if (!rc_package_spec_equal (&pdepi->spec, &pkg->spec))
                prov->virtual = VIRTUAL_SET;
            else
                prov->virtual = VIRTUAL_UNSET;
        }

        found = g_hash_table_lookup_extended (rcdr->provide_hash, pdepi->spec.name,
                                              (gpointer *) &name, (gpointer *) &dpsl);
        if (!found) {
            /* This provide doesn't exist. Create a new entry for it. */
            name = g_strdup (prov->name);
            dpsl = g_slist_append (NULL, prov);
            g_hash_table_insert (rcdr->provide_hash, name, dpsl);
        } else {
            /* This one was found. We need to check and see if any of the
             * packages that provide this have the same name as pkg -- if so,
             * we need to unprovide those things and resolve the upstream req's.
             */
            RCDepProvideSList *dpsl_it = dpsl;
            dpsl = g_slist_append (dpsl, prov);
            while (dpsl_it) {
                RCDepProvide *prov = (RCDepProvide *) dpsl_it->data;
                if (prov != prov && !strcmp (prov->provided_by->spec.name, pkg->spec.name)) {
                    /* Same package */
                    deps_provide_unprovide (rcdr, prov);
                }
                dpsl_it = dpsl_it->next;
            }

//        rc_debug (RC_DEBUG_LEVEL_DEBUG, "provide inserting: '%s' -> %p\n", name, dpsl);
            g_hash_table_insert (rcdr->provide_hash, name, dpsl);
        }

        pdep = pdep->next;
    }

    DEPS_EXIT;
}

static gint
packagedepitem_compare (gconstpointer ptra, gconstpointer ptrb)
{
    RCPackageDepItem *a = (RCPackageDepItem *) ptra;
    RCPackageDepItem *b = (RCPackageDepItem *) ptrb;

    return rc_packman_version_compare (das_global_packman, &a->spec, &b->spec);
}

static gchar *
deps_join_names (GSList *name_list)
{
    GString *new_string = g_string_sized_new (100);
    gchar *ret;
    while (name_list) {
        g_string_append (new_string, (gchar *) name_list->data);
        if (name_list->next)
            g_string_append (new_string, ":");
        name_list = name_list->next;
    }
    ret = new_string->str;
    g_string_free (new_string, FALSE);
    return ret;
}


/*
 */
static gchar *
deps_make_names (RCPackageDep *dep, GSList **list_out)
{
    gchar *ret;
    GSList *ret_list = NULL;
    RCPackageDep *sorted, *sorted_dep;
    GString *gs = g_string_sized_new (50);
    gchar *last_name = NULL;

    sorted_dep = g_slist_sort (g_slist_copy (dep), packagedepitem_compare);
    sorted = sorted_dep;
    while (sorted) {
        RCPackageDepItem *pdepi = (RCPackageDepItem *) sorted->data;

        if (last_name && !strcmp (pdepi->spec.name, last_name)) {
            dep = dep->next;
            continue;
        } else if (!last_name) {
            last_name = pdepi->spec.name;
        }

        if (list_out)
            ret_list = g_slist_append (ret_list, g_strdup (pdepi->spec.name));
        g_string_append (gs, pdepi->spec.name);
        if (sorted->next)
            g_string_append (gs, ":"); /* : is used since it's not common in pkg names (illegal in rpm?) */
        sorted = sorted->next;
    }

    g_slist_free (sorted_dep);
    ret = g_strdup (gs->str);
    if (list_out)
        *list_out = ret_list;
    g_string_free (gs, TRUE);
    return ret;
}

static void
deps_remerge_reqs (RCDepResolution *rcdr,
                   RCDepRequire *req)
{
    RCDepAndPackageSList *daps = req->required_list;

    DEPS_ENTER;
    DEPS_DEBUG_REQ("req", req);

    if (req->deps) {
        rc_package_dep_slist_free (req->deps);
        req->deps = NULL;
    }

    while (daps) {
        RCDepAndPackage *dap = (RCDepAndPackage *) daps->data;
        deps_require_merge_requirement (req, dap->dep);
        daps = daps->next;
    }

    DEPS_EXIT;
}

static gboolean
deps_require_merge_requirement (RCDepRequire *req,
                                RCPackageDep *rcd)
{
    GSList *p;
    RCPackageDepSList *new_deps_list = NULL;
    gboolean still_provide = TRUE;

    DEPS_ENTER;
    DEPS_DEBUG_PD("rcd", rcd);

    if (req->deps == NULL) {
        req->deps = deps_dep_to_or_slist (rcd);
        /* nothing to merge with. we're happy. */
        DEPS_DEBUG_BOOL("ret", TRUE);
        DEPS_EXIT;
        return TRUE;
    }

//    rc_debug (RC_DEBUG_LEVEL_DEBUG, "deps_require_merge_req: starting...\n");
    p = rcd;
    while (p) {
        RCPackageDepItem *di = (RCPackageDepItem *) p->data;
        /* each di in this list is an OR */
        RCPackageDepSList *orig_deps = req->deps;

        while (orig_deps) {
            RCPackageDep *orig_dep = (RCPackageDep *) orig_deps->data;
            RCPackageDep *new_dep = NULL;
            gboolean merged_older = FALSE;
            gboolean merged_newer = FALSE;

            /* Each of the orig_dep items SHOULD have the same name, so we
             * check the first one and then ignore it if it doesn't match */

            if (strcmp (((RCPackageDepItem *)orig_dep->data)->spec.name, di->spec.name))
            {
                gchar *cur_name = ((RCPackageDepItem *)orig_dep->data)->spec.name;
                RCPackageDep *dep_it = rcd;
                /* We need to go through rcd and make sure that cur_name isn't in there,
                 * so that we know to copy this one
                 */
                while (dep_it) {
                    RCPackageDepItem *dit = (RCPackageDepItem *) dep_it->data;
                    if (!strcmp (dit->spec.name, cur_name)) break;
                    dep_it = dep_it->next;
                }
                if (!dep_it) {
                    new_deps_list = g_slist_append (new_deps_list, rc_package_dep_copy (orig_dep));
                }
                orig_deps = orig_deps->next;
                continue;
            }

            /* Each orig_dep is an OR */
            while (orig_dep) {
                RCPackageDepItem *odi = (RCPackageDepItem *) orig_dep->data;
                /* Each odi is an AND */

                /* We need to go through and merge each one */
                switch (rc_package_dep_item_is_subset (di, odi)) {
                    case 1:
                        /* Can be merged; use new */
                        if (odi == NULL) abort ();
                        new_dep = g_slist_append (new_dep, rc_package_dep_item_copy (odi));
                        merged_newer = TRUE;
//                        rc_debug (RC_DEBUG_LEVEL_DEBUG, "case 1: list: %d\n", g_slist_length (new_dep));
                        break;
                    case 0:
                        /* Can be merged; use old */
                        if (!merged_older) {
                            if (di == NULL) abort ();
                            new_dep = g_slist_append (new_dep, rc_package_dep_item_copy (di));
                            merged_older = TRUE;
                        }
//                        rc_debug (RC_DEBUG_LEVEL_DEBUG, "case 0: list: %d\n", g_slist_length (new_dep));
                        break;
                    case -1:
                        /* Don't add the new one if it's already been merged as NEW --
                         * it's already in the list */
                        if (!merged_older) {
                            if (di == NULL) abort ();
                            new_dep = g_slist_append (new_dep, rc_package_dep_item_copy (di));
                            merged_older = TRUE;
                        }
                        if (odi == NULL) abort ();
                        new_dep = g_slist_append (new_dep, rc_package_dep_item_copy (odi));

//                        rc_debug (RC_DEBUG_LEVEL_DEBUG, "case -1: list: %d\n", g_slist_length (new_dep));
                        /* There was a conflict, so we tell the caller to clear the
                         * provide and refind one
                         */
                        still_provide = FALSE;
                        break;
                }

                if (merged_newer) {
                    /* This means that the provide might not be valid any more. */
                    still_provide = FALSE;
                }

                orig_dep = orig_dep->next;
            }

            new_deps_list = g_slist_append (new_deps_list, new_dep);

            orig_deps = orig_deps->next;
        }

        p = p->next;
    }

    rc_package_dep_slist_free (req->deps);
    req->deps = new_deps_list;

    /* This might be better handled by simply checking the current
     * provided_by against the new list
     */
    DEPS_DEBUG_BOOL("ret(still_provide)", still_provide);
    DEPS_EXIT;
    return still_provide;
}

static gboolean
deps_require_add_package (RCDepRequire *req,
                          RCPackage *pkg,
                          RCPackageDep *rcd,
                          gboolean force)
{
    RCDepAndPackage *dap;

    DEPS_ENTER;
    DEPS_DEBUG_PKG ("pkg", pkg);
    DEPS_DEBUG_PD ("rcd", rcd);

    dap = g_new0 (RCDepAndPackage, 1);
    dap->package = pkg;
    dap->dep = rcd;
    dap->force = force;

//    rc_debug (RC_DEBUG_LEVEL_DEBUG, "*** Package: %s\n", pkg->spec.name);

    req->required_list = g_slist_append (req->required_list, dap);

    if (!deps_require_merge_requirement (req, rcd)) {
        /* The merge failed -- we got a -1 back from the merge function */
        DEPS_DEBUG_BOOL("ret", FALSE);
        DEPS_EXIT;
        return FALSE;
    }

    DEPS_DEBUG_BOOL("ret", TRUE);
    DEPS_EXIT;
    return TRUE;
}

static void
deps_require_unprovide (RCDepResolution *rcdr, RCDepRequire *req, gboolean as_conflict)
{
    RCDepProvide *prov;

    DEPS_ENTER;
    DEPS_DEBUG_REQ ("req", req);

    if (req->top_provide == REQUIRE_HAS_CONFLICT ||
        req->top_provide == REQUIRE_ALL_REMOVED)
    {
        if (as_conflict)
            req->top_provide = REQUIRE_HAS_CONFLICT;
        else
            req->top_provide = NULL;

        DEPS_EXIT;
        return;
    }

    prov = req->top_provide;
    if (!prov) {
        DEPS_EXIT;
        return;                 /* wasn't provided in the first place! */
    }

    if (rc_deps_debug_level >= 20) {
        GSList *l = req->names;
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "Unprovide: unproviding ");
        while (l) { rc_debug (RC_DEBUG_LEVEL_DEBUG, "%s%s", (gchar *)l->data, l->next?":":""); l = l->next; }
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "\n");
    }

    prov->required_by = g_slist_remove (prov->required_by, req);

    if (as_conflict)
        req->top_provide = REQUIRE_HAS_CONFLICT;
    else
        req->top_provide = NULL;

#if 0
    /* XXX Gross hack attempt */
    {
        RCDepAndPackageSList *daps = req->required_list;
        while (daps) {
            RCDepAndPackage *dap = (RCDepAndPackage *) daps->data;
            RCDepRequire *found_req;
            found_req = g_hash_table_lookup (rcdr->require_hash, dap->package->spec.name);
            if (found_req && found_req->top_provide)
                deps_require_unprovide (rcdr, found_req, FALSE);
            daps = daps->next;
        }
    }
#endif

    DEPS_EXIT;
}


static void
package_push_require (RCDepResolution *rcdr,
                      RCPackage *pkg,
                      RCPackageDep *pdep,
                      gboolean forced)
{
    gchar *name;
    gchar *orig_name;
    GSList *names;
    RCDepRequire *req;
    RCDepRequireSList *req_list;

    DEPS_ENTER;
    DEPS_DEBUG_PKG ("pkg", pkg);
    DEPS_DEBUG_PD ("pdep", pdep);

    name = deps_make_names (pdep, &names);

    if (!g_hash_table_lookup_extended (rcdr->require_hash, name,
                                       (gpointer *) &orig_name, (gpointer *) &req_list))
    {
        GSList *names_it;
        /* This thing doesn't exist. Make it. */
        req = deps_require_new (names, pdep);
        if (!deps_require_add_package (req, pkg, pdep, forced)) {
            deps_require_unprovide (rcdr, req, FALSE);
        }

        /* Then stick it in the hash */
        names_it = names;
        while (names_it) {
            gchar *cur_name = (gchar *) names_it->data;
            RCDepRequireSList *req_list_indiv;
            gchar *out_name;
            gboolean found;

            found = g_hash_table_lookup_extended (rcdr->require_hash, cur_name,
                                                  (gpointer *) &out_name, (gpointer *) &req_list_indiv);

            if (!found) {
                req_list_indiv = NULL;
                out_name = g_strdup (cur_name);
            }

            req_list_indiv = g_slist_append (req_list_indiv, req);
            g_hash_table_insert (rcdr->require_hash, out_name, req_list_indiv);

            names_it = names_it->next;
        }

        orig_name = name;
        req_list = g_slist_append (NULL, req);
    } else {
        /* It already exists. This means that we just need to update it,
         * as it's already in the right lists [since we search for the
         * full name in all it's OR glory]
         */
        GSList *list_it = req_list;

        /* if this is an OR, there should be only one element in list_it */
        while (list_it) {
            req = (RCDepRequire *) list_it->data;

            if (!deps_require_add_package (req, pkg, pdep, forced)) {
                deps_require_unprovide (rcdr, req, FALSE);
            }

            list_it = list_it->next;
        }
    }

    g_hash_table_insert (rcdr->require_hash, orig_name, req_list);
    if (orig_name != name)
        g_free (name);

    DEPS_EXIT;
}



/************************


This is the magic random throughts comment that's been sneakily placed in the
middle of code to give the illusion that this code has no issues.

Do we match up provides with requires as we're adding the requires to the list,
or do we wait until the end, and then just go through all the requires and find
the provide that matches them? [at the end]

Life would be much easier if we didn't have to deal with stupid OR dependencies.
How do I handle these when merging requirements? [we don't. We do the
nasty OR-AND-SList business.]

dependency: exim-mua-pine
if things need exim > 0.5 | mua | pine > 1.0, and something else needs
exim > 1.0 | mua | pine > 0.5, this should be merged to exim > 1.0 | mua | pine > 1.0.
[maybe?]

In a DepRequire
the deps DepSList isn't the same as a traditional DepSList in that each element in the
DepSList is an 'OR', and then each Dep's DepItem's are to be considered in an AND.
The usual (RCPackage) way is the opposite. [Or at least this is the way it shuld work.]

conflicts -- weak relations only take us so far.

If we have package FOO that conflicts with BAR, and nothing requires BAR,
FOO's weak requirement on BAR will not be satisfied, unless weak requirements
are still checked against the provides list (i.e. if something provides/is BAR)

-- conflicts need to be done --
started... weak dependends should be cached; a flag in the req structure when
something is created based solely on weak deps, maybe

[X] debian self conflicts need to be handled
conflicts -- I think it's safe to ignore debian self-conflicts. But I think they
can be handled by a flag in the RCDepProvide saying "to fulfill this, you must use
this provide".

[X] write pkginfo_find_package that takes a Dep, not a DepItem, and takes
    a boolean saying whether to treat it as an AND or OR
[X] figure out a way to do reverse provides (the foo and foo-devel case)

Removes are going to be fun. There really isn't a difference between
removing a package from a resolution and removing an already installed
package; it's basically a matter of whether you add it to the remove list
or not. (so, if it's on the pkgs_added list, remove it from there, but don't
put it on the remove list.) The hard part comes in deciding how to handle
things which depend on something that you remove. Once you remove something
and everything it provides from the correct lists, the packages that
depended on this thing (and when there's nothing to step in to fulfill the
dependency) are going to have null top_p's. This means that as it stands,
the code will helpfully try to go out and find a package to install that
will fulfill this dependency. But we don't want it to do that, since we're
trying to uninstall the package in the first place! One way to fix this is
to have the code that does the remove actually walk up the path -- but that's
broken in and of itself, in case there is something that is not to be removed
which will provide the same thing... i.e. gv and xpdf both provide pdf-viewer.
Package foo depends on pdf-viewer. You remove gv. You should NOT remove foo,
because xpdf is still there. So this means that we need a way
to flag a provide as "don't use this", instead of just simply removing. (!!)
[This is the way it's currently done -- there is a removed flag inside the
RCDepProvide structure that marks this provide as something that was provided
earlier, but isn't any more -- if it's true, this provide shouldn't be
used for providing anything. (so all the reqs up above it are unprovided).

*************************/

static void
deps_provide_require_iterator (gpointer key, gpointer value, gpointer data)
{
    gchar *name = (gchar *) key;
    RCDepRequireSList *req_list = (RCDepRequireSList *) value;
    RCDepResolution *rcdr = (RCDepResolution *) data;
#if 0
    GSList *reqs_to_remove = NULL;
    GSList *iter;
    gboolean ret = FALSE;
#endif
    int res;

    DEPS_ENTER;
    DEPS_DEBUG_STRING("name(key)", name);

    deps_bt (name);

    while (req_list) {
        RCDepRequire *req = (RCDepRequire *) req_list->data;
        res = deps_provide_require (rcdr, req, name, FALSE);
        if (res == -1) {
            RCDepAndPackageSList *dap_iter;

            /* IS THIS RIGHT?! FIXME */
            if (rc_deps_debug_level > 20) {
                rc_debug (RC_DEBUG_LEVEL_DEBUG, "---- setting require key %s to ALL REMOVED\n", name);
            }


            check_daps_for_removal (rcdr, req);
            req->top_provide = REQUIRE_ALL_REMOVED;

#if 0
            /* Double-check all the daps in this req and see if anything is forced */
            if (!rc_package_dep_slist_is_fully_weak (req->deps)) {
                dap_iter = req->required_list;
                while (dap_iter) {
                    RCDepAndPackage *dap = (RCDepAndPackage *) dap_iter->data;
                    if (dap->force) {
                        /* make a conflict, no? */
                        if (g_slist_find (rcdr->packages_added, dap->package)) {
                            deps_add_conflict (rcdr,
                                               RC_CONFLICT_WHO_KNOWS,
                                               debug_rc_package_dep_str (dap->dep),
                                               dap->package,
                                               NULL,
                                               NULL);
                        } else {
                            /* We assume it will be automagically taken care of */
                        }
                    }
                    dap_iter = dap_iter->next;
                }
            }
#endif

            rcdr->reqs_to_remove = g_slist_append (rcdr->reqs_to_remove, req);
        }
        req_list = req_list->next;
    }

#if 0
#if 1
    iter = reqs_to_remove;
    while (iter) {
        RCDepRequire *r_req = (RCDepRequire *) iter->data;
        if (deps_remove_require (rcdr, r_req, FALSE)) {
            g_free (name);
            deps_require_free (r_req);
            ret = TRUE;
        }
        iter = iter->next;
    }
    g_slist_free (reqs_to_remove);
    return ret;
#else
    if (reqs_to_remove) {
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "--> %d to remove\n", g_slist_length (reqs_to_remove));
        while (reqs_to_remove) {
            RCDepRequire *r_req = (RCDepRequire *) reqs_to_remove->data;
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "----> %s\n", deps_join_names (r_req->names));
            reqs_to_remove = reqs_to_remove->next;
        }
    }

    return ret;
#endif
#endif
    DEPS_EXIT;
}

static int
deps_provide_require (RCDepResolution *rcdr, RCDepRequire *req, gchar *name,
                      gboolean invalidate_provide)
{
    RCDepProvide *prov;
    GSList *provs;
    int prov_res;

    DEPS_ENTER;
    DEPS_DEBUG_REQ ("req", req);
    DEPS_DEBUG_STRING ("name", name);
    DEPS_DEBUG_BOOL ("invalidate_provide", invalidate_provide);

    if (rc_package_dep_slist_is_fully_weak (req->deps)) {
        /* ignorable */
        DEPS_DEBUG_INT("ret", 1);
        DEPS_EXIT;
        return 1;
    }

    if (invalidate_provide || req->top_provide == REQUIRE_ALL_REMOVED) {
#if 1
        req->top_provide = NULL;
#else
        if (req->top_provide &&
            req->top_provide != REQUIRE_ALL_REMOVED &&
            req->top_provide != REQUIRE_HAS_CONFLICT)
        {
            deps_require_unprovide (rcdr, req, FALSE);
        } else {
            req->top_provide = NULL;
        }
#endif
    }

/*
    if (req->top_provide) {
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "-- %s is already provided by %s\n", name, req->top_provide->name);
        return;
    }
*/

#if 0
    /* Check and see if we're trying to remove this thing already */
    if (rcdr->packages_removed) {
        if (g_slist_find_custom (rcdr->packages_removed, name, package_find_by_name)) {
            /* already removed, doh */
            fprintf (stderr, "already found %s to remove, so skipping!\n", name);
            return 0;
        }
    }
#endif

    prov_res = deps_find_best_provide (rcdr, req, &prov);
    if (prov_res == 0) {
#if 1
        req->top_provide = NULL;
#else
        deps_require_unprovide (rcdr, req, FALSE);
#endif
        DEPS_DEBUG_INT("ret", 0);
        DEPS_EXIT;
        return 0;
    } else if (prov_res == -1) {
        DEPS_DEBUG_INT("ret", -1);
        DEPS_EXIT;
        return -1;
    }

    /* Make sure that this prov is unique */
    provs = g_hash_table_lookup (rcdr->provide_hash, prov->name);
    while (provs) {
        RCDepProvide *p = (RCDepProvide *) provs->data;
        if (p == prov) {
            provs = provs->next;
            continue;
        }

#if 1        
        deps_replace_provide (rcdr, p, prov);
#else
        if (p->required_by) {
            RCDepRequireSList *reqs = g_slist_copy (p->required_by);
            RCDepRequireSList *origlist = reqs;
            while (reqs) {

                deps_require_unprovide (rcdr, (RCDepRequire *) reqs->data, FALSE);
                reqs = reqs->next;
            }
            g_slist_free (origlist);
        }
#endif

        provs = provs->next;
    }

//    rcdr->unprovided_packages--;
    if (req->top_provide) {
        /* Already provided by something */
        deps_require_unprovide (rcdr, req, FALSE);
    }

    if (deps_provide_require_check (rcdr, req, prov)) {
        /* Now we update prov */
        if (!g_slist_find (prov->required_by, req))
            prov->required_by = g_slist_append (prov->required_by, req);
        DEPS_DEBUG_INT("ret", 0);
        return 0;
    }

    DEPS_DEBUG_INT("ret", 1);
    DEPS_EXIT;
    return 1;
}

static gboolean
deps_verify_and_dep_slist (RCPackageDepSList *dep_list, RCPackageDepItem *item, GSList **failures)
{
    DEPS_ENTER;
    DEPS_DEBUG_PDI("item", item);
    
    while (dep_list) {
        RCPackageDep *pd = (RCPackageDep *) dep_list->data;
        RCPackageDepItem *first_item = (RCPackageDepItem *) pd->data;
        RCPackageDep *failure = NULL;

        if (strcmp (first_item->spec.name, item->spec.name) != 0) {
            dep_list = dep_list->next;
            continue;
        }

        /* If we get this far, then this dep has the right name. We
         * verify it as an and.
         */
        if (item->relation == RC_RELATION_EQUAL || item->relation == RC_RELATION_ANY) {
            if (rc_package_dep_verify_and_relation (pd, &item->spec, &failure, (item->relation == RC_RELATION_ANY)))
            {
                /* It's good.. */
                DEPS_DEBUG_BOOL("ret", TRUE);
                DEPS_EXIT;
                return TRUE;
            } else {
                if (failures) {
                    *failures = g_slist_concat (*failures, rc_package_dep_copy (failure));
                }
                failure = NULL;
            }
        } else {
            g_error ("Provide %s has a relation other than EQUAL or ANY", rc_package_spec_to_str (&item->spec));
        }

        dep_list = dep_list->next;
    }

    DEPS_DEBUG_BOOL("ret", FALSE);
    DEPS_EXIT;
    return FALSE;
}

/*
 * This function is the core of the dependency resolution code.
 *
 * It is given a rcdr with a deps that has to be resolved. It
 * needs to find anything that can provide the thing named by
 * the deps, and then find the first one that resolves it.
 *
 * There are number of issues that must be considered here. First,
 * in the case of OR dependencies, some counting needs to take place
 * to get the most optimal provide for the package (i.e. if package foo
 * needs A | B, and package bar needs B | C, you'd want to resolve this
 * with B, not with A && C, unless you already had A && C installed.
 * It's tricky. ** Note that it's sort of okay to to request extra
 * packages initially. Just as long as everything gets resolved.
 *
 * The other issue is handling the foo & foo-devel case, namely,
 * the reverse provides case. This has to be done recursively,
 * so if foo is installed, the foo provide is looked at, and each
 * of its requires is checked to make sure this foo can still
 * satisfy the deps. If this foo cannot satisfy the deps, then new
 * versions of the packages with failed deps need to be found.
 * [Note that we aren't going to support the notion of a downgrade,
 * although this might fall out of the algorithm naturally].
 *
 */


static int
deps_find_provide (RCDepResolution *rcdr, RCDepRequire *req, RCDepProvide **prov_out)
{
    GSList *names_it;
    gboolean use_forced = TRUE;
    gboolean use_provided = TRUE;
    gboolean avoid_avoided = TRUE;
    gboolean no_provides = TRUE;

    RCDepProvide *found_provide = NULL;
    RCDepProvideSList *prov_list;
    GSList *fails = NULL;

    /* TRUE if each of the things that could provide this was removed */
    gboolean all_removed = TRUE;

    DEPS_ENTER;
    DEPS_DEBUG_REQ("req", req);

    if (rc_deps_debug_level > 10) {
        gchar *names = deps_join_names (req->names);
        deps_bt (names);
    }

    /* Check to see if everything was removed */
    names_it = req->names;
    while (names_it)
    {
        gchar *n = (gchar *) names_it->data;
        prov_list = g_hash_table_lookup (rcdr->provide_hash, n);
        if (!prov_list) {
            names_it = names_it->next;
            continue;
        }
        while (prov_list) {
            RCDepProvide *prov = (RCDepProvide *) prov_list->data;

            if (!prov->removed) {
                all_removed = FALSE;
                break;
            }
            prov_list = prov_list->next;
        }
        if (!all_removed)
            break;
        names_it = names_it->next;
    }

    /* First we see if we can resolve this with a real, installed package (not a virtual one) */
    names_it = req->names;
    while (names_it) 
    {
        gchar *n = (gchar *) names_it->data;

        prov_list = g_hash_table_lookup (rcdr->provide_hash, n);
        while (prov_list) {
            RCDepProvide *prov = (RCDepProvide *) prov_list->data;

            if (prov->removed) {
                prov_list = prov_list->next;
                continue;
            }

            if (avoid_avoided && prov->avoid) {
                prov_list = prov_list->next;
                continue;
            }

            if (prov->virtual == VIRTUAL_UNSET) {
                /* This is a real thing! Find me! Find me! */
                if (deps_verify_and_dep_slist (req->deps, prov->provision, &fails) &&
                    deps_provide_check_must_provide (rcdr, req, prov))
                {
                    /* We win! */
                    found_provide = prov;
                    break;
                }
            }

            prov_list = prov_list->next;
        }
        if (found_provide)
            break;

        names_it = names_it->next;
    }

    /* We try to find something that we must use */
    if (!found_provide) {
        GSList *maybe_list = NULL;

        names_it = req->names;

        while (names_it) {
            gchar *n = (gchar *) names_it->data;
            GSList *lnk;
            
            if ((lnk = g_slist_find_custom (rcdr->must_provide_list, n, dap_find_by_name)) != NULL) {
                RCDepAndPackage *ndap = (RCDepAndPackage *) lnk->data;
                prov_list = g_hash_table_lookup (rcdr->provide_hash, n);
                while (prov_list) {
                    RCDepProvide *prov = (RCDepProvide *) prov_list->data;
                    if (prov->provided_by == ndap->package) {
                        if (deps_verify_and_dep_slist (req->deps, prov->provision, &fails))
                        {
                            /* yay */
                            found_provide = prov;
                            break;
                        } else {
                            /* We don't win. We make a conflict, and pretend we won. */
                            /* Hrrrm. If this package can't provide it, then we remove it,
                             * unless we're trying to install it, in which case we need to remove
                             * all the -other- things.
                             */
                            maybe_list = g_slist_append (maybe_list, prov);
                        }
                    }

                    prov_list = prov_list->next;
                }
            }
            names_it = names_it->next;
        }

        if (maybe_list) {
            GSList *itz = maybe_list;
            while (itz) {
                RCDepProvide *prov = (RCDepProvide *) itz->data;
                if (!found_provide || (prov->provided_by != found_provide->provided_by)) {
                    if (g_slist_find (rcdr->packages_added, prov->provided_by)) {
                        *prov_out = NULL;
                        
                        DEPS_DEBUG_STRING ("ret -1", "prov_out == NULL");
                        DEPS_EXIT;
                        return -1;
                    } else {
                        /* arrgh. as soon as we do this once, we give up all our pointers. */
                        deps_remove_package (rcdr, prov->provided_by, FALSE);
//                            deps_update_remove_list (rcdr, prov->provided_by);
                        found_provide = NULL;
                        break;
                    }
                }
                itz = itz->next;
            }
        }
    }

    /* Try things that are forced */
    if (!found_provide) {
        GSList *maybe_list = NULL;

        names_it = req->names;
        while (names_it) {
            gchar *n = (gchar *) names_it->data;

            prov_list = g_hash_table_lookup (rcdr->provide_hash, n);
            if (prov_list) no_provides = FALSE;

            while (prov_list) {
                RCDepProvide *prov = (RCDepProvide *) prov_list->data;

                if (prov->removed) {
                    prov_list = prov_list->next;
                    continue;
                }

                if (avoid_avoided && prov->avoid) {
                    prov_list = prov_list->next;
                    continue;
                }

                /* If we're looking for forced things, then see if it's forced */
                if (prov->force)
                {
                    if (deps_verify_and_dep_slist (req->deps, prov->provision, &fails)) {
                        /* We win! */
                        found_provide = prov;
                        break;
                    } else {
                        /* We don't win. We make a conflict, and pretend we won. */
                        /* we maybe list it */
                        maybe_list = g_slist_append (maybe_list, prov);
                    
                        found_provide = prov;
                        break;
                    }
                }

                prov_list = prov_list->next;
            }

            if (found_provide)
                break;
            names_it = names_it->next;
        }

        if (maybe_list) {
            GSList *itz = maybe_list;
            while (itz) {
                RCDepProvide *prov = (RCDepProvide *) itz->data;
                if (!found_provide || (prov->provided_by != found_provide->provided_by)) {
                    if (g_slist_find (rcdr->packages_added, prov->provided_by)) {
                        *prov_out = NULL;
                        
                        DEPS_DEBUG_STRING ("ret -1", "prov_out == NULL");
                        DEPS_EXIT;
                        return -1;
                    } else {
                        /* arrgh. as soon as we do this once, we give up all our pointers. */
                        deps_remove_package (rcdr, prov->provided_by, FALSE);
//                            deps_update_remove_list (rcdr, prov->provided_by);
                        found_provide = NULL;
                        break;
                    }
                }
                itz = itz->next;
            }
        }
    }

    /* Now look for already provided things */
    if (!found_provide) {
        names_it = req->names;
        while (names_it) {
            gchar *n = (gchar *) names_it->data;

            prov_list = g_hash_table_lookup (rcdr->provide_hash, n);
            if (prov_list) no_provides = FALSE;

            while (prov_list) {
                RCDepProvide *prov = (RCDepProvide *) prov_list->data;

                if (prov->removed) {
                    prov_list = prov_list->next;
                    continue;
                }

                if (avoid_avoided && prov->avoid) {
                    prov_list = prov_list->next;
                    continue;
                }

                /* If we're not looking for forced things, but looking for provided things,
                   see if it's in use by something. */
                if (prov->required_by != NULL)
                {
                    if (deps_verify_and_dep_slist (req->deps, prov->provision, &fails)) {
                        /* We win! */
                        found_provide = prov;
                        break;
                    }
                }

                prov_list = prov_list->next;
            }

            if (found_provide)
                break;
            names_it = names_it->next;
        }
    }

    /* Just find something (as long as it's not avoided) */
    if (!found_provide) {
        names_it = req->names;
        while (names_it) {
            gchar *n = (gchar *) names_it->data;

            prov_list = g_hash_table_lookup (rcdr->provide_hash, n);
            if (prov_list) no_provides = FALSE;

            while (prov_list) {
                RCDepProvide *prov = (RCDepProvide *) prov_list->data;

                if (prov->removed) {
                    prov_list = prov_list->next;
                    continue;
                }

                if (avoid_avoided && prov->avoid) {
                    prov_list = prov_list->next;
                    continue;
                }

                /* If we're not looking for either, grab something that works. */
                if (deps_verify_and_dep_slist (req->deps, prov->provision, &fails)) {
                    /* We win! */
                    found_provide = prov;
                    break;
                }

                prov_list = prov_list->next;
            }

            if (found_provide)
                break;

            names_it = names_it->next;
        }
    }

    /* Just find something, for crying out loud! anything! */
    if (!found_provide) {
        names_it = req->names;
        while (names_it) {
            gchar *n = (gchar *) names_it->data;

            prov_list = g_hash_table_lookup (rcdr->provide_hash, n);
            if (prov_list) no_provides = FALSE;

            while (prov_list) {
                RCDepProvide *prov = (RCDepProvide *) prov_list->data;

                if (prov->removed) {
                    prov_list = prov_list->next;
                    continue;
                }

                /* If we're not looking for either, grab something that works. */
                if (deps_verify_and_dep_slist (req->deps, prov->provision, &fails)) {
                    /* We win! */
                    found_provide = prov;
                    break;
                }

                prov_list = prov_list->next;
            }

            if (found_provide)
                break;

            names_it = names_it->next;
        }
    }

    if (all_removed && !no_provides) {
        *prov_out = NULL;

        /* no, no, no! if everything was removed, this means we need to try to find
         * something that will satisfy this, if this package was something whose
         * installation was requested.
         */
#if 1
        /* Everything that could provide this has been removed, so we need
         * to remove this package.
         */
        DEPS_DEBUG_STRING ("ret -1", "prov_out == NULL");
        DEPS_EXIT;
        return -1;
#else
        DEPS_DEBUG_STRING ("ret 0", "prov_out == NULL");
        DEPS_EXIT;
        return 0;
#endif
    }

    if (fails) {
        GSList *iter;
        RCPackageSList *newer_pkgs = NULL;
        RCPackage *newer;
        RCPackage *forced_newer;

        /* Try to resolve it in one fell swoop */
        forced_newer = deps_find_best_package_for_dep (fails, rcdr->current_chs, rcdr->user_pref);
        if (forced_newer) {
            newer_pkgs = g_slist_append (newer_pkgs, forced_newer);
        }


        /* First check if the failures have new versions available in an attempt to
         * fix this. If we can't find any, we make a conflict.
         */
        if (!forced_newer) {
            iter = fails;
            while (iter) {
                RCPackageDepItem *pdi = (RCPackageDepItem *) iter->data;
                RCDepConflict *conf = NULL;
                RCDepAndPackageSList *daps;

                if (pdi->relation & RC_RELATION_NONE ||
                    (pdi->relation & RC_RELATION_NOT_EQUAL && pdi->spec.version == NULL))
                {
                    /* So, this failed because we don't -want- this thing. */
                    /* Let's try uninstalling */
                    /* Mantra: if all else fails, uninstall. */
                    /* hmm. just in case. if (!strcmp (pdi->spec.name, "glibc")) {.. NO! BAD! BAD! ;-) */
#if 0
                    *prov_out = NULL;
                    return -1;
#else
                    RCDepAndPackageSList *daps = find_daps_for_pdi (rcdr, req, pdi);
                    GSList *removes = NULL;
                    GSList *local_iter;
                    gboolean sense = TRUE;
                    gboolean toggled = FALSE;

                    /* First check to see if all the daps are forced. If they are,
                     * then we need to honor them over the others.
                     */
                    local_iter = daps;
                    while (local_iter) {
                        RCDepAndPackage *dap = (RCDepAndPackage *) local_iter->data;
                        if (!dap->force) {
                            sense = FALSE;
                            toggled = TRUE;
                            break;
                        }
                        local_iter = local_iter->next;
                    }

                    if (!toggled) sense = FALSE;

                    if (sense) {
                        local_iter = daps;
                    } else {
                        local_iter = req->required_list;
                    }
                    while (local_iter) {
                        RCDepAndPackage *dap = (RCDepAndPackage *) local_iter->data;
                        if ((dap->force == sense) && !g_slist_find (removes, dap->package)) {
                            removes = g_slist_append (removes, dap->package);
                            if (sense) {
                                /* Don't get too overzealous with removing; just uninstall the first thing. */
                                /* (If we're uninstalling non-forced things) */
                                break;
                            }
                        }
                        local_iter = local_iter->next;
                    }
                    g_slist_free (daps);
                    local_iter = removes;
                    while (local_iter) {
                        RCPackage *pkg = (RCPackage *) local_iter->data;
                        deps_remove_package (rcdr, pkg, FALSE);
                        local_iter = local_iter->next;
                    }
                    g_slist_free (removes);

                    *prov_out = NULL;

                    DEPS_DEBUG_STRING ("ret 0", "prov_out == NULL");
                    DEPS_EXIT;

                    return 0;
#endif
                }

                newer = deps_find_newer_package_by_name (rcdr, pdi->spec.name);
                
                if (newer) {
                    if (rc_deps_debug_level > 10) {
                        rc_debug (RC_DEBUG_LEVEL_DEBUG, ">>> newer package %s for pdi %s\n",
                                  rc_package_spec_to_str (&newer->spec),
                                  rc_package_spec_to_str (&pdi->spec));
                    }
                
                    if (!g_slist_find (newer_pkgs, newer))
                        newer_pkgs = g_slist_append (newer_pkgs, newer);
                } else {
                    daps = find_daps_for_pdi (rcdr, req, pdi);
                
                    if (rc_deps_debug_level > 10) {
                        rc_debug (RC_DEBUG_LEVEL_DEBUG, ">>> failure: %s, n_daps: %d\n", debug_rc_package_dep_item_str (pdi),
                                  g_slist_length (daps));
                    }

                    while (daps) {
                        RCDepAndPackage *dap = (RCDepAndPackage *) daps->data;
                    
                        if (rc_deps_debug_level > 10) {
                            rc_debug (RC_DEBUG_LEVEL_DEBUG, "   dap: pkg: %s what: ", rc_package_spec_to_str (&dap->package->spec));
                            debug_rc_package_dep_dump (RC_DEBUG_LEVEL_DEBUG, dap->dep);
                        }
                    
#if 1
                        deps_upgrade_or_remove_package (rcdr, dap->package, dap->force);
//                        deps_remove_package (rcdr, dap->package, TRUE);
#else
                        /* Make a conflict */
                        deps_add_conflict (rcdr,
                                           RC_CONFLICT_WHO_KNOWS,
                                           deps_join_names (req->names),
                                           dap->package,
                                           rc_package_dep_item_copy (pdi),
                                           NULL);
#endif

                        daps = daps->next;
                    }
                }
                iter = iter->next;
            }
        }

#if 0
        iter = fails;
        while (iter) {
            RCPackageDep *pd = (RCPackageDepItem *) iter->data;
            if (pdi->relation == RC_RELATION_WEAK_NOT_EQUAL &&
                pdi->spec.version == NULL)
            {
                /* This smells like a conflict. This is a bit hacky,
                 * I think, but we'll give it a shot. FIXME
                 */
                g_slist_free (fails);
                *prov_out = NULL;
                DEPS_DEBUG_STRING ("ret -1", "prov_out == NULL");
                DEPS_EXIT;
                return -1;
            }
            iter = iter->next;
        }
#endif

        iter = newer_pkgs;
        while (iter) {
            RCPackage *newer_pkg = (RCPackage *) newer_pkgs->data;
            gboolean same_as_old = FALSE;
            RCPackageSList *old_pkgs;

            /* Check and see if the newer package is indeed newer than what is
             * already installed.
             */
            if (rcdr->system_packages) {
                old_pkgs = g_slist_find_custom (rcdr->system_packages, newer_pkg->spec.name,
                                                package_find_by_name);
                if (old_pkgs) {
                    RCPackage *older_pkg = (RCPackage *) old_pkgs->data;
                    
                    if (rc_package_spec_equal (&older_pkg->spec, &newer_pkg->spec)) {
                        /* it's the same package */
                        same_as_old = TRUE;
                    }
                }
            }
            if (rcdr->packages_added) {
                old_pkgs = g_slist_find_custom (rcdr->packages_added, newer_pkg->spec.name,
                                                package_find_by_name);
                if (old_pkgs) {
                    RCPackage *older_pkg = (RCPackage *) old_pkgs->data;
                    
                    if (rc_package_spec_equal (&older_pkg->spec, &newer_pkg->spec)) {
                        /* it's the same package */
                        same_as_old = TRUE;
                    }
                }
            }

            if (!same_as_old) {
                deps_add_package (rcdr, newer_pkg, TRUE);
            }
            iter = iter->next;
        }

        g_slist_free (fails);
    }

    if (!found_provide) {
        *prov_out = NULL;
        DEPS_DEBUG_STRING ("ret 0", "prov_out == NULL");
        DEPS_EXIT;
        return 0;
    }

#if BADCODE
    if (use_forced) {
        /* find the old provide, if any */
        prov_list = g_hash_table_lookup (rcdr->provide_hash, found_provide->name);
        while (prov_list) {
            RCDepProvide *prov = (RCDepProvide *) prov_list->data;
            if (prov != found_provide && prov->required_by) {
                old_provide = prov;
                break;
            }
            prov_list = prov_list->next;
        }
    }
        
    if (old_provide)
        deps_replace_provide (rcdr, old_provide, found_provide);
#endif

    *prov_out = found_provide;
    DEPS_DEBUG_PROV("prov_out", *prov_out);
    DEPS_EXIT;

    return 1;
}

static gboolean
deps_provide_check_must_provide (RCDepResolution *rcdr, RCDepRequire *req, RCDepProvide *maybe_provide)
{
    GSList *lnk;

    if ((lnk = g_slist_find_custom (rcdr->must_provide_list, maybe_provide->name, dap_find_by_name)) != NULL) {
        RCDepAndPackage *ndap = (RCDepAndPackage *) lnk->data;
        if (ndap->package != maybe_provide->provided_by)
            return FALSE;
    }

    return TRUE;
}

static gboolean
deps_provide_require_check (RCDepResolution *rcdr, RCDepRequire *req, RCDepProvide *prov)
{
    RCPackageDepSList *rds;
    RCPackageDepItemSList *failures = NULL;
    gboolean success = FALSE;

    DEPS_ENTER;
    DEPS_DEBUG_REQ ("req", req);
    DEPS_DEBUG_PROV ("prov", prov);

    if (rc_deps_debug_level > 10) {
        gchar *names = deps_join_names (req->names);
        deps_bt (names);
    }

    /* So something failed here -- we try to fix it */
    rds = req->deps;
    while (rds) {
        RCPackageDep *dep = (RCPackageDep *) rds->data;
        RCPackageDep *fail_out = NULL;
        
        /* See if anything fails in this one */
        if (prov->provision->relation == RC_RELATION_EQUAL ||
            prov->provision->relation == RC_RELATION_ANY)
        {
            if (!rc_package_dep_verify_and_relation (dep, &prov->provision->spec, &fail_out,
                                                     (prov->provision->relation == RC_RELATION_ANY)))
            {
                /* Check all the things that the provided package provides */
                RCPackageDepSList *other_provides = prov->provided_by->provides;
                gboolean all_failed = TRUE;
                while (other_provides) {
                    RCPackageDep *other_prov = (RCPackageDep *) other_provides->data;
                    while (other_prov) {
                        RCPackageDepItem *pdi = (RCPackageDepItem *) other_prov->data;
                        if (rc_package_dep_verify_and_relation (dep, &pdi->spec, NULL,
                                                                (pdi->relation == RC_RELATION_ANY)))
                        {
                            all_failed = FALSE;
                            break;
                        }
                        other_prov = other_prov->next;
                    }
                    if (!all_failed)
                        break;
                    other_provides = other_provides->next;
                }

                if (all_failed) {
                    /* What failed? */
                    if (rc_deps_debug_level > 10) {
                        rc_debug (RC_DEBUG_LEVEL_DEBUG, "@@@@ FAIL: dep item failed: %s (AVOIDING)\n",
                                 debug_rc_package_dep_str (fail_out));
                        print_require (RC_DEBUG_LEVEL_DEBUG, req);
                        print_provide (RC_DEBUG_LEVEL_DEBUG, prov);
                    }

                    failures = g_slist_concat (failures, rc_package_dep_copy (fail_out));
                    prov->avoid = TRUE;
                    success = FALSE;
                } else {
                    /* it really passes */
                    success = TRUE;
                    break;
                }
            } else {
                /* This one passes -- so one thing in the deps OR list is valid */
                success = TRUE;
                break;
            }
        } else {
            g_error ("Provide %s has a relation other than EQUAL or ANY", prov->name);
        }
        rds = rds->next;
    }

    if (!success) {
        int unfixed;
        /* remove the bits that failed */
        unfixed = deps_require_remove_deps (rcdr, req, failures, FALSE);
        /* now remerge the original require */
        deps_remerge_reqs (rcdr, req);

#if 1
        req->top_provide = NULL;
#else
        deps_require_unprovide (rcdr, req, FALSE);
#endif

        if (unfixed == g_slist_length (failures)) {
            /* Eeek.  We have issues here we can't deal with. */
            /* We're getting failures, but all of them are forced. This means that
             * we need to start removing the other things. HANDLE THIS RIGHT -- FIXME
             */
            return FALSE;
        }
        /* Only run it through again if all the failing things were removed.  If they
         * weren't, we don't run it again and leave it as unprovided.
         */
        if (!unfixed) {
            /* Make sure it still succeeds */
            deps_provide_require_check (rcdr, req, prov);
        }
        
        g_slist_free (failures);

        /* will be set by whichever check succeeds */
        DEPS_EXIT;
        return unfixed ? FALSE : TRUE;
    }

    req->top_provide = prov;
    DEPS_EXIT;
    return TRUE;
}    

static int
deps_require_remove_deps (RCDepResolution *rcdr, RCDepRequire *req,
                          RCPackageDepItemSList *depis,
                          gboolean force)
{
    int unfixed = 0;

    DEPS_ENTER;
    DEPS_DEBUG_REQ("req", req);
    DEPS_DEBUG_PD("depis", depis);

    while (depis) {
        RCPackageDepItem *depi = (RCPackageDepItem *) depis->data;
        RCDepAndPackageSList *daps = req->required_list;
        gboolean this_one_counted = FALSE;

        while (daps) {
            RCDepAndPackage *dap = (RCDepAndPackage *) daps->data;
            GSList *nextp = daps->next;

            if (dap->force && !force) {
                daps = nextp;
                if (!this_one_counted)
                    unfixed++;
                this_one_counted = TRUE;
                continue;
            } else if (dap->force) {
                if (rc_package_dep_is_item_subset (dap->dep, depi) == 0) {
                    rc_debug (RC_DEBUG_LEVEL_DEBUG, "dap->dep, depi... foo_forced!\n");
                }
            }

            switch (rc_package_dep_is_item_subset (dap->dep, depi)) {
                case 1:
                case -1:
                    /* Ignore. */
                    break;

                    /* if it's -1, we don't know if it's a subset, but we do know
                     * that it failed. (Do we?)
                     */

                case 0:

                    /* Remove this DaP and append to failures */
                    req->required_list = g_slist_remove_link (req->required_list, daps);
                    rcdr->failed_daps = g_slist_append (rcdr->failed_daps, dap);
                    g_slist_free_1 (daps);
                    daps = nextp;
                    continue;
            }

            daps = nextp;
        }
        depis = depis->next;
    }

    DEPS_EXIT;
    return unfixed;
}

static void
deps_replace_provide (RCDepResolution *rcdr, RCDepProvide *oldp, RCDepProvide *newp)
{
    RCDepRequireSList *old_reqs;

    DEPS_ENTER;
    DEPS_DEBUG_PROV("oldp", oldp);
    DEPS_DEBUG_PROV("newp", newp);

    old_reqs = oldp->required_by;
    while (old_reqs) {
        RCDepRequire *req = (RCDepRequire *) old_reqs->data;

        /* First check and see if something will fail */
        if (deps_verify_and_dep_slist (req->deps, newp->provision, NULL)) {
            /* It's ok */
            oldp->required_by = g_slist_remove (oldp->required_by, req);
            newp->required_by = g_slist_append (newp->required_by, req);
            req->top_provide = newp;
            old_reqs = oldp->required_by;
            continue;
        }

        deps_provide_require_check (rcdr, req, newp);

        old_reqs = old_reqs->next;
    }
    DEPS_EXIT;
}

/* Fix the failed fail_out dependency in req (caused by newp)
 * by finding newer versions of everything that depends on fail_out
 */

static RCPackage *
deps_find_newer_package_for_dap (RCDepResolution *rcdr,
                                 RCDepAndPackage *dap)
{
    RCPackage *pkg = dap->package;
    RCPackage *new_pkg;
    GSList *new_pkgs = NULL;
    GSList *iter;
    RCPackageDep *newer_dep;

    DEPS_ENTER;
    DEPS_DEBUG_DAP("dap", dap);

    newer_dep = rc_package_dep_new_with_item
        (rc_package_dep_item_new_from_spec (&dap->package->spec,
                                            RC_RELATION_GREATER));

    new_pkg = deps_find_newer_package_for_dep (rcdr, newer_dep);

    return new_pkg;
}

static RCPackage *
deps_find_newer_package (RCDepResolution *rcdr,
                         RCPackage *pkg)
{
    RCPackage *new_pkg;
    GSList *new_pkgs = NULL;
    GSList *iter;
    RCPackageDep *newer_dep;

    DEPS_ENTER;

    newer_dep = rc_package_dep_new_with_item
        (rc_package_dep_item_new_from_spec (&pkg->spec,
                                            RC_RELATION_GREATER));

    new_pkg = deps_find_newer_package_for_dep (rcdr, newer_dep);

    return new_pkg;
}

static RCPackage *
deps_find_newer_package_for_dep (RCDepResolution *rcdr,
                                 RCPackageDep *newer_dep)
{
    RCPackage *new_pkg;
    GSList *new_pkgs = NULL;
    GSList *iter;

    DEPS_ENTER;

    new_pkg = deps_find_best_package_for_dep (newer_dep, rcdr->preferred_chs, rcdr->user_pref);
    if (new_pkg)
        new_pkgs = g_slist_append (new_pkgs, new_pkg);
    new_pkg = deps_find_best_package_for_dep (newer_dep, rcdr->current_chs, rcdr->user_pref);
    if (new_pkg)
        new_pkgs = g_slist_append (new_pkgs, new_pkg);
    new_pkg = deps_find_best_package_for_dep (newer_dep, rcdr->other_chs, rcdr->user_pref);
    if (new_pkg)
        new_pkgs = g_slist_append (new_pkgs, new_pkg);

    iter = new_pkgs;
    while (iter) {
        new_pkg = (RCPackage *) iter->data;
        if (!rc_package_dep_verify_relation (newer_dep, &new_pkg->spec)) {
            iter = iter->next;
            continue;
        }

        if (rc_deps_debug_level > 20) {
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "-- found package for fixing DAP %s\n", rc_package_spec_to_str (&new_pkg->spec));
        }

        /* Found this one */
        DEPS_DEBUG_PKG("ret", new_pkg);
        DEPS_EXIT;
        return new_pkg;
    }

    if (rc_deps_debug_level > 20)
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "-- didn't find any package\n");

    DEPS_EXIT;
    return NULL;
}

#if NOTUSED
static RCPackage *
deps_find_newer_package_for_pdi (RCDepResolution *rcdr,
                                 RCPackageDepItem *pdi)
{
    RCPackage *new_pkg;
    RCPackageDep *newer_dep;

    DEPS_ENTER;
    DEPS_DEBUG_PDI("pdi", pdi);

    newer_dep = rc_package_dep_new_with_item
        (rc_package_dep_item_new_from_spec (&pdi->spec, RC_RELATION_GREATER));

    new_pkg = deps_find_newer_package_for_dep (rcdr, newer_dep);

    return new_pkg;
}
#endif

static RCPackage *
deps_find_newer_package_by_name (RCDepResolution *rcdr,
                                 gchar *name)
{
    RCPackageDep *newer_dep;
    RCPackage *new_pkg;

    DEPS_ENTER;
    DEPS_DEBUG_STRING("name", name);

    newer_dep = rc_package_dep_new_with_item
        (rc_package_dep_item_new (name, 0, NULL, NULL, RC_RELATION_ANY));

    new_pkg = deps_find_newer_package_for_dep (rcdr, newer_dep);

    return new_pkg;
}

static void
deps_update_remove_list (RCDepResolution *rcdr, RCPackage *pkg)
{
    GSList *it;

    DEPS_ENTER;
    DEPS_DEBUG_PKG("pkg", pkg);

    /* Check if pkg is either providing anything, or if it was
     * added. If so, clear out its provides, requirements, and
     * the things that it was fulfilling, and get rid of it
     * from the added list (unless it was forced!!)
     */
    if (rc_deps_debug_level > 10)
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "REMOVE: Removing %s\n", rc_package_spec_to_str (&pkg->spec));
    /* Check the install list to make sure it's not on there as well */
    /* FIXME -- this probably should be done elsewhere, in case we really
     * want to remove something we previously added. But then the upstream
     * code could just remove it from that list first.
     */

    it = rcdr->packages_added;
    while (it) {
        RCPackage *instpkg = (RCPackage *) it->data;
        if (pkg != instpkg && !strcmp (pkg->spec.name, instpkg->spec.name)) {
            /* This package is in the to install list, so we ignore this remove
             * (it's an upgrade)
             */
            DEPS_EXIT;
            return;
        }
        it = it->next;
    }

    /* We need to pull out this package from the requires/provides lists */
    deps_untangle_package (rcdr, pkg);
    if (g_slist_find (rcdr->packages_added, pkg)) {
        GSList *lnk;
        rcdr->packages_added = g_slist_remove (rcdr->packages_added, pkg);
    } else {
        rcdr->packages_removed = g_slist_append (rcdr->packages_removed, pkg);
    }

    DEPS_EXIT;
}


/* Pull this package out and [hopefully] leave no trace */
/* this is actually possible now! */
static void
deps_untangle_package (RCDepResolution *rcdr, RCPackage *pkg)
{
    GSList *it;
    GSList *tmpreq;
    RCPackageDep *smack_me_up;


    DEPS_ENTER;
    DEPS_DEBUG_PKG("pkg", pkg);

    it = pkg->provides;
    while (it) {
        RCPackageDep *pd = (RCPackageDep *) it->data;
        while (pd) {
            RCPackageDepItem *pdi = (RCPackageDepItem *) pd->data;
            GSList *prov_it;

            prov_it = g_hash_table_lookup (rcdr->provide_hash, pdi->spec.name);
            while (prov_it) {
                RCDepProvide *prov = (RCDepProvide *) prov_it->data;

                if (prov->provided_by == pkg) {
                    RCDepRequireSList *reqs = prov->required_by;

                    prov->removed = TRUE;

                    while (reqs) {
                        RCDepRequire *req = (RCDepRequire *) reqs->data;
                        deps_require_unprovide (rcdr, req, FALSE);

                        deps_remerge_reqs (rcdr, req);

                        if (!g_slist_find (prov->required_by, req))
                            reqs = prov->required_by;
                        else
                            reqs = reqs->next;
                    }
                }
                prov_it = prov_it->next;
            }
            pd = pd->next;
        }
        it = it->next;
    }

    tmpreq = g_slist_copy (pkg->requires);
    /* Get rid of meself from the require list! */
    smack_me_up = rc_package_dep_new_with_item
        (rc_package_dep_item_new_from_spec (&pkg->spec, RC_RELATION_EQUAL));
    tmpreq = g_slist_append (tmpreq, smack_me_up);
    it = tmpreq;
    while (it) {
        RCPackageDep *pd = (RCPackageDep *) it->data;
        RCDepRequireSList *reqlist;

        gchar *nunu = deps_make_names (pd, NULL);
        reqlist = g_hash_table_lookup (rcdr->require_hash, nunu);
        g_free (nunu);
        while (reqlist) {
            RCDepRequire *req = (RCDepRequire *) reqlist->data;
            GSList *dapit = req->required_list;
            while (dapit) {
                RCDepAndPackage *od = (RCDepAndPackage *) dapit->data;
                if (od->package == pkg) {
                    req->required_list = g_slist_remove (req->required_list, od);
                    g_free (od);
                    dapit = req->required_list;
                } else {
                    dapit = dapit->next;
                }
            }
            deps_remerge_reqs (rcdr, req);
            reqlist = reqlist->next;
        }
        it = it->next;
    }
    g_slist_free (tmpreq);
    rc_package_dep_free (smack_me_up);

    it = pkg->conflicts;
    while (it) {
        RCPackageDep *pd = (RCPackageDep *) it->data;
        RCDepRequireSList *reqlist;

        gchar *nunu = deps_make_names (pd, NULL);
//        while (pd) {
//            RCPackageDepItem *pdi = (RCPackageDepItem *) pd->data;
//            gchar *nunu = pdi->spec.name;

        reqlist = g_hash_table_lookup (rcdr->require_hash, nunu);
        g_free (nunu);
        while (reqlist) {
            RCDepRequire *req = (RCDepRequire *) reqlist->data;
            GSList *dapit = req->required_list;
            while (dapit) {
                RCDepAndPackage *od = (RCDepAndPackage *) dapit->data;
                if (od->package == pkg) {
                    req->required_list = g_slist_remove (req->required_list, od);
                    g_free (od);
                    dapit = req->required_list;
                    continue;
                }
                dapit = dapit->next;
            }
            deps_remerge_reqs (rcdr, req);
            reqlist = reqlist->next;
//        }
//        pd = pd->next;
        }
        it = it->next;
    }

    it = rcdr->must_provide_list;
    while (it) {
        RCDepAndPackage *dap = (RCDepAndPackage *) it->data;
        if (dap->package == pkg) {
            GSList *nextp = it->next;
            g_free (it->data); /* dap */
            rcdr->must_provide_list = g_slist_remove_link (rcdr->must_provide_list, it);
            it = nextp;
            continue;
        }
        it = it->next;
    }

    DEPS_EXIT;
}

static void
deps_update_install_list (RCDepResolution *rcdr, RCPackage *pkg)
{
    /* Check if pkg is on the remove list; if so, get rid of it
     * from there.
     */

    rc_debug (RC_DEBUG_LEVEL_DEBUG, "INSTALL: Installing %s", rc_package_spec_to_str (&pkg->spec));
    if (!g_slist_find (rcdr->packages_added, pkg)) {
        rcdr->packages_added = g_slist_append (rcdr->packages_added, pkg);
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "\n");
    } else {
        rc_debug (RC_DEBUG_LEVEL_DEBUG, " (already there)\n");
    }
}

static void
deps_provide_unprovide (RCDepResolution *rcdr, RCDepProvide *prov)
{
    RCDepProvideSList *prov_list;

    DEPS_ENTER;

    prov_list = g_hash_table_lookup (rcdr->provide_hash, prov->name);
    while (prov_list) {
        RCDepProvide *oldp = (RCDepProvide *) prov_list->data;
        if (oldp == prov) {
            prov_list = prov_list->next;
            continue;
        }

        if (prov->required_by) {
            /* Nuke these */
            while (oldp->required_by) {
                RCDepRequire *req = (RCDepRequire *) oldp->required_by->data;
                deps_require_unprovide (rcdr, req, FALSE);
                /* req unprovide will modify prov->required_by */
            }
        }
        prov_list = prov_list->next;
    }

    DEPS_EXIT;
}


static int
deps_find_best_provide (RCDepResolution *rcdr, RCDepRequire *req, RCDepProvide **prov_out)
{
    int prov_result;
    RCDepProvide *prov;
    RCPackage *new_pkg;
    GSList *new_pkgs = NULL;
    GSList *iter;

    DEPS_ENTER;

    if (rc_deps_debug_level > 5) {
        gchar *s = deps_join_names (req->names);
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "+++++ STARTING find_best_provide for require: %s\n",
                 s);
        deps_bt (s);
    }

    /* First we check to make sure we use something that's already provided */
    prov_result = deps_find_provide (rcdr, req, &prov);
    switch (prov_result) {
        case 1:
            *prov_out = prov;
            return 1;
            break;
        case -1:
            /* This means this req should be removed */
            iter = req->required_list;
            while (iter) {
                RCDepAndPackage *dap = (RCDepAndPackage *) iter->data;
                /* We don't remove weak dependencies, since they're not real deps */
                if (!rc_package_dep_is_fully_weak (dap->dep)) {
                    if (!dap->force) {
                        deps_upgrade_or_remove_package (rcdr, dap->package, dap->force);
//                        deps_remove_package (rcdr, dap->package, FALSE);
                    }
                    /* we can only remove one at a time -- req isn't valid
                     * after this point, really.
                     */
                    break;
                }
                iter = iter->next;
            }
            *prov_out = NULL;
            return -1;
            break;
        case 0:
            break;
    }

    /* Not found? Okay, we need to try to find a package that satisfies this dep */
    new_pkg = deps_find_best_package_for_dep_slist (req->deps, rcdr->preferred_chs, rcdr->user_pref);
    if (new_pkg)
        new_pkgs = g_slist_append (new_pkgs, new_pkg);
    new_pkg = deps_find_best_package_for_dep_slist (req->deps, rcdr->current_chs, rcdr->user_pref);
    if (new_pkg)
        new_pkgs = g_slist_append (new_pkgs, new_pkg);
    new_pkg = deps_find_best_package_for_dep_slist (req->deps, rcdr->other_chs, rcdr->user_pref);
    if (new_pkg)
        new_pkgs = g_slist_append (new_pkgs, new_pkg);

    iter = new_pkgs;
    while (iter) {
        RCPackageDepItem *di;

        new_pkg = (RCPackage *) iter->data;
        if (!deps_verify_package_for_and_dep_slist (rcdr, new_pkg, req->deps)) {
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "+++++ found package %s FAILED!\n", rc_package_spec_to_str (&new_pkg->spec));
            iter = iter->next;
            new_pkg = NULL;
            continue;
        }

        if (rc_deps_debug_level > 5) {
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "+++++ ADDING package %s\n", rc_package_spec_to_str (&new_pkg->spec));
        }
        deps_add_package (rcdr, new_pkg, TRUE);
        deps_find_provide (rcdr, req, &prov);

        if (prov) {
            if (rc_deps_debug_level > 5) {
                rc_debug (RC_DEBUG_LEVEL_DEBUG, "+++++ PUSHING provide for %s\n", prov->name);
            }
            *prov_out = prov;
            DEPS_EXIT;
            return 1;
        } else {
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "Pushed new package %s for req [first name] %s (%d), but didn't find provide!\n",
                     rc_package_spec_to_str (&new_pkg->spec), (gchar *) req->names->data,
                     g_slist_length (req->names));
        }
        iter = iter->next;
    }

    rc_debug (RC_DEBUG_LEVEL_DEBUG, "$$$$ Not found for req %p\n", req);
    print_require (RC_DEBUG_LEVEL_DEBUG, req);
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "$$$$\n");
    *prov_out = NULL;

    DEPS_EXIT;
    return 0;
}

static RCPackage *
deps_find_best_package_for_dep_slist (RCPackageDepSList *deps, RCChannelSList *chs, gint user_pref)
{
    RCPackage *retpkg = NULL;

    if (chs == NULL)
        return NULL;

    while (deps) {
        RCPackageDep *dep = (RCPackageDep *) deps->data;
        RCPackage *pkg;

        pkg = deps_find_best_package_for_dep (dep, chs, user_pref);

        if (retpkg) {
            if (rc_packman_version_compare (das_global_packman, &pkg->spec, &retpkg->spec) > 0) {
                retpkg = pkg;
            }
        } else {
            retpkg = pkg;
        }

        if (pkg) return pkg;

        deps = deps->next;
    }

    return NULL;
}

static RCPackage *
deps_find_best_package_for_dep (RCPackageDep *dep, RCChannelSList *chs, gint user_pref)
{
    RCPackageSList *found_packages = NULL;
    RCPackage *pkg;
    gchar *first_name = ((RCPackageDepItem *) dep->data)->spec.name;

    while (chs) {
        RCChannel *ch = (RCChannel *) chs->data;

        pkg = pkginfo_find_package_with_constraint (ch,
                                                    first_name,
                                                    user_pref,
                                                    dep,
                                                    TRUE);
#if 0
        fprintf (stderr, "finding: %s  ", first_name);
#endif

        if (pkg) {
            found_packages = g_slist_append (found_packages, pkg);
        } else {
            /* Try to find it in the channel's dep_table */
            RCSubchannel *subch = NULL;
            {
                RCSubchannelSList *subchs = ch->subchannels;
                int best_pref = 999999;
                while (subchs) {
                    RCSubchannel *s = (RCSubchannel *) subchs->data;
                    if (s->preference >= user_pref &&
                        s->preference < best_pref)
                    {
                        best_pref = s->preference;
                        subch = s;
                    }
                    subchs = subchs->next;
                }
                if (best_pref == 0) {
                    rc_debug (RC_DEBUG_LEVEL_WARNING, "Couldn't find subchannel for preference %d, channel %d\n", user_pref, ch->id);
                }
            }

            if (subch) {
                RCPackageDep *dep_list = dep;
                while (dep) {
                    RCPackageDepItem *depi = (RCPackageDepItem *) dep_list->data;

                    pkg = g_hash_table_lookup (subch->dep_table,
                                               &depi->spec);
                    if (pkg && rc_package_dep_verify_and_relation (dep, &pkg->spec, NULL,
                                                                   0))
                    {
                        found_packages = g_slist_append (found_packages, pkg);
                    }
                    dep = dep->next;
                }
            }
        }

        chs = chs->next;
    }
    
    if (found_packages != NULL) {
        /* We found at least one package that will satisfy everything. */
        found_packages = rc_package_slist_sort_by_spec (found_packages);
        pkg = (RCPackage *)(g_slist_last (found_packages) -> data);
        g_slist_free (found_packages);
#if 0
        fprintf (stderr, "found: %s\n", rc_package_spec_to_str (&pkg->spec));
#endif
        return pkg;
    }
#if 0
    fprintf (stderr, "found: NULL\n");
#endif
    return NULL;
}

/*
 * Find the list of DAPs that contribute to this pdi
 *
 * This is used to figure out what packages are causing a
 * conflict.
 */
static RCDepAndPackageSList *
find_daps_for_pdi (RCDepResolution *rcdrr, RCDepRequire *req, RCPackageDepItem *pdi)
{
    RCDepAndPackageSList *dap_iter;
    RCDepAndPackageSList *out_list = NULL;

    DEPS_ENTER;
    DEPS_DEBUG_REQ("req", req);
    DEPS_DEBUG_PDI("pdi", pdi);

    dap_iter = req->required_list;
    while (dap_iter) {
        RCDepAndPackage *dap = (RCDepAndPackage *) dap_iter->data;
        RCPackageDep *dep = dap->dep;
        while (dep) {
            RCPackageDepItem *pkg_di = (RCPackageDepItem *) dep->data;
            if (rc_package_dep_item_equal (pkg_di, pdi)) {
                out_list = g_slist_append (out_list, dap);
            }
        /* FIXME */
/*         else {
 *             switch (rc_package_dep_item_is_subset (dap->dep, pdi)) {
 *                 case 0:
 *                 case 1:
 *                 case -1:
 *             }
 *         }
 */


            dep = dep->next;
        }
        dap_iter = dap_iter->next;
    }

    DEPS_DEBUG_INT ("out_list'length", g_slist_length (out_list));
    DEPS_EXIT;
    return out_list;
}


/*
 * If we have anything that failed, we need to go through this list
 * and try to find newer packages, if possible, or remove them
 * if nothing newer is found.
 */

static gboolean
process_failed_list (RCDepResolution *rcdr)
{
    GSList *it;

    if (rcdr->failed_daps == NULL)
        return FALSE;

    it = rcdr->failed_daps;
    while (it) {
        RCDepAndPackage *fail = (RCDepAndPackage *) it->data;
        RCPackage *newer_pkg;

        if (g_slist_find (rcdr->packages_added, fail->package)) {
            /* We're trying to install another version of this */
            /* Keep on' movin' */
            it = it->next;
            continue;
        }

        /* Now try to find a newer version of every package in this list */
        newer_pkg = deps_find_newer_package_for_dap (rcdr, fail);
        if (newer_pkg) {
            if (rc_deps_debug_level > 10) {
                rc_debug (RC_DEBUG_LEVEL_DEBUG, ">> found newer package %s to replace %s\n",
                         rc_package_spec_to_str (&newer_pkg->spec),
                         rc_package_spec_to_str (&fail->package->spec));
            }

            /* Check if this is in the remove list already */
            if (g_slist_find_custom (rcdr->packages_removed, newer_pkg, package_find_by_name)) {
                /* We're trying to remove this package, doh. */
                if (rc_deps_debug_level > 10) {
                    rc_debug (RC_DEBUG_LEVEL_DEBUG, "  >> not installing because we're removing it!\n");
                }
                /* nothing */
                ;
            } else {
                deps_add_package (rcdr, newer_pkg, TRUE); // force this?
            }
        } else {
            if (rc_deps_debug_level > 20) {
                rc_debug (RC_DEBUG_LEVEL_DEBUG, ">> found NO package to replace %s\n",
                         rc_package_spec_to_str (&fail->package->spec));
            }
            deps_remove_package (rcdr, fail->package, FALSE);
//            deps_update_remove_list (rcdr, fail->package);
        }

        g_free (fail);
        it = it->next;
    }
    g_slist_free (rcdr->failed_daps);
    rcdr->failed_daps = NULL;

    return TRUE;
}

static void
process_provides_iterator (gpointer key, gpointer value, gpointer data)
{
    gchar *name = (gchar *) key;
    RCDepProvideSList *provs = (RCDepProvideSList *) value;
    RCDepResolution *rcdr = (RCDepResolution *) data;
    GSList *lnk = rcdr->must_provide_list;


    while ((lnk = g_slist_find_custom (lnk, name, dap_find_by_name)) != NULL)
    {
        RCDepAndPackage *must_dap = (RCDepAndPackage *) lnk->data;
        if (must_dap->force) {
            while (provs) {
                RCDepProvide *prov = (RCDepProvide *) provs->data;

                if (!prov->removed && prov->provided_by != must_dap->package) {
                    deps_remove_package (rcdr, prov->provided_by, FALSE);
                }
                provs = provs->next;
            }
        }

        lnk = lnk->next;
    }
}


static void
process_provides_list (RCDepResolution *rcdr)
{
    g_hash_table_foreach (rcdr->provide_hash, process_provides_iterator, rcdr);
}


static void
process_missing_iterator (gpointer key, gpointer value, gpointer data)
{
    gchar *name = (gchar *) key;
    RCDepRequireSList *reqs = (RCDepRequireSList *) value;
    RCDepResolution *rcdr = (RCDepResolution *) data;

    /* First check and see if this thing is provided */
    while (reqs) {
        RCDepRequire *req = (RCDepRequire *) reqs->data;
        if (req->top_provide != NULL &&
            req->top_provide != REQUIRE_ALL_REMOVED &&
            req->top_provide != REQUIRE_HAS_CONFLICT)
        {
            return;
        }
        reqs = reqs->next;
    }

    /* If not, then we check if any of these are the result of an
     * inconsistent weak dependency. [thus indicating a conflict]
     */
    reqs = (RCDepRequireSList *) value;
    while (reqs) {
        RCDepRequire *req = (RCDepRequire *) reqs->data;
        RCPackageDepSList *req_deps = req->deps;
        GSList *reqit;

        if (rc_package_dep_slist_is_fully_weak (req_deps)) {
            reqs = reqs->next;
            continue;
        }

        rc_debug (RC_DEBUG_LEVEL_DEBUG, "Missing: %s\n", deps_join_names (req->names));

/* Try #0 */

#if 1
        while (req_deps) {
            RCPackageDep *reqdep = (RCPackageDep *) req_deps->data;
            gboolean nuked = FALSE;

            while (reqdep) {
                RCPackageDepItem *reqdepi = (RCPackageDepItem *) reqdep->data;

                if (reqdepi->relation == RC_RELATION_WEAK_NONE ||
                    reqdepi->relation == RC_RELATION_NONE)
                {
                    /* WTF? Get rid of this crap. */
                    RCDepAndPackageSList *dsl = find_daps_for_pdi (rcdr, req, reqdepi);
                    RCDepAndPackageSList *ds = dsl;
                    while (ds) {
                        RCDepAndPackage *dap = (RCDepAndPackage *) ds->data;
                        if (!dap->force) {
                            deps_remove_package (rcdr, dap->package, TRUE);
                            /* Note that we can't do much else other than break right now, since our dap
                             * might have gotten nuked from under us (!)
                             */
                            nuked = TRUE;
                            break;
                        } else {
                            /* We need to get rid of the /other/ stuff. So just undo one and hope for the best. */
                            RCDepAndPackageSList *pkg_daps = req->required_list;
                            while (pkg_daps) {
                                if (!g_slist_find (ds, pkg_daps->data)) {
                                    /* good enough for me! */
                                    RCDepAndPackage *old_bad = (RCDepAndPackage *) pkg_daps->data;
                                    if (!old_bad->force) {
                                        deps_remove_package (rcdr, old_bad->package, TRUE);
                                        nuked = TRUE;
                                        break;
                                    }
                                }
                                pkg_daps = pkg_daps->next;
                            }
                        }
                        if (nuked)
                            break;
                        ds = ds->next;
                    }
                    g_slist_free (dsl);
                }
                if (nuked)
                    break;
                reqdep = reqdep->next;
            }
            if (nuked)
                break;
            req_deps = req_deps->next;
        }
#endif

/* Try #1 */
#if 0
        while (req_deps) {
            RCPackageDep *dep = (RCPackageDep *) req_deps->data;
            while (dep) {
                RCPackageDepItem *depi = (RCPackageDepItem *) dep->data;

                if (depi->relation & RC_RELATION_WEAK_NONE || depi->relation == RC_RELATION_NONE) {
                    RCDepAndPackageSList *daps = req->required_list;
                    RCDepAndPackageSList *working_daps = NULL;
                    RCDepAndPackageSList *conflicting_daps = NULL;
                    RCPackage *conf_pkg = NULL;
                    gboolean make_conflict = TRUE;

                    /* weak relation. How do we know that this is what
                     * conflicts? We really don't, but we have to guess.
                     */
                    /* Find the package that has this weak bit */
                    while (daps) {
                        RCDepAndPackage *dap = (RCDepAndPackage *) daps->data;
                        RCPackageDep *dapdep = dap->dep;

                        while (dapdep) {
                            RCPackageDepItem *dapdepi = (RCPackageDepItem *) dapdep->data;

                            if ((dapdepi->relation == depi->relation) &&
                                rc_package_spec_equal (&dapdepi->spec, &depi->spec))
                            {
                                /* This is the thing with the weak dependency. */
                                conflicting_daps = g_slist_append (conflicting_daps, dap);
                            } else {
                                working_daps = g_slist_append (working_daps, dap);
                            }

                            dapdep = dapdep->next;
                        }
                        daps = daps->next;
                    }

                    /* VVVV */

                    while (conflicting_daps) {
                        RCDepAndPackage *cdap = (RCDepAndPackage *) conflicting_daps->data;
                        if (cdap->force) {
                            make_conflict = FALSE;
                            break;
                        } else {
                            conf_pkg = cdap->package;
                        }
                        conflicting_daps = conflicting_daps->next;
                    }

                    if (make_conflict) {
                        deps_remove_package (rcdr, conf_pkg, TRUE);
#if 0
                        deps_add_conflict (rcdr,
                                           RC_CONFLICT_WHO_KNOWS,
                                           NULL,
                                           conf_pkg,
                                           NULL, /* rc_package_dep_item_copy (depi), */
                                           NULL,
                                           NULL);
#endif
                    } else {
                        RCPackageSList *psl = NULL;
                        RCPackageSList *pslit;

                        /* We have to keep the things that are causing the conflict, so
                         * remove the other things instead.
                         */
                        while (working_daps) {
                            RCDepAndPackage *wdap = (RCDepAndPackage *) working_daps->data;
                            if (!g_slist_find (psl, wdap->package)) {
                                psl = g_slist_append (psl, wdap->package);
                            }
                            working_daps = working_daps->next;
                        }
                        pslit = psl;
                        while (pslit) {
                            RCPackage *pkg = (RCPackage *) pslit->data;
                            deps_remove_package (rcdr, pkg, TRUE);
                            pslit = pslit->next;
                        }
                        if (g_slist_find (psl, conf_pkg))
                            conf_pkg = NULL;
                        g_slist_free (psl);
                    }

                    if (req->top_provide) {
                        deps_remove_package (rcdr, req->top_provide->provided_by, TRUE);
                        if (conf_pkg && conf_pkg != req->top_provide->provided_by)
                            deps_remove_package (rcdr, conf_pkg, TRUE);
		    }

                    /* Try removing the top thing here? */
#if 1
                    req->top_provide = NULL;
#else
                    deps_require_unprovide (rcdr, req, FALSE);
#endif

		    /* We unprovide this thing now. */
                }
                dep = dep->next;
            }
            req_deps = req_deps->next;
        }
#endif

/* Try #2 */
#if 0
        while (req_deps) {
            RCPackageDep *dep = (RCPackageDep *) req_deps->data;
            while (dep) {
                RCPackageDepItem *depi = (RCPackageDepItem *) dep->data;

                if (depi->relation & RC_RELATION_WEAK) {
                    RCDepAndPackageSList *daps = req->required_list;
                    RCPackage *conf_pkg = NULL;
                    /* weak relation. How do we know that this is what
                     * conflicts? We really don't, but we have to guess.
                     */
                    /* Find the package that has this weak bit */
                    while (daps) {
                        RCDepAndPackage *dap = (RCDepAndPackage *) daps->data;
                        RCPackageDep *dapdep = dap->dep;
                        while (dapdep) {
                            RCPackageDepItem *dapdepi = (RCPackageDepItem *) dapdep->data;
                            if ((dapdepi->relation == depi->relation) &&
                                rc_package_spec_equal (&dapdepi->spec, &depi->spec))
                                conf_pkg = dap->package;
                            dapdep = dapdep->next;
                        }
                        if (conf_pkg)
                            break;

                        daps = daps->next;
                    }

                    deps_add_conflict (rcdr,
                                       RC_CONFLICT_WHO_KNOWS,
                                       NULL,
                                       conf_pkg,
                                       rc_package_dep_item_copy (depi),
                                       NULL);

		    if (req->top_provide) {
                        deps_remove_package (rcdr, req->top_provide->provided_by, TRUE);
                        if (conf_pkg && conf_pkg != req->top_provide->provided_by)
                            deps_remove_package (rcdr, conf_pkg, TRUE);
		    }

                    req->top_provide = REQUIRE_HAS_CONFLICT;
		    /* We unprovide this thing now. */
		    
                }
                dep = dep->next;
            }
            req_deps = req_deps->next;
        }
#endif

        reqs = reqs->next;
    }
}

static void
process_missing_list (RCDepResolution *rcdr)
{
    g_hash_table_foreach (rcdr->require_hash, process_missing_iterator, rcdr);
}


static RCPackageDepSList *
deps_dep_to_or_slist (RCPackageDep *dep)
{
    RCPackageDepSList *new_slist = NULL;

    while (dep) {
        RCPackageDepItem *old_item = (RCPackageDepItem *) dep->data;

        RCPackageDep *new_dep = rc_package_dep_new_with_item (rc_package_dep_item_copy (old_item));
        new_slist = g_slist_append (new_slist, new_dep);
        dep = dep->next;
    }

    return new_slist;
}

static void
deps_add_conflict (RCDepResolution *rcdr,
                   RCDepConflictType type,
                   gchar *what_str,
                   RCPackage *what,
                   RCPackageDepItem *what_conflicts,
                   RCPackage *with_other_package)
{
    RCDepConflict *conf;
    GSList *iter;

    DEPS_ENTER;

    iter = rcdr->packages_conflicting;
    while (iter) {
        RCDepConflict *conflict = (RCDepConflict *) iter->data;
        if (conflict->type == type &&
            (!conflict->what_str || !strcmp(conflict->what_str, what_str)) &&
            (!conflict->what || conflict->what == what) &&
            (!conflict->what_conflicts || conflict->what_conflicts == what_conflicts) &&
            (!conflict->with_other_package || conflict->with_other_package == with_other_package))
        {
            /* same exists */
            return;
        }
        iter = iter->next;
    }

    conf = g_new0 (RCDepConflict, 1);

    conf->type = type;
    conf->what_str = what_str;
    conf->what = what;
    conf->what_conflicts = what_conflicts;
    conf->with_other_package = with_other_package;

    rcdr->packages_conflicting = g_slist_append (rcdr->packages_conflicting, conf);

    DEPS_EXIT;
}


/* Debugging misc */
void
debug_deps_dump_dap (RCDepAndPackage *dap)
{
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "DAP: pkg: %s: ", rc_package_spec_to_str (&dap->package->spec));
    debug_rc_package_dep_dump (RC_DEBUG_LEVEL_DEBUG, dap->dep);
}


/* Evil evil itp hack because Vlad wouldn't listen to me ;-) */
void rc_dep_set_packman (RCPackman *packman)
{
    das_global_packman = packman;
}

/* Other stuff */
/* For debugging */
void
debug_rc_dep_print_results (int dlev, RCDepResolutionResult *rr)
{
    GSList *ql;

    rc_debug (dlev, "--- Dep Result ---\n");
    rc_debug (dlev, "Unresolved packages: %d\n",
              g_slist_length (rr->unresolved_reqs));
    ql = rr->unresolved_reqs;
    while (ql) {
        RCDepRequire *req = (RCDepRequire *) (ql->data);
        print_require (dlev, req);
        ql = ql->next;
    }

    rc_debug (dlev, "Conflicts: %d\n",
            g_slist_length (rr->packages_conflicting));
    ql = rr->packages_conflicting;
    while (ql) {
        RCDepConflict *dc = (RCDepConflict *) (ql->data);
        
        rc_debug (dlev, "  %s conflicts with %s (type %d): %s\n", dc->what_str, // rc_package_spec_to_str (&dc->what->spec),
                  dc->with_other_package ? rc_package_spec_to_str (&dc->with_other_package->spec) : "blah",
                  dc->type,
                  "(fixme)" /* debug_rc_package_dep_item_str (dc->what_conflicts) */);
        ql = ql->next;
    }

    rc_debug (dlev, "Packages to be installed: %d\n",
            g_slist_length (rr->packages_to_install));
    ql = rr->packages_to_install;
    while (ql) {
        RCPackage *pkg = (RCPackage *) ql->data;
        gchar *s = rc_package_spec_to_str(&pkg->spec);
        rc_debug (dlev, "  %s\n", s);
        g_free(s);
        ql = ql->next;
    }

#if 0
    rc_debug (dlev, "Packages to be replaced: %d\n",
            g_slist_length (rr->packages_to_replace));
    ql = rr->packages_to_replace;
    while (ql) {
        PackageReplaces *pl = (PackageReplaces *) ql->data;
        gchar *n = rc_package_spec_to_str (&pl->new_package->spec);
        gchar *o = rc_package_spec_to_str (&pl->old_package->spec);
        printf("  %s (replaces %s)\n", n, o);
        g_free(n);
        g_free(o);
        ql = ql->next;
    }
#endif

    rc_debug (dlev, "Packages to be removed: %d\n",
            g_slist_length (rr->packages_to_remove));
    ql = rr->packages_to_remove;
    while (ql) {
        RCPackage *pkg = (RCPackage *) ql->data;
        gchar *s = rc_package_spec_to_str(&pkg->spec);
        rc_debug (dlev, "  %s\n", s);
        g_free(s);
        ql = ql->next;
    }

#if 0
    rc_debug (dlev, "Packages with suggestions: %d\n", g_slist_length (rr->suggestions));
    ql = rr->suggestions;
    while (ql) {
        PackageSuggests *sug = (PackageSuggests *) ql->data;
        RCPackageDep *sugs = sug->suggested;
        gchar *n = rc_package_spec_to_str (&sug->pkg->spec);
        rc_debug (dlev, "  %s suggests: ", n);
        while (sugs) {
            RCPackageDep *dep = (RCPackageDep *) sugs->data;
            if (sugs != sug->suggested) rc_debug (dlev, ", ");
            while (dep) {
                RCPackageDepItem *it = (RCPackageDepItem *) dep->data;
                gchar *s = rc_package_spec_to_str (&it->spec);
                if (dep != sugs->data) rc_debug (dlev, " | ");
                rc_debug (dlev, "%s", s);
                g_free (s);
                dep = dep->next;
            }
            sugs = sugs->next;
        }
        rc_debug (dlev, "\n");
        g_free (n);
        ql = ql->next;
    }
    
    rc_debug (dlev, "Packages with recommendations: %d\n", g_slist_length (rr->recommendations));
    ql = rr->recommendations;
    while (ql) {
        PackageRecommends *rec = (PackageRecommends *) ql->data;
        RCPackageDep *recs = rec->recommended;
        gchar *n = rc_package_spec_to_str (&rec->pkg->spec);
        rc_debug (dlev, "  %s suggests: ", n);
        while (recs) {
            RCPackageDep *dep = (RCPackageDep *) recs->data;
            if (recs != rec->recommended) rc_debug (dlev, ", ");
            while (dep) {
                RCPackageDepItem *it = (RCPackageDepItem *) dep->data;
                gchar *s = rc_package_spec_to_str (&it->spec);
                if (dep != recs->data) rc_debug (dlev, " | ");
                rc_debug (dlev, "%s", s);
                g_free (s);
                dep = dep->next;
            }
            recs = recs->next;
        }
        rc_debug (dlev, "\n");
        g_free (n);
        ql = ql->next;
    }

#endif

    if (g_slist_length (rr->unresolved_reqs)) {
        rc_debug (dlev, "\nResolution FAILED\n");
    } else {
        rc_debug (dlev, "\nResolution SUCCEEDED\n");
    }
}






gchar *
rc_dep_require_to_string (RCDepRequire *req)
{
    GSList *wanted_by_packages = NULL;
    GSList *requirement_strings_lists = NULL;

    GSList *iter;
    GString *gstr;

    gchar *outstr;

    iter = req->deps;
    while (iter) {
        RCPackageDep *dep = (RCPackageDep *) iter->data;
        GSList *nu = NULL;

        while (dep) {
            RCPackageDepItem *di = (RCPackageDepItem *)dep->data;
            gchar *item_str;

            item_str = debug_rc_package_dep_item_str (di);
            nu = g_slist_append (nu, item_str);
            dep = dep->next;
        }

        requirement_strings_lists = g_slist_append (requirement_strings_lists, nu);
        
        iter = iter->next;
    }

    iter = req->required_list;
    while (iter) {
        RCDepAndPackage *dap = (RCDepAndPackage *) iter->data;

        if (!g_slist_find (wanted_by_packages, dap->package))
            wanted_by_packages = g_slist_append (wanted_by_packages, dap->package);

        iter = iter->next;
    }

    gstr = g_string_sized_new (100);
    iter = requirement_strings_lists;
    while (iter) {
        GSList *strs = (GSList *) iter->data;
        while (strs) {
            gchar *str = (gchar *) strs->data;
            g_string_append (gstr, str);
            g_free (str);

            if (strs->next)
                g_string_append (gstr, " or ");
            strs = strs->next;
        }

        g_slist_free (iter->data);

        if (iter->next)
            g_string_append (gstr, ", ");

        iter = iter->next;
    }

    g_slist_free (requirement_strings_lists);

    g_string_append (gstr, " is needed by ");

    iter = wanted_by_packages;
    while (iter) {
        RCPackage *pkg = (RCPackage *) iter->data;

        g_string_append (gstr, rc_package_spec_to_str (&pkg->spec));
        if (iter->next)
            g_string_append (gstr, ", ");
        iter = iter->next;
    }

    g_slist_free (wanted_by_packages);

    outstr = g_strdup (gstr->str);
    g_string_free (gstr, TRUE);

    return outstr;
}

RCPackageSList *
deps_require_list_packages (RCDepRequire *req)
{
    RCPackageSList *psl = NULL;
    RCDepAndPackageSList *daps;

    daps = req->required_list;
    while (daps) {
        RCDepAndPackage *dap = (RCDepAndPackage *) daps->data;
        if (!g_slist_find (psl, dap->package))
            psl = g_slist_append (psl, dap->package);
        daps = daps->next;
    }

    return psl;
}

static void free_provides_iterator (gpointer key, gpointer data, gpointer closure)
{
    gchar *name = (gchar *) key;
    RCDepProvideSList *provs = (RCDepProvideSList *) data;

    while (provs) {
        RCDepProvide *prov = (RCDepProvide *) provs->data;
        g_free (prov->name);
        g_slist_free (prov->required_by);
        provs = provs->next;
    }
    g_free (name);
    g_slist_free (data);
}

static void free_requires_iterator (gpointer key, gpointer data, gpointer closure)
{
    gchar *name = (gchar *) key;
    RCDepRequireSList *reqs = (RCDepRequireSList *) data;

    while (reqs) {
        RCDepRequire *req = (RCDepRequire *) reqs->data;
        // deps_require_free is buggy
        // deps_require_free (req);
        reqs = reqs->next;
    }
    g_free (name);
    g_slist_free (data);
}


void
rc_dep_resolution_free (RCDepResolution *rcdr)
{
    GSList *iter;

    g_hash_table_foreach (rcdr->provide_hash, free_provides_iterator, NULL);
    g_hash_table_foreach (rcdr->require_hash, free_requires_iterator, NULL);
    g_hash_table_destroy (rcdr->provide_hash);
    g_hash_table_destroy (rcdr->require_hash);

    g_slist_free (rcdr->preferred_chs);
    g_slist_free (rcdr->current_chs);
    g_slist_free (rcdr->other_chs);
    g_slist_free (rcdr->conflict_deps);
    g_slist_free (rcdr->packages_added);
    g_slist_free (rcdr->packages_removed);

    iter = rcdr->must_provide_list;
    while (iter) {
        g_free (iter->data);
        iter = iter->next;
    }
    g_slist_free (rcdr->must_provide_list);

    g_free (rcdr);
}

gboolean
deps_verify_package_for_and_dep_slist (RCDepResolution *rcdr, RCPackage *pkg, RCPackageDepSList *deps)
{
#if 0
    fprintf (stderr, ">>>> deps_verify_package_for_and_dep_slist (pkg: %s)\n",
             rc_package_spec_to_str (&pkg->spec));
#endif

    while (deps) {
        RCPackageDep *dep = (RCPackageDep *) deps->data;
        gboolean this_fails = FALSE;

        while (dep) {
            gchar *this_name;
            GSList *prov_iter;

            this_name = ((RCPackageDepItem *) (dep->data))->spec.name;
#if 0
            fprintf (stderr, "  this_name: %s\n", this_name);
#endif
            /* Find this provide in the package's thing */
            prov_iter = pkg->provides;
            while (prov_iter) {
                RCPackageDep *pp = (RCPackageDep *) prov_iter->data;
                while (pp) {
                    RCPackageDepItem *prov_item = (RCPackageDepItem *) pp->data;
#if 0
                    fprintf (stderr, "  checking provide: %s\n", debug_rc_package_dep_item_str (prov_item));
#endif
                    if (!strcmp (prov_item->spec.name, this_name)) {
                        /* Check with this one */
                        if (!rc_package_dep_verify_relation (dep, &prov_item->spec)) {
#if 0
                            fprintf (stderr, "  FAIL\n");
#endif
                            this_fails = TRUE;
                            break;
                        }
                    }
                    if (!this_fails)
                        break;
                    pp = pp->next;
                }
                if (!this_fails)
                    break;
                prov_iter = prov_iter->next;
            }

            if (this_fails)
                break;
            dep = dep->next;
        }

        if (!this_fails) {
#if 0
            fprintf (stderr, "<<<< TRUE\n");
#endif
            return TRUE;
        }

        deps = deps->next;
    }

#if 0
    fprintf (stderr, "<<<< FALSE\n");
#endif
    return FALSE;
}


/* We are about to remove this thing. We want to see whether there is something
 * that strongly depends on it (and not that it should be removed :-) and if so
 * make a conflict.
 */
static void
check_daps_for_removal (RCDepResolution *rcdr, RCDepRequire *req)
{
    GSList *strongly_want = NULL;
    GSList *strongly_dont_want = NULL;

    RCDepAndPackageSList *reqs;

    reqs = req->required_list;
    while (reqs) {
        RCDepAndPackage *dap = (RCDepAndPackage *) reqs->data;

        if (dap->force) {
            RCPackageDepItem *di = (RCPackageDepItem *) dap->dep->data;
            /* You know, I don't care about ORs at this point. */

            if (di->relation & RC_RELATION_NONE) {
                /* This dep is trying to get rid of this req */
                if (!g_slist_find (strongly_dont_want, dap->package))
                    strongly_dont_want = g_slist_append (strongly_dont_want, dap->package);
            } else if (!(di->relation & RC_RELATION_WEAK)) {
                /* This dep is trying to get this thing [and not weakly] */
                if (!g_slist_find (strongly_want, dap->package))
                    strongly_want = g_slist_append (strongly_want, dap->package);
            }
        }

        reqs = reqs->next;
    }

    if (strongly_want) {
        GSList *want_iter = strongly_want;

        while (want_iter) {
            RCPackage *want_pkg = (RCPackage *) want_iter->data;
            GSList *dont_want_iter = strongly_dont_want;

            while (dont_want_iter) {
                RCPackage *dont_want_pkg = (RCPackage *) dont_want_iter->data;

                deps_add_conflict (rcdr,
                                   RC_CONFLICT_PACKAGES,
                                   NULL,
                                   want_pkg,
                                   NULL,
                                   dont_want_pkg);
                dont_want_iter = dont_want_iter->next;
            }

            if (!strongly_dont_want) {
                /* So, something strongly wants this, but for some reason it's getting removed */
                /* and it's not because of another package here. */
                /* We make a conflict here. */
                deps_add_conflict (rcdr,
                                   RC_CONFLICT_REQUIREMENT,
                                   g_strconcat (deps_join_names (req->names), " is being removed", NULL),
                                   want_pkg,
                                   NULL,
                                   NULL);
            }

            want_iter = want_iter->next;
        }

    }

    g_slist_free (strongly_want);
    g_slist_free (strongly_dont_want);
}

static void
deps_upgrade_or_remove_package (RCDepResolution *rcdr, RCPackage *pkg, gboolean force)
{
    RCPackage *newer;

    newer = deps_find_newer_package (rcdr, pkg);
    if (newer) {
        deps_add_package (rcdr, newer, force);
    } else {
        deps_remove_package (rcdr, pkg, FALSE);
    }
}
