/*
 * rc-deps.h: Dependency resolution biits
 *
 * Copyright (C) 2000 Helix Code, Inc.
 *
 * Authors: Vladimir Vukicevic <vladimir@helixcode.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef _RC_DEPS_H
#define _RC_DEPS_H


#include <glib.h>

#include "rc-package.h"
#include "rc-package-dep.h"
#include "rc-channel.h"
#include "xml-util.h"

typedef struct _RCDepAndPackage RCDepAndPackage;
typedef GSList RCDepAndPackageSList;

typedef struct _RCDepRequire RCDepRequire;
typedef GSList RCDepRequireSList;

typedef struct _RCDepProvide RCDepProvide;
typedef GSList RCDepProvideSList;

typedef GHashTable RCDepRequireHashByName;
typedef GHashTable RCDepProvideHashByName;

typedef enum _RCDepConflictType RCDepConflictType;
typedef struct _RCDepConflict RCDepConflict;
typedef GSList RCDepConflictSList;

typedef struct _RCPackageAndInfo RCPackageAndInfo;

typedef struct _RCDepResolution RCDepResolution;
typedef struct _RCDepResolutionResult RCDepResolutionResult;


struct _RCDepAndPackage
{
    /* DO NOT FREE */
    RCPackageDep *dep;
    /* DO NOT FREE */
    RCPackage *package;
    gboolean force;
};


/* Used in the top_p field to indicate that a requre has a conflict */
#define REQUIRE_HAS_CONFLICT   ((RCDepProvide *) 0x1)
/* Used to indicate that this thing has been removed */
#define REQUIRE_ALL_REMOVED    ((RCDepProvide *) 0x2)

struct _RCDepRequire
{
    /* The name of this require */
    /* FREE */
    GSList *names;

    /* Restrictive list of deps that must be satisfied by this module */
    /* DEEP FREE */
    RCPackageDepSList *deps;    /* so that we can handle ORs. KILL ME NOW */

    /* list of deps that are satisfied by this module, along with their
     * packages
     */
    /* FREE SLIST & STRUCT */
    RCDepAndPackageSList *required_list;

    /* the topmost provide (the one that can currently satisfy 
     * all the deps_list)
     */
    /* DO NOT FREE (freed in hash) */
    RCDepProvide *top_provide;
};

typedef enum _RCDepProvideVirtualState RCDepProvideVirtualState;

enum _RCDepProvideVirtualState {
    VIRTUAL_UNKNOWN,
    VIRTUAL_UNSET,
    VIRTUAL_SET
};

struct _RCDepProvide
{
    /* The name of the provide (should match require) */
    /* FREE */
    gchar *name;

    /* what is provided */
    /* DO NOT FREE -- comes from RCPackage */
    RCPackageDepItem *provision;

    /* DO NOT FREE */
    RCPackage *provided_by;

    /* FREE SList */
    RCDepRequireSList *required_by;

    /* to resolve this name, this provide /must/ be used */
    gboolean force;

    /* this provide was removed in response to a remove request. it should
     * not be used in resolving dependencies.
     */
    gboolean removed;

    /* we want to avoid using this package, for whatever reason.
     */
    gboolean avoid;

    RCDepProvideVirtualState virtual;
};

enum _RCDepConflictType {
    RC_CONFLICT_WHO_KNOWS,
    RC_CONFLICT_PACKAGES,
    RC_CONFLICT_REQUIREMENT
};

struct _RCDepConflict
{
    RCDepConflictType type;

    gchar *what_str;
    RCPackage *what;
    RCPackageDepItem *what_conflicts;
    RCPackage *with_other_package;
};

struct _RCDepResolution {
    /* FREE KEYS */
    RCDepProvideHashByName *provide_hash;
    /* FREE KEYS */
    RCDepRequireHashByName *require_hash;

    RCPackageSList *depended_packages;
    RCPackageSList *processed_packages;

    /* Copy of passed in system packages */
    RCPackageSList *system_packages;

    /* Search order.. first preferred,
     * then current, then other */
    /* FREE SLists */
    RCChannelSList *preferred_chs;
    RCChannelSList *current_chs;
    RCChannelSList *other_chs;
    RCChannel *current_ch;
    gint user_pref;

    gint error;                 /* if an error occured */
    gint unprovided_packages;

    /* Things so that we can do some housekeeping */
    GSList *conflict_deps;

    /* FREE SList */
    RCPackageSList *packages_added;
    RCPackageSList *packages_removed;

    RCDepConflictSList *packages_conflicting;

    /* FREE SList-and-DAP */
    RCDepAndPackageSList *failed_daps;
    RCDepRequireSList *failed_reqs;

    /* FREE SList-and-DAP */
    RCDepAndPackageSList *must_provide_list;
    /* Only used during matching, should always be NULL unless
     * an iteration is in progress
     */
    RCDepRequireSList *reqs_to_remove;

    /* The history of failures on this package, so that
     * we can later figure out what happened if we can't
     * provide it
     */
    RCPackageDepItemSList *failures;
};

struct _RCDepResolutionResult {
    RCPackageSList *packages_to_install;
    RCPackageSList *packages_to_remove;
    RCDepRequireSList *unresolved_reqs;

    RCDepConflictSList *packages_conflicting;

    gboolean abort;
};



void deps_conflicts_use_virtual_packages (gboolean value);


RCDepResolution *
rc_dep_resolution_new (RCPackageSList *system_packages,
                       RCChannel *current_ch,
                       RCChannelSList *preferred_chs,
                       RCChannelSList *other_chs,
                       gint user_pref);

void
rc_dep_resolution_free (RCDepResolution *rcdr);

void
rc_dep_resolution_add_package (RCDepResolution *rcdr,
                               RCPackage *pkg);

void
rc_dep_resolution_add_package_by_name (RCDepResolution *rcdr,
                                       gchar *pkg_name);


void
rc_dep_resolution_remove_package (RCDepResolution *rcdr,
                                  RCPackage *pkg);

void
rc_dep_resolution_remove_package_by_name (RCDepResolution *rcdr,
                                          gchar *pkg_name);

RCDepResolutionResult *
rc_dep_resolution_get_result (RCDepResolution *rcdr);
void
rc_dep_resolution_result_free (RCDepResolutionResult *res);

gchar *
rc_dep_require_to_string (RCDepRequire *req);

/* debugging misc */
void debug_deps_dump_dap (RCDepAndPackage *dap);
void debug_rc_dep_print_results (int dlev, RCDepResolutionResult *rr);


#endif /* _RC_DEPS_H */
