/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* rc-package-info.c
 * Copyright (C) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gnome-xml/xmlmemory.h>
#include <gnome-xml/parser.h>

#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "rc-package-info.h"
#include "rc-debman-general.h"
#include "rc-debug.h"
#include "rc-util.h"
#include "xml-util.h"
#include "rc-package-update.h"

typedef enum _DepType DepType;

enum _DepType {
    DEP_TYPE_REQUIRES,
    DEP_TYPE_PROVIDES,
    DEP_TYPE_CONFLICTS,
    DEP_TYPE_SUGGESTS,
    DEP_TYPE_RECOMMENDS
};

static guint rc_subchannel_parse_helix (RCChannel *, gchar *);
static guint rc_subchannel_parse_debian (RCChannel *, gchar *);
static guint rc_subchannel_parse_redhat (RCChannel *, gchar *);

guint
rc_subchannel_parse (RCChannel *channel, gchar *tbuf, int compressed_length)
{
    gchar *buf;
    GByteArray *byte_array = NULL;
    guint ret = 0;

    g_assert (tbuf);

    if (compressed_length) {
        if (rc_uncompress_memory (tbuf, compressed_length, &byte_array)) {
            g_warning ("Uncompression failed");
            return 1;
        }

        buf = byte_array->data;
    } else {
        buf = tbuf;
    }

    switch (channel->type) {
    case RC_CHANNEL_TYPE_HELIX:
        ret = rc_subchannel_parse_helix (channel, buf);
        break;
    case RC_CHANNEL_TYPE_DEBIAN:
        ret = rc_subchannel_parse_debian (channel, buf);
        break;
    case RC_CHANNEL_TYPE_REDHAT:
        ret = rc_subchannel_parse_redhat (channel, buf);
        break;
    default:
        g_error ("Trying to parse unknown channel type!");
        break;
    }

    if (byte_array) {
        g_byte_array_free (byte_array, TRUE);
    }

    return (ret);
}

static guint
rc_subchannel_parse_helix (RCChannel *channel, gchar *buf)
{
    xmlDocPtr xml_doc;
    gint ret = 0;

    g_assert (buf);

    xml_doc = xmlParseMemory (buf, strlen (buf));

    if (!xml_doc) {
        ret = -1;
    }

    ret = rc_xml_node_to_channel (channel, xml_doc->root);

    xmlFreeDoc (xml_doc);

    return (ret);
}

static void debian_packages_helper (gchar *mbuf, RCPackage *pkg, gchar *url_prefix);

static guint
rc_subchannel_parse_debian (RCChannel *rcc, char *buf)
{
    /* Note that buf is NOT ours -- we can't munge it, and strdup'ing the
     * whole thing would be ufly. So, we let packages_query_helper handle
     */
    RCSubchannel *si = NULL;
    GSList *ps;

    gchar *b, *ob;

    si = rc_subchannel_new ();
    si->channel = rcc;
    si->name = g_strdup ("normal");
    si->preference = 100;
    si->packages = g_hash_table_new (g_str_hash, g_str_equal);

    si->dep_table = g_hash_table_new (rc_package_spec_hash,
                                      rc_package_spec_equal);

    /* Keep looking for a "Packages: " */
    ob = buf;
    while ((b = strstr (ob, "Package: ")) != NULL) {
        RCPackage *p;
        if (b != buf && b[-1] != '\r' && b[-1] != '\n') {
            ob = b+1;
            continue;
        }
        p = g_new0 (RCPackage, 1);
        debian_packages_helper (b, p, rcc->file_path);
        p->subchannel = si;
        ((RCPackageUpdate *)(p->history->data))->package = p;
        ob = b+1;

        ps = p->provides;
        while (ps) {
            RCPackageDep *dep = (RCPackageDep *) ps->data;
            while (dep) {
                RCPackageDepItem *depi = (RCPackageDepItem *) dep->data;
                g_hash_table_insert (si->dep_table, &depi->spec, p);
                dep = dep->next;
            }
            ps = ps->next;
        }
        g_hash_table_insert (si->packages, p->spec.name, p);
        g_hash_table_insert (si->dep_table, &p->spec, p);
    }
    rcc->subchannels = g_slist_append (rcc->subchannels, si);
    return 0;
}

/*
 * This code needs to be kept super fast
 */

static void
debian_packages_helper (gchar *mbuf, RCPackage *pkg, gchar *url_prefix)
{
    gchar *ibuf = mbuf;
    gchar *buf, *p;
    RCPackageUpdate *up = NULL;
    gboolean desc = FALSE;
    int ilen;

    up = rc_package_update_new ();

    while (1) {
        /* FIXME -- handle all empty line with spaces */
        if (*ibuf == '\0' || *ibuf == '\n') break;

        buf = strchr (ibuf, '\n');
        if (buf == NULL) break; /* something bizzare  happened */

        ilen = buf - ibuf;      /* the length of the line */
        buf = ibuf;             /* start of buf again */
        buf[ilen] = '\0';       /* we probably shouldn't modify the string in-place, but
                                 * we restore it at the end */
        ibuf += ilen + 1;

        /* Downcase the sucker - I hate debian - it's SUPPOSED to be nicely
         * capitalized, but.. */
        p = buf;
        if (*p != ' ')
            while (*p != ':') *p++ = tolower(*p);

        if (!strncmp (buf, "package: ", strlen ("package: "))) {
            pkg->spec.name = g_strndup (buf + strlen ("package: "),
                                        ilen - strlen ("package: "));
        } else if (!strncmp (buf, "installed-size: ", strlen ("installed-size: "))) {
            pkg->installed_size = strtoul (buf +
                                           strlen ("installed-size: "),
                                           NULL, 10) * 1024;
        } else if (!strncmp (buf, "size: ", strlen ("size: "))) {
            up->package_size = strtoul(buf + strlen("size: "), NULL, 10);
        } else if (!strncmp (buf, "description: ", strlen ("description: "))) {
            pkg->summary = g_strndup (buf + strlen ("description: "),
                                      ilen - strlen ("description: "));
            desc = TRUE;
        } else if (*buf == ' ' && desc) {
            if (ilen == 2 && buf[1] == '.') {
                /* It's just a newline */
                if (pkg->description) {
                    int l = strlen (pkg->description);
                    gchar *tmp = g_malloc (l + 2);
                    strcpy (tmp, pkg->description);
                    tmp[l] = '\n';
                    tmp[l+1] = '\0';
                    g_free (pkg->description);
                    pkg->description = tmp;
                } else {
                    pkg->description = g_strndup ("\n", 1);
                }
            } else {
                /* More than a newline */
                if (pkg->description) {
                    int l = strlen (pkg->description);
                    gchar *tmp = g_malloc (l + ilen + 1);
                    strcpy (tmp, pkg->description);
                    tmp[l] = '\n';
                    strncpy (tmp + l + 1, buf+1, ilen-1);
                    tmp[l + ilen] = '\0';
                    g_free (pkg->description);
                    pkg->description = tmp;
                } else {
                    pkg->description = g_strndup (buf+1, ilen-1);
                }
            }
        } else if (!strncmp (buf, "version: ", strlen ("version: "))) {
            rc_debman_parse_version (buf + strlen ("Version: "),
                                     &pkg->spec.epoch, &pkg->spec.version,
                                     &pkg->spec.release);
        } else if (!strncmp (buf, "section: ", strlen ("section: "))) {
            gchar **tokens = g_strsplit (buf + strlen ("section: "), "/", -1);
            gint i = 0;
            gchar *sec;
            while (tokens[i+1] != NULL) i++;
            sec = g_strchomp (tokens[i]);
            p = sec;
            while (*p) *p++ = tolower(*p);
            pkg->section = rc_debman_section_to_package_section (sec);
#if 0
            if (!strcmp (sec, "admin")) {
                pkg->section = RC_SECTION_SYSTEM;
            } else if (!strcmp (sec, "base")) {
                pkg->section = RC_SECTION_SYSTEM;
            } else if (!strcmp (sec, "comm")) {
                pkg->section = RC_SECTION_INTERNET;
            } else if (!strcmp (sec, "devel")) {
                pkg->section = RC_SECTION_DEVEL;
            } else if (!strcmp (sec, "doc")) {
                pkg->section = RC_SECTION_DOC;
            } else if (!strcmp (sec, "editors")) {
                pkg->section = RC_SECTION_UTIL;
            } else if (!strcmp (sec, "electronics")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "games")) {
                pkg->section = RC_SECTION_GAME;
            } else if (!strcmp (sec, "graphics")) {
                pkg->section = RC_SECTION_IMAGING;
            } else if (!strcmp (sec, "hamradio")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "interpreters")) {
                pkg->section = RC_SECTION_DEVELUTIL;
            } else if (!strcmp (sec, "libs")) {
                pkg->section = RC_SECTION_LIBRARY;
            } else if (!strcmp (sec, "mail")) {
                pkg->section = RC_SECTION_PIM;
            } else if (!strcmp (sec, "math")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "misc")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "net")) {
                pkg->section = RC_SECTION_INTERNET;
            } else if (!strcmp (sec, "news")) {
                pkg->section = RC_SECTION_INTERNET;
            } else if (!strcmp (sec, "oldlibs")) {
                pkg->section = RC_SECTION_LIBRARY;
            } else if (!strcmp (sec, "otherosfs")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "science")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "shells")) {
                pkg->section = RC_SECTION_SYSTEM;
            } else if (!strcmp (sec, "sound")) {
                pkg->section = RC_SECTION_MULTIMEDIA;
            } else if (!strcmp (sec, "tex")) {
                pkg->section = RC_SECTION_MISC;
            } else if (!strcmp (sec, "text")) {
                pkg->section = RC_SECTION_UTIL;
            } else if (!strcmp (sec, "utils")) {
                pkg->section = RC_SECTION_UTIL;
            } else if (!strcmp (sec, "web")) {
                pkg->section = RC_SECTION_INTERNET;
            } else if (!strcmp (sec, "x11")) {
                pkg->section = RC_SECTION_XAPP;
            } else {
                pkg->section = RC_SECTION_MISC;
            }
#endif
            g_strfreev (tokens);
        } else if (!strncmp (buf, "depends: ", strlen ("depends: "))) {
            pkg->requires = rc_debman_fill_depends (buf + strlen ("depends: "),
                                                    pkg->requires);
        } else if (!strncmp (buf, "recommends: ", strlen ("recommends: "))) {
            pkg->recommends = rc_debman_fill_depends (buf +
                                                      strlen ("recommends: "),
                                                      pkg->recommends);
        } else if (!strncmp (buf, "suggests: ", strlen ("suggests: "))) {
            pkg->suggests = rc_debman_fill_depends (buf +
                                                    strlen ("suggests: "),
                                                    pkg->suggests);
        } else if (!strncmp (buf, "pre-depends: ", strlen ("pre-depends: "))) {
            pkg->requires = rc_debman_fill_depends (buf +
                                                    strlen ("pre-depends: "),
                                                    pkg->requires);
        } else if (!strncmp (buf, "conflicts: ", strlen ("conflicts: "))) {
            pkg->conflicts = rc_debman_fill_depends (buf +
                                                     strlen ("conflicts: "),
                                                     pkg->conflicts);
        } /* else if (!strncmp (buf, "replaces: ", strlen ("replaces: "))) {
             pkg->conflicts = rc_debman_fill_depends (buf +
             strlen ("replaces: "),
             pkg->conflicts);
        } */ else if (!strncmp (buf, "provides: ", strlen ("provides: "))) {
            pkg->provides = rc_debman_fill_depends (buf +
                                                    strlen ("provides: "),
                                                    pkg->provides);

        } else if (!strncmp (buf, "filename: ", strlen ("filename: "))) {
            /* Build a new update with just this version */
            if (url_prefix) {
                up->package_url = g_strconcat (url_prefix, "/",
                                               buf + strlen ("filename: "),
                                               NULL);
            } else {
                up->package_url = g_strndup (buf + strlen ("filename: "),
                                             ilen - strlen ("filename: "));
            }
        } else if (!strncmp (buf, "md5sum: ", strlen ("md5sum: "))) {
            up->md5sum = g_strndup (buf + strlen ("md5sum: "),
                                    ilen - strlen ("md5sum: "));
        }

        buf[ilen] = '\n';
    }

    up->importance = RC_IMPORTANCE_SUGGESTED;
    up->description = g_strdup ("Upstream Debian release");
    rc_package_spec_copy (&up->spec, &pkg->spec);
    pkg->history = g_slist_append (pkg->history, up);

    /* Make sure to provide myself, for the dep code! */
    pkg->provides =
        g_slist_append (pkg->provides, rc_package_dep_new_with_item
                        (rc_package_dep_item_new_from_spec
                         (&pkg->spec,
                          RC_RELATION_EQUAL)));
}


static guint
rc_subchannel_parse_redhat (RCChannel *rcc, char *buf)
{
    g_error ("No redhat subchannel support!");
    return 0;
}

/* Helper functions */

RCPackage *
pkginfo_find_package_with_constraint (RCChannel *ch,
                                      gchar *name,
                                      gint max_pref,
                                      RCPackageDep *constraint,
                                      gboolean is_and)

{
    RCPackage *rp, *candidate = NULL;
    RCPackageUpdate *upi;
    RCSubchannel *si;
    GSList *my_chs;
    RCSubchannelSList *chs;

    chs = ch->subchannels;

    g_return_val_if_fail (ch != NULL, NULL);
    
#if 0
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "pfpwc: name: %s constr: ", name);
    rc_package_dep_dump (constraint);
#endif

    /* First try channel with exact preference */
    my_chs = chs;
    while (my_chs && ((RCSubchannel *) my_chs->data)->preference != max_pref)
        my_chs = my_chs->next;
    if (my_chs) {
        rp = g_hash_table_lookup (((RCSubchannel *) my_chs->data)->packages,
                                  name);
        if (rp) {
#if DEBUG > 10
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "   found %s in hash...\n", rp->spec.name);
#endif
            if (constraint) {
                g_assert(rp->history);
                g_assert(rp->history->data);
                
                upi = (RCPackageUpdate *) rp->history->data;
                if (is_and) {
                    if (rc_package_dep_verify_and_relation
                        (constraint, &(upi->spec), NULL, 0))
                    {
                        candidate = rp;
                    }
                } else {
                    if (rc_package_dep_verify_relation
                        (constraint, &(upi->spec)))
                    {
                        candidate = rp;
                    }
                }
            }
            else {
                candidate = rp;
            }
            if (candidate)
                return candidate;
#if DEBUG > 10
            if (constraint)
                rc_debug (RC_DEBUG_LEVEL_DEBUG, "   failed constraint test.\n");
            else
                rc_debug (RC_DEBUG_LEVEL_DEBUG, "   didn't exist!\n");
#endif
        }
    }

    my_chs = chs;
    while (my_chs) {
        si = (RCSubchannel *) my_chs->data;
        if (si->preference != max_pref) {
            rp = g_hash_table_lookup (((RCSubchannel *) my_chs->data)->packages,
                                      name);
            if (rp) {
                if (constraint) {
                    g_assert(rp->history);
                    g_assert(rp->history->data);
		    
                    upi = (RCPackageUpdate *) rp->history->data;
                    if (is_and) {
                        if (rc_package_dep_verify_and_relation
                            (constraint, &(upi->spec), NULL, 0))
                        {
                            candidate = rp;
                        }
                    } else {
                        if (rc_package_dep_verify_relation
                            (constraint, &(upi->spec)))
                        {
                            candidate = rp;
                        }
                    }
                }
                else {
                    candidate = rp;
                }
                if (candidate)
                    return candidate;
            }
        }
        my_chs = my_chs->next;
    }

    return NULL;
}

RCPackage *
pkginfo_find_package (RCChannel *chs,
                      gchar *name,
                      gint max_pref)
{
    return pkginfo_find_package_with_constraint (chs, name, max_pref,
                                                 NULL, FALSE);
}
