/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * cache.c: Caching of channel info
 *
 * Copyright (c) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <gnome.h>
#include <time.h>
#include <stdio.h>
#include <unistd.h>

#include "cache.h"
#include "gui-prefs.h"

#include <errno.h>

CacheName override_cache_name_func = NULL;

void
gui_cache_override_cache_name_func(CacheName cache_func)
{
    override_cache_name_func = cache_func;
} /* gui_cache_override_cache_name_func */

char *
gui_cache_get_filename(const char *filename)
{
    char *cachedir;
    char *modname, *fullpath;

    g_return_val_if_fail(filename, NULL);
    
    gui_prefs_get_cache_info(&cachedir, NULL, NULL, NULL);

    if (filename[0] == '/')
        filename++;

    if (!filename[0])
        filename = "index";

    modname = g_strdelimit(g_strdup(filename), "/", '-');
    fullpath = g_strconcat(cachedir, "/", modname, NULL);
    g_free(cachedir);
    g_free(modname);

    return fullpath;
} /* gui_cache_get_filename */

gboolean
gui_cache_fresh(const char *filename, time_t timestamp, int filesize)
{
    int cachediff;
    char *fullpath;
    struct stat file_stat;
    time_t cachetime;
    int r;

    if (override_cache_name_func)
        fullpath = (*override_cache_name_func)(filename);
    else
        fullpath = gui_cache_get_filename(filename);

    r = stat(fullpath, &file_stat);
    g_free(fullpath);
    if (r < 0)
        return FALSE;

    gui_prefs_get_cache_info(NULL, &cachediff, NULL, NULL);

    cachetime = file_stat.st_mtime;
    cachetime += cachediff;

    rc_debug(
        RC_DEBUG_LEVEL_DEBUG, 
        "::::::::::\n"
        "FNAME : %s\nCACHE : %d (%d)\n"
        "SERVER: %d\n"
        "---\n"
        "CACHE : %d\n"
        "SERVER: %d\n"
        "::::::::::\n\n", 
        filename, (int) cachetime, cachediff, (int) timestamp,
        (int) file_stat.st_size, filesize);

    if (cachetime >= timestamp && 
        (file_stat.st_size == filesize || filesize == -1))
        return TRUE;
    else
        return FALSE;
} /* gui_cache_fresh */

int
gui_cache_add(const char *filename, GByteArray *data)
{
    char *fullpath;
    FILE *f;
    char *cachedir;

    g_return_val_if_fail(data, -1);

    gui_prefs_get_cache_info(&cachedir, NULL, NULL, NULL);

    if (!g_file_exists(cachedir)) {
        if (rc_mkdir(cachedir, 0755)) {
            rc_debug (RC_DEBUG_LEVEL_WARNING, "Unable to create cache directory %s", cachedir);
            g_free(cachedir);
            return -1;
        }
    }
    g_free(cachedir);

    if (override_cache_name_func)
        fullpath = (*override_cache_name_func)(filename);
    else
        fullpath = gui_cache_get_filename(filename);

    f = fopen(fullpath, "w");
    g_free(fullpath);
    if (!f)
        return -1;
    fwrite(data->data, data->len, 1, f);
    fclose(f);
    return 0;
} /* gui_cache_add */

GByteArray *
gui_cache_get(const char *filename)
{
    char *fullpath;
    GByteArray *data;
    int fd;
    char buf[4096];
    int bytes;

    g_return_val_if_fail(filename, NULL);

    if (override_cache_name_func)
        fullpath = (*override_cache_name_func)(filename);
    else
        fullpath = gui_cache_get_filename(filename);
    fd = open(fullpath, O_RDONLY);
    if (fd < 0)
        return NULL;
    data = g_byte_array_new();
    do {
        bytes = read(fd, buf, 4096);
        if (bytes)
            data = g_byte_array_append(data, buf, bytes);
    } while (bytes > 0);

    return data;
} /* gui_cache_get */
