/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-channel.c: GUIChannel definition and associated functions
 *
 * Copyright 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <fcntl.h>
#include <crypt.h>
#include <time.h>

#include <gnome.h>
#include <glade/glade.h>
#include <gtkhtml/gtkhtml.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <libgnomeui/gnome-window-icon.h>

#include <gtkhtml/htmlimage.h>
#include <gtkhtml/htmlengine-search.h>

#include "gui.h"
#include "gui-transfer.h"
#include "gui-prefs.h"
#include "cache.h"
#include "gui-install.h"
#include "mirror.h"
#include "subscription.h"
#include <gal/widgets/e-scroll-frame.h>
#include <gal/shortcut-bar/e-shortcut-bar.h>

#include "libredcarpet.h"

#include "gui-channel.h"
#include "gui-util.h"
#include "gui-uri.h"
#include "gui-html.h"

#include "subscription.h"


/* Local callbacks */

static void sub_button_clicked(GtkWidget *widget, GUIChannel *channel);
static void unsub_button_clicked(GtkWidget *widget, GUIChannel *channel);
static void notebook_page_switched(GtkNotebook *notebook, 
                                   GtkNotebookPage *page, int page_num, 
                                   GUIChannel *channel);

static void channel_refresh_clicked(GtkWidget *item, GUIChannel *channel);
static void rc_gui_populate_option_menu(GtkOptionMenu *option_menu, 
                                        GUIChannel *channel);
static void set_subchannel_pref(GtkWidget *item, GUIChannel *channel);

GUIChannel *
rc_gui_channel_new (RCChannel *base_ch)
{
    GUIChannel *n;

    g_assert (base_ch);

    n = g_new0 (GUIChannel, 1);
    n->rcchannel = *base_ch;    /* ick */

    return n;
}

void
rc_gui_channel_clear_data (GUIChannel *channel)
{
    g_return_if_fail(channel);
    
    if (channel->html_data) {
        g_byte_array_free(channel->html_data, TRUE);
        channel->html_data = NULL;
    }

#if 0
    /* This crashes it right now, but shouldn't be strictly necessary, beacuse
       this info should be freed below in rc_subchannel_slist_free */
    rc_package_slist_free(channel->available_updates);
    rc_package_slist_free(channel->installed_packages);
    rc_package_slist_free(channel->uninstalled_packages);
    channel->available_updates = NULL;
    channel->installed_packages = NULL;
    channel->uninstalled_packages = NULL;
#endif

#if 1
    /* So I think this is what you want -- itp */
    g_slist_free (channel->available_updates);
    g_slist_free (channel->installed_packages);
    g_slist_free (channel->uninstalled_packages);
#endif

    rc_subchannel_slist_free(RCCHANNEL(channel)->subchannels);
    
    RCCHANNEL(channel)->subchannels = NULL;
} /* rc_gui_channel_clear_data */

/* This function will fill in a GByteArray of the channel's HTML info as well
   as fill in the subchannels field of the RCChannel. */
int
rc_gui_channel_fetch_data(GUIChannel *channel)
{
    RCChannel *rcchannel = RCCHANNEL(channel);
    GByteArray *data;
    const char *m;
    char *url;
    GUITransfer *t;
    gchar *info_url = NULL;
    int err;
    GUITransferError terr;

    RC_ENTRY;

    /* Set the subscription flag */
    /* This little if thingy here is a bit of a hack. Because
       gui_channel_subscribed always pulls it from the subscription file, even
       if you "subscribe" to a channel as non-root, whenever this function
       gets called, it'll return to you the real value (non-root can't write
       to this file to actually change the subscription). So, what we're gonna
       do is see if the channel->subscribed flag has been toggled by the
       "subscription" and if it is, don't bother to re-read from disk. However,
       if it isn't get the real value from disk. The only time this won't work
       correctly is if non-root is subscribed to a channel, becomes root,
       changes the subscriptions.xml file and then the non-root user 
       refreshes. But you'd be stupid if you did that, and fuck you if you 
       did: just use the damned GUI */
    if (!channel->subscribed || !appinfo->not_root)
        channel->subscribed = rc_gui_subscribed(rcchannel);

    if (channel->subscribed) {
        info_url = rcchannel->subs_url;
    } else {
        info_url = rcchannel->unsubs_url;
    }

    /* Fetch the info HTML */
    if (info_url) {
        /* Do nothing; we'll defer fetching and rendering the html until the
           user actually clicks on this channel */
    } else {
        /* No info HTML file? Well, let's create an array... */
        char *message;

        channel->html_data = g_byte_array_new();
        message = "No channel info available";
        channel->html_data = g_byte_array_append(
            channel->html_data, message, strlen(message));
    }

    /* Fetch the subchannel/package/dependency info and fill it in */
    m = gui_prefs_get_mirror_url();
    if (m && rcchannel->mirrored) {
        char *mirror_url;
        mirror_url = merge_mirror_url_and_path(m, rcchannel->path);
        /* XXX FIXME -- the above isn't really using pkginfo_file, that's
         * bad.
         */
        url = rc_build_url(
            NULL, NULL, mirror_url, rcchannel->pkginfo_file);
        g_free (mirror_url);
    } else {
        url = rc_build_url(
            "http", rc_gui_get_host (), rcchannel->path,        
            rcchannel->pkginfo_file);
    }

    t = gui_transfer_new(TRUE, FALSE);
    if (appinfo->splash) {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(e_splash_get_progress_bar (appinfo->splash)));
    } else {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(GET_WIDGET("single_progress")));
    }
    data = gui_transfer(t, url);
    g_free(url);

    terr = gui_transfer_get_error(t);
    gtk_object_destroy(GTK_OBJECT(t));
    if (terr == GUI_TRANSFER_ERROR_RETRY) {
        if (data)
            g_byte_array_free(data, FALSE);
        return rc_gui_channel_fetch_data(channel);
    }
    else if (terr == GUI_TRANSFER_ERROR_ABORT) {
        /* This channel thing is inaccessible, so we just skip it. */
        return -1;
    }

    data = g_byte_array_append(data, "\0", 1);

    err = rc_subchannel_parse(
        rcchannel, data->data, rcchannel->pkginfo_compressed ? data->len : 0);
    g_byte_array_free(data, TRUE);
    if (err) {
        gchar *s;
        gboolean retry;

        /* An error occurred. */
        s = g_strdup_printf(
            "<p>An error occured trying to parse channel information.</p>"
            "<p>This may be caused either by an old version of Red Carpet "
            "that is unable to read newer data files or by an error on the "
            "server. If you cancel, Red Carpet will quit.</p>"
            "<p>You should report this error to "
            "<a href=\"mailto:distribution@ximian.com\">"
            "distribution@ximian.com</a>.<br>"
            "Error code: CP-%d</p>", err);
        retry = rc_gui_error_box(s, RC_ERROR_RETRY);
        g_free(s);
        if (retry) {
            return rc_gui_channel_fetch_data(channel);;
        }
        else
            gtk_exit(1);
    }

#ifdef GUI_USE_PACKAGESETS
    /* Fetch the packageset data */
    m = gui_prefs_get_mirror_url();
    if (m && rcchannel->mirrored) {
        char *mirror_url;
        mirror_url = merge_mirror_url_and_path(m, rcchannel->path);
        /* XXX FIXME -- the above isn't really using pkginfo_file, that's
         * bad.
         */
        url = rc_build_url(
            NULL, NULL, mirror_url, "packagesets.xml");
        g_free (mirror_url);
    } else {
        url = rc_build_url(
            "http", rc_gui_get_host (), rcchannel->path, "packagesets.xml.gz");
    }
    t = gui_transfer_new(TRUE, TRUE);
    gui_transfer_set_progress(t, GTK_PROGRESS(GET_WIDGET("single_progress")));
    data = gui_transfer(t, url);
    g_free(url);
    gtk_object_destroy(GTK_OBJECT(t));

    /* If the transfer succeeds, then we have sets. otherwise, they're null */
    if (data && data->data) {
        data = g_byte_array_append(data, "\0", 1);

        rcchannel->package_sets = rc_package_set_parse (data->data, data->len);
        g_byte_array_free (data, TRUE);
    } else {
        g_byte_array_free (data, FALSE);
    }
#endif

    rc_gui_channel_update_package_lists(channel);
    
    RC_EXIT;
    return 0;
} /* rc_gui_channel_fetch_data */

void
rc_gui_channel_fetch_all_data(void)
{
    GSList *i;
    GtkProgress *progress = NULL, *single_progress = NULL;
    GtkLabel *progress_label = NULL;
    gint count = 0;

    RC_ENTRY;

    if (!appinfo->splash) {
        progress = GTK_PROGRESS(GET_WIDGET("total_progress"));
        single_progress = GTK_PROGRESS(GET_WIDGET("single_progress"));
        progress_label = GTK_LABEL(GET_WIDGET("progress_label"));
    }

    if (progress) {
        gtk_progress_configure(GTK_PROGRESS (progress), 0.0, 0.0,
                               (gfloat) g_slist_length (appinfo->channel_list));
    }

    i = appinfo->channel_list;
    while (i) {
        GUIChannel *c = i->data;
        char *label;

        rc_debug (RC_DEBUG_LEVEL_INFO, "Fetching %s\n", RCCHANNEL(c)->name);
        label = g_strdup_printf ("Downloading channel data (%s)",
                                 RCCHANNEL(c)->name);
        if (appinfo->splash) {
            e_splash_set_progress_text (appinfo->splash, label);
        } else {
            gtk_label_set_text (progress_label, label);
            gtk_progress_set_value (single_progress, 0.0);
        }

        g_free(label);

        if (rc_gui_channel_fetch_data(c)) {
            char *err;
            gboolean retry;

            err = g_strdup_printf(
                "<p>An error has occurred while downloading channel data.</p>"
                "<p>Red Carpet was unable to fetch information for the "
                "\"%s\" channel from the server. Red Carpet can continue "
                "without this information, but you will be unable to access "
                "that channel during this session (unless you refresh).</p>"
                "<p>If this problem persists, please report it to "
                "<a href=\"mailto:distribution@ximian.com\">"
                "distribution@ximian.com</a>.</p>",
                RCCHANNEL(c)->name);
            retry = rc_gui_error_box(err, RC_ERROR_RETRY);
            g_free(err);
            if (retry)
                continue;
            else {
                /* This failed for some reason; skip this channel. */
                /* this is semi-dangerous */
                rc_debug(
                    RC_DEBUG_LEVEL_WARNING, 
                    "Channel %d failed to download, skipping\n", 
                    RCCHANNEL(c)->id);
                i = i->next;
                appinfo->channel_list = g_slist_remove(
                    appinfo->channel_list, c);
                continue;
            }
        }

        if (progress) {
            gtk_progress_set_format_string (single_progress, "%p%%");
            gtk_progress_set_value (single_progress, 0.0);
            gtk_progress_set_value (progress, (gfloat) ++count);
            gtk_widget_draw (GTK_WIDGET (progress), NULL);       
        }
        
        i = i->next;
    }
    
    RC_EXIT;
} /* rc_gui_fetch_channel_all_data */

RCSubchannel *
rc_gui_channel_get_preferred_subchannel(GUIChannel *channel)
{
    RCSubchannel *lowest = NULL;
    GSList *i;

    i = RCCHANNEL(channel)->subchannels;
    while (i) {
        RCSubchannel *s = i->data;

        if (!lowest || s->preference < lowest->preference) {
            lowest = s;
        }

        if (s->preference == channel->subchannel_preference)
            return s;

        i = i->next;
    }

    /* If we can't match the preferred subchannel, return the lowest
       priority */

    return lowest;
} /* rc_gui_channel_get_preferred_subchannel */

/* FIXME: I moved this out of rc-channel.c, because it doesn't really
 * belong there, and it needs knowledge of the packman, which it can
 * get here.  Someday this belongs somewhere more general than here,
 * but not rc-channel.c ... */

static RCPackage *
rc_find_best_package (RCPackageDepItem *pdep, RCChannelSList *chs, gint user_pref)
{
    RCPackage *ret = NULL;
    RCPackageDep *dep = rc_package_dep_new_with_item (pdep);

    while (chs) {
        RCChannel *ch = (RCChannel *) chs->data;
        RCPackage *found;
        found = pkginfo_find_package_with_constraint (ch,
                                                      pdep->spec.name,
                                                      user_pref,
                                                      dep,
                                                      FALSE);
#if 0
        if (!found) {
            found = g_hash_table_lookup (ch->dep_table,
                                         &pdep->spec);
        }
#endif
        if (found) {
            if (ret) {
                if (rc_packman_version_compare (appinfo->pman, &found->spec,
                                                &ret->spec) > 0) {
                    ret = found;
                }
            } else {
                ret = found;
            }
        }
        chs = chs->next;
    }

    g_slist_free (dep);
    return ret;
}

void
rc_gui_channel_update_package_lists(GUIChannel *channel)
{
    RCPackageSList *channel_packages = NULL;
    RCSubchannel *subchannel;
    GHashTable *hash;
    GSList *i;
    RCChannelSList *sublist = NULL;
    RCChannelSList *unsublist = NULL;

    /* Get the list of packages in our preferred subchannel */
    subchannel = rc_gui_channel_get_preferred_subchannel(channel);
    g_return_if_fail(subchannel);
    hash = rc_hash_table_copy (subchannel->packages, g_str_hash, g_str_equal);

    /* Get the sub channels */
    rc_gui_split_channel_list (appinfo->channel_list, &sublist, &unsublist);

    /* Split up the package list into updates, installed, and
       available lists */

    channel->available_updates = NULL;
    channel->installed_packages = NULL;
    channel->uninstalled_packages = NULL;

    i = appinfo->system_packages;
    while (i) {
        RCPackage *package = i->data;
        RCPackage *p;

        p = g_hash_table_lookup(hash, RC_PACKAGE_SPEC(package)->name);

        if (!p) {
            i = i->next;
            continue;
        }

        package->section = p->section;

        channel->installed_packages = g_slist_append(
            channel->installed_packages, package);

        if (!package->hold && rc_packman_version_compare (
                appinfo->pman, RC_PACKAGE_SPEC (package),
                RC_PACKAGE_SPEC (p)) < 0) {
            /* There is an update for this package in this channel.
             * We need to make sure that there isn't another update for this 
             * same package in another channel.
             */
            RCPackage *other_pkg;
            RCPackageDepItem *di;

            di = rc_package_dep_item_new_from_spec(
                RC_PACKAGE_SPEC(p), RC_RELATION_GREATER);
            other_pkg = rc_find_best_package(
                di, sublist, channel->subchannel_preference);
            if (!other_pkg) {
                channel->available_updates = g_slist_append(
                    channel->available_updates, p);
            }
        }

        g_hash_table_remove(hash, RC_PACKAGE_SPEC(package)->name);

        i = i->next;
    }

    channel->uninstalled_packages = rc_package_hash_table_by_spec_to_list(
        hash);
    g_hash_table_destroy(hash);

    g_slist_free(channel_packages);
    
    g_slist_free (sublist);
    g_slist_free (unsublist);
} /* rc_gui_channel_update_package_lists */


static void
rc_gui_channel_connect_signals (GUIChannel *channel)
{
    GtkWidget *w;

    /* Connect the "switch-page" signal to the notebook */
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("notebook")), "switch-page",
        GTK_SIGNAL_FUNC(notebook_page_switched), channel);

    /* Connect the "link_clicked" signal to the HTML widgets */
    gtk_signal_connect(
        GTK_OBJECT(appinfo->update_html), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->installed_html), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->available_html), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->info_html), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);

    /* Connect clicked and selected signals to the subscription toolbar and
       menu items */
    w = GET_WIDGET("subscribe1");
    g_assert(w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(sub_button_clicked), 
        channel);
    
    w = GET_WIDGET("unsubscribe1");
    g_assert(w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(unsub_button_clicked),
        channel);

    /* Show and connect the signal to the channel refresh menu item and
       hide the summary refresh menu item */
    rc_gui_set_refresh_item(REFRESH_CHANNEL);
    w = GET_WIDGET("channel_refresh");
    g_assert(w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(channel_refresh_clicked), 
        channel);
} /* rc_gui_channel_connect_signals */

static void
rc_gui_channel_disconnect_signals(GUIChannel *channel)
{
    GtkWidget *w;

    /* not an error condition */
    if (!channel)
        return;

    gtk_signal_disconnect_by_func(
        GTK_OBJECT(GET_WIDGET("notebook")),
        GTK_SIGNAL_FUNC(notebook_page_switched), channel);
    
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->update_html), 
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->installed_html), 
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->available_html), 
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->info_html), 
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), channel);

    w = GET_WIDGET("subscribe1");
    g_assert(w);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(w), GTK_SIGNAL_FUNC(sub_button_clicked), channel);
    
    w = GET_WIDGET("unsubscribe1");
    g_assert(w);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(w), GTK_SIGNAL_FUNC(unsub_button_clicked), channel);

    w = GET_WIDGET("channel_refresh");
    g_assert(w);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(w), GTK_SIGNAL_FUNC(channel_refresh_clicked), channel);

} /* rc_gui_channel_disconnect_signals */

void
rc_gui_channel_show(GUIChannel *channel)
{
    GtkWidget *notebook;
    GtkWidget *w;

    RC_ENTRY;

    g_return_if_fail(channel);

    /* Get the notebook object */
    notebook = GET_WIDGET("notebook");
    g_assert(notebook);

    /* Hide the package HTML */
    gtk_widget_hide(GET_WIDGET("package_vbox"));

    /* First check to see if we are subscribed and sensitize the buttons */
    if (channel->subscribed) {
        rc_gui_html_sensitize_subscribe(FALSE);
        rc_gui_html_sensitize_unsubscribe(TRUE);
        gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);
        if (!getenv ("RC_DEBUG_TABS"))
            gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
    }
    else {
        rc_gui_html_sensitize_subscribe(TRUE);
        rc_gui_html_sensitize_unsubscribe(FALSE);
        if (!getenv ("RC_DEBUG_TABS"))
            gtk_notebook_set_show_tabs(GTK_NOTEBOOK(notebook), FALSE);
        gtk_notebook_set_show_border(GTK_NOTEBOOK(notebook), FALSE);
    }

    rc_gui_set_page ("minor_notebook", PAGE_DETAILS);

    /* Get the search menu item and sensitize it */
    gtk_widget_set_sensitive(GET_WIDGET("search"), TRUE);
    
    /* Populate the preferences tab. */
    if (g_slist_length(RCCHANNEL(channel)->subchannels) == 1) {
        w = gtk_notebook_get_nth_page(
            GTK_NOTEBOOK(GET_WIDGET("notebook")), PAGE_PREFERENCES);
        g_assert(w);
        gtk_widget_hide(w);
    }
    else {
        w = gtk_notebook_get_nth_page(
            GTK_NOTEBOOK(GET_WIDGET("notebook")), PAGE_PREFERENCES);
        g_assert(w);
        gtk_widget_show(w);
        w = GET_WIDGET("subchannel_optionmenu");
        g_assert(w);
        rc_gui_populate_option_menu(GTK_OPTION_MENU(w), channel);
    }
    
    /* If we are subscribed, then we want to make sure we have all of the
       tabs. If we are unsubscribed, then we just want to show the info page */
    
    if (channel->subscribed) {
        rc_gui_set_page ("notebook", PAGE_UPDATES);
    }
    else {
        rc_gui_set_page ("notebook", PAGE_INFO);
    }

    while (gtk_events_pending ())
        gtk_main_iteration ();

    if (channel->subscribed) {
        rc_gui_render_updates_html(channel);
    }
    else {
        rc_gui_render_info_html(channel);
    }
} /* rc_gui_channel_show */

void
rc_gui_channel_subscribe(GUIChannel *channel)
{
    rc_debug (RC_DEBUG_LEVEL_INFO, "Subscribing to: %s\n", RCCHANNEL(channel)->name);

    if (!appinfo->not_root)
        rc_gui_subscribe(RCCHANNEL(channel));

    channel->subscribed = TRUE;
    
    rc_gui_set_action_html(channel, ACTION_UPDATE, 0);
    rc_gui_channel_refresh_updates ();
    rc_gui_channel_show(channel);
} /* rc_gui_channel_subscribe */

void
rc_gui_channel_unsubscribe(GUIChannel *channel)
{
    rc_debug (RC_DEBUG_LEVEL_INFO, "Unsubscribing to: %s\n", RCCHANNEL(channel)->name);

    if (!appinfo->not_root)
        rc_gui_unsubscribe(RCCHANNEL(channel));

    channel->subscribed = FALSE;

    rc_gui_set_action_html(channel, ACTION_INFO, 0);
    rc_gui_channel_refresh_updates ();
    rc_gui_channel_show(channel);
} /* rc_gui_channel_unsubscribe */

/* CB */
static void
channel_refresh_clicked(GtkWidget *item, GUIChannel *channel)
{
    GtkWidget *window;

    /* Hide the package HTML */
    gtk_widget_hide(GET_WIDGET("package_vbox"));

    gtk_widget_set_sensitive(GET_WIDGET("channel_refresh"), FALSE);
    gtk_widget_set_sensitive(GET_WIDGET("search"), FALSE);
    if (channel->subscribed)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), FALSE);
    else
        gtk_widget_set_sensitive(GET_WIDGET("subscribe1"), FALSE);

    rc_gui_channel_clear_data(channel);

    gtk_progress_configure(
        GTK_PROGRESS(GET_WIDGET("single_progress")), 0.0, 0.0, 1.0);
    gtk_progress_configure(
        GTK_PROGRESS(GET_WIDGET("total_progress")), 0.0, 0.0, 1.0);
    gtk_widget_hide(GET_WIDGET("total_progress"));

    gtk_widget_show(GET_WIDGET("startup_pause"));
    gtk_widget_hide(GET_WIDGET("startup_resume"));

    window = GET_WIDGET("startup_window");
    gtk_label_set_text(
        GTK_LABEL(GET_WIDGET("progress_label")),
        "Downloading channel data...");
    gtk_window_set_modal(GTK_WINDOW(window), TRUE);
    gtk_widget_show(window);

    rc_gui_channel_fetch_data(channel);
    rc_gui_fetch_icon(channel);

    gtk_label_set_text(GTK_LABEL(GET_WIDGET("progress_label")), "Finished.");
    gtk_widget_hide(window);
    gtk_window_set_modal(GTK_WINDOW(window), FALSE);
    gtk_widget_show(GET_WIDGET("total_progress"));

    gtk_widget_set_sensitive(GET_WIDGET("channel_refresh"), TRUE);
    gtk_widget_set_sensitive(GET_WIDGET("search"), TRUE);
    if (channel->subscribed)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), TRUE);
    else
        gtk_widget_set_sensitive(GET_WIDGET("subscribe1"), TRUE);    
 
    rc_gui_channel_update_package_lists(channel);
    rc_gui_channel_show(channel);
} /* channel_refresh_clicked */

static void
notebook_page_switched(GtkNotebook *notebook, GtkNotebookPage *page,
                       int page_num, GUIChannel *channel)
{
    ActionType types[] = {
        ACTION_UPDATE,
        ACTION_REMOVE,
        ACTION_INSTALL,
        ACTION_PREFERENCES,
        ACTION_INFO,
        ACTION_SEARCH
    };

    if (types[page_num] == ACTION_UPDATE)
        rc_gui_render_updates_html(channel);
    else
        rc_gui_set_action_html(channel, types[page_num], 0);
} /* notebook_page_switched */

static void
sub_button_clicked(GtkWidget *widget, GUIChannel *channel)
{
    rc_gui_channel_subscribe(channel);
} /* sub_button_clicked */

static void
unsub_button_clicked(GtkWidget *widget, GUIChannel *channel)
{
    rc_gui_channel_unsubscribe(channel);
} /* unsub_button_clicked */

/*
 * rc_gui_channel_refresh_updates
 *
 * Go through each subscribed channel and referesh the available/other updates
 */
void rc_gui_channel_refresh_updates (void)
{
    GUIChannelSList *chiter;

    chiter = appinfo->channel_list;
    while (chiter) {
        GUIChannel *guic = (GUIChannel *) chiter->data;

        if (guic->subscribed)
            rc_gui_channel_update_package_lists (guic);

        chiter = chiter->next;
    }
}

/*
 * rc_gui_populate_option_menu
 *
 * populates option menu with each individual subchannel
 * and its name
 */
static void
rc_gui_populate_option_menu(GtkOptionMenu *option_menu, GUIChannel *channel)
{
    GtkWidget *menu;
    GtkWidget *item;
    GSList *i;
    int index = 0;

    menu = gtk_menu_new();

    i = RCCHANNEL(channel)->subchannels;
    while (i) {
        RCSubchannel *s = i->data;
        
        item = gtk_menu_item_new_with_label(s->name);
        gtk_object_set_data(GTK_OBJECT(item), "subchannel", s);
        gtk_signal_connect(
            GTK_OBJECT(item), "activate", 
            GTK_SIGNAL_FUNC(set_subchannel_pref), channel);
        gtk_widget_show(item);
        gtk_menu_append(GTK_MENU(menu), item);
        
        if (channel->subchannel_preference == s->preference)
            gtk_menu_set_active(GTK_MENU(menu), index);
		
        index++;
        i = i->next;
    }

    gtk_option_menu_set_menu(option_menu, menu);
} /* rc_gui_populate_option_menu */

/* CB from option menu */
static void
set_subchannel_pref(GtkWidget *item, GUIChannel *channel)
{
    RCSubchannel *s;

    s = gtk_object_get_data(GTK_OBJECT(item), "subchannel");
    channel->subchannel_preference = s->preference;
    rc_gui_channel_update_package_lists(channel);
    rc_gui_channel_show(channel);
} /* new_subchannel_pref */

void
rc_gui_channel_select(GUIChannel *channel)
{
    rc_gui_channel_disconnect_signals(appinfo->current_channel);
    appinfo->current_channel = channel;
    if (channel) {
        rc_gui_channel_connect_signals(channel);
        rc_gui_channel_show(channel);
    }
} /* rc_gui_channel_select */
