/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-init.c: GUI Initialization
 *
 * Copyright 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <fcntl.h>
#include <crypt.h>
#include <time.h>

#include <gnome.h>
#include <glade/glade.h>
#include <gtkhtml/gtkhtml.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <libgnomeui/gnome-window-icon.h>

#include <gal/e-paned/e-paned.h>

#include <gtkhtml/htmlimage.h>
#include <gtkhtml/htmlengine-search.h>

#include <gal/widgets/e-scroll-frame.h>
#include <gal/shortcut-bar/e-shortcut-bar.h>

#include "gui.h"
#include "gui-transfer.h"
#include "gui-prefs.h"
#include "cache.h"
#include "gui-install.h"
#include "mirror.h"
#include "subscription.h"

#include "libredcarpet.h"

#include "gui-register.h"
#include "gui-channel.h"
#include "gui-util.h"
#include "gui-html.h"
#include "gui-glade.h"
#include "gui-uri.h"
#include "gui-download.h"

#include "e-splash.h"

static guint display_level = RC_DEBUG_LEVEL_NONE;
static guint log_level = RC_DEBUG_LEVEL_INFO;

static struct poptOption options[] = {
    { "display-level", 'd', POPT_ARG_INT, &display_level, 0,
      "Visible debugging level, ranges from 0 (nothing) to 6 (everything)",
      NULL },
    { "log-level", 'l', POPT_ARG_INT, &log_level, 0,
      "Log file debugging level, ranges from 0 (nothing) to 6 (everything)",
      NULL },
    { NULL, '\0', POPT_ARG_NONE, NULL, 0, NULL, NULL }
};

AppInfo *appinfo;

static void rc_glade_init(void);
static void rc_setup_e_paned (void);
static void check_for_root(void);
static void content_area_init(void);
static void populate_navigation_bar(void);

static void cancel_clicked_cb(GtkWidget *button);
static void pause_clicked_cb(GtkWidget *button, gpointer data);
static void resume_clicked_cb(GtkWidget *button, gpointer data);

static GdkPixbuf *sb_icon_cb(EShortcutBar *sb, const gchar *url, gpointer data);

static void summary_refresh_clicked(GtkWidget *widget);
static void news_refresh_clicked(GtkWidget *widget);

static void rc_gui_channel_list_fetch(void);
static void rc_gui_news_fetch(void);
static void rc_gui_fetch_pretty_names (void);

static void
rc_glade_init(void)
{
    char *file;

    /* appinfo is a global variable */
    appinfo = g_new0(AppInfo, 1);

    file = "../redcarpet.glade";
    if (!g_file_exists (file))
        file = SHAREDIR"/redcarpet.glade";
        
    appinfo->xml = glade_xml_new(file, NULL);
    if (!appinfo->xml) {
        GtkWidget *w;
		char *s = g_strdup_printf (_("Could not find '%s'.\n"
                                     "Please make sure Red Carpet was "
                                     "installed correctly."), file);
		w = gnome_error_dialog (s);
		gnome_dialog_run_and_close (GNOME_DIALOG (w));
		exit (0);
	}

    glade_xml_signal_autoconnect(appinfo->xml);
} /* rc_glade_init */

static void
rc_setup_e_paned ()
{
    GtkWidget *container, *child, *paned;

    container = GET_WIDGET ("extra_widgets");
    paned = GET_WIDGET ("e-paned");

    child = GET_WIDGET ("shortcut_thingie");
    gtk_object_ref (GTK_OBJECT (child));
    gtk_container_remove (GTK_CONTAINER (container), child);
    e_paned_add1 (E_PANED (paned), child);
    gtk_object_unref (GTK_OBJECT (child));

    child = GET_WIDGET ("content_area");
    gtk_object_ref (GTK_OBJECT (child));
    gtk_container_remove (GTK_CONTAINER (container), child);
    e_paned_add2 (E_PANED (paned), child);
    gtk_object_unref (GTK_OBJECT (child));
}

static void
initialize_packman(void)
{
        /* Initialize the packman structure */
    do {
        RCPackmanError err;

        appinfo->pman = rc_distman_new();
        err = rc_packman_get_error(appinfo->pman);

        if (err) {
            char *reason;

            reason = g_strdup_printf(
                "Unable to access packaging subsystem:<br>"
                "%s", rc_packman_get_reason(appinfo->pman));
            rc_gui_error_box(
                reason, err == RC_PACKMAN_ERROR_FATAL ? RC_ERROR_FATAL :
                RC_ERROR_RETRY);

            g_free(reason);
            if (err == RC_PACKMAN_ERROR_FATAL)
                gtk_exit(1);
            else {
                gtk_object_unref(GTK_OBJECT(appinfo->pman));
                appinfo->pman = NULL;
            }
        }
    } while (!appinfo->pman);

    rc_dep_set_packman (appinfo->pman);
} /* initialize_packman */

static gboolean window_hideable = FALSE;

static gboolean
splash_hide_timeout (gpointer data)
{
    GtkWidget *splash = GTK_WIDGET (data);

    if (window_hideable) {
        gtk_widget_destroy (GTK_WIDGET (splash));
        appinfo->splash = NULL;
    } else {
        window_hideable = TRUE;
    }

    return FALSE;
}

static void
download_channel_info(void)
{
    ESplash *splash;
    GdkPixbuf *summary_icon, *news_icon, *featured_icon, *subscribed_icon;

    summary_icon = gdk_pixbuf_new_from_file (SHAREDIR "/summary.png");
    news_icon = gdk_pixbuf_new_from_file (SHAREDIR "/news.png");
    featured_icon = gdk_pixbuf_new_from_file (SHAREDIR "/featured.png");
    subscribed_icon = gdk_pixbuf_new_from_file (SHAREDIR "/subscribed.png");

    splash = E_SPLASH (e_splash_new (TRUE));
    appinfo->splash = splash;

    gtk_widget_show (GTK_WIDGET (splash));

    g_timeout_add (5000, (GSourceFunc) splash_hide_timeout, splash);

    e_splash_add_icon (splash, summary_icon);
    e_splash_add_icon (splash, news_icon);
    e_splash_add_icon (splash, subscribed_icon);
    e_splash_add_icon (splash, featured_icon);

    gdk_pixbuf_unref (summary_icon);
    gdk_pixbuf_unref (news_icon);
    gdk_pixbuf_unref (subscribed_icon);
    gdk_pixbuf_unref (featured_icon);

    e_splash_set_icon_highlight (splash, 0, TRUE);
    e_splash_set_progress_text (splash, "Downloading channel list...");

    rc_gui_channel_list_fetch ();
    e_splash_set_progress (splash, 0.0);

    e_splash_set_icon_highlight (splash, 1, TRUE);
    e_splash_set_progress_text (splash, "Downloading news...");

    rc_gui_news_fetch ();
    e_splash_set_progress (splash, 0.0);

    e_splash_set_icon_highlight (splash, 2, TRUE);
    e_splash_set_progress_text (splash, "Downloading channel data...");

    rc_gui_channel_fetch_all_data();
    e_splash_set_progress (splash, 0.0);

    e_splash_set_icon_highlight (splash, 3, TRUE);
    e_splash_set_progress_text (splash, "Downloading channel artwork...");

    rc_gui_fetch_all_icons();
    e_splash_set_progress (splash, 0.0);

    rc_gui_fetch_pretty_names();
    e_splash_set_progress (splash, 0.0);

    if (window_hideable) {
        gtk_widget_destroy (GTK_WIDGET (splash));
        appinfo->splash = NULL;
    } else { 
        window_hideable = TRUE;
    }
} /* download_channel_info */

static void
display_main_window(void)
{
    GtkWidget *main_window;

    main_window = GET_WIDGET("main_window");
    g_assert(main_window);

    gtk_widget_hide(GET_WIDGET("ok_button"));
    gtk_widget_hide(GET_WIDGET("cancel_button"));
    gtk_widget_hide(GET_WIDGET("resume_button"));
    gtk_widget_show(GET_WIDGET("pause_button"));
    rc_gui_set_refresh_item(REFRESH_SUMMARY);

    gtk_widget_set_sensitive(GET_WIDGET("summary_refresh"), FALSE);
    gtk_widget_set_sensitive(GET_WIDGET("search"), FALSE);
    gtk_widget_set_sensitive(GET_WIDGET("subscribe1"), FALSE);
    gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), FALSE);

    /* Create the model (ie. the channel lists) */
    gtk_widget_show_now(main_window);
    gtk_widget_draw(main_window, NULL);

    gtk_label_set_text(GTK_LABEL(GET_WIDGET("progress_label")), "");

    gtk_widget_set_sensitive(GET_WIDGET("summary_refresh"), TRUE);
    gtk_widget_set_sensitive(GET_WIDGET("search"), TRUE);

    /* Create the view */
#if 0
    populate_channel_list();
#else
    populate_navigation_bar();
#endif

    rc_gui_channel_select(NULL);
    rc_gui_show_executive_summary();
} /* display_main_window */    

static void
check_for_root(void)
{
    uid_t uid;

    uid = geteuid();
    if (uid) {
        GtkWidget *not_root_dialog;
        if (!getenv ("RC_ME_EVEN_HARDER")) {
            not_root_dialog = gnome_message_box_new(
                "You are not running Red Carpet as root.\n\n"
                "You will be unable to install or remove\n"
                "packages from your system and subscription\n"
                "information will not be saved during this\n"
                "session.", GNOME_MESSAGE_BOX_INFO, GNOME_STOCK_BUTTON_OK, NULL);
            gtk_window_set_modal(GTK_WINDOW(not_root_dialog), TRUE);
            gtk_widget_show(not_root_dialog);
            appinfo->not_root = TRUE;
        }
    }
} /* check_for_root */

static void
rc_gui_channel_list_fetch(void)
{
    GByteArray *data = NULL;
    RCChannelSList *rcchannel_list = NULL;
    GUIChannelSList *guichannel_list = NULL;
    GSList *i;
    const char *m;
    char *url;
    GUITransfer *t = NULL;
    GUITransferError terr;
    RCDistroType *dt;

    RCChannel *rc;
    GUIChannel *gc;

    RC_ENTRY;

    t = gui_transfer_new(TRUE, TRUE);
    if (appinfo->splash) {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS (e_splash_get_progress_bar (appinfo->splash)));
    } else {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(GET_WIDGET("single_progress")));
        gui_transfer_set_progress_label(
            t, GTK_LABEL(GET_WIDGET("progress_label")));
    }

    m = gui_prefs_get_mirror_url();

    if (m) {
        url = g_strconcat(m, "/channels.xml.gz", NULL);
    } else {
        url = rc_build_url(
            "http", rc_gui_get_host (), NULL, "channels.xml.gz");
    }

    data = gui_transfer(t, url);
    g_free(url);

    terr = gui_transfer_get_error(t);
    gtk_object_destroy(GTK_OBJECT(t));

    if (terr == GUI_TRANSFER_ERROR_RETRY) {
        if (data)
            g_byte_array_free(data, FALSE);
        return rc_gui_channel_list_fetch();
    }
    else if (terr == GUI_TRANSFER_ERROR_ABORT) {
        gtk_exit(1);
    }
    
    rcchannel_list = rc_channel_parse_xml(data->data, data->len);
    g_byte_array_free(data, TRUE);

    dt = rc_figure_distro ();
    rc_debug (RC_DEBUG_LEVEL_INFO, "Distro is %s\n", dt->unique_name);

    /* We need to construct GUIChannels from the RCChannels. */
    for (i = rcchannel_list; i; i = i->next) {
        rc = i->data;

        if (!getenv ("RC_ME_HARDER")) {
            char *compare_string;

            if (dt->pretend_name)
                compare_string = dt->pretend_name;
            else
                compare_string = dt->unique_name;

            if (!rc->distro_target) {
                /* Channel with no target? */
                rc_channel_free (rc);
                continue;
            }

            if (!g_slist_find_custom(
                    rc->distro_target, compare_string,
                    (GCompareFunc) strcmp)) {
                rc_channel_free (rc);
                continue;
            }
        }

        gc = rc_gui_channel_new (rc);

        guichannel_list = g_slist_append(guichannel_list, gc);
    }
    g_slist_free(rcchannel_list);

    appinfo->channel_list = guichannel_list;

    RC_EXIT;

} /* rc_gui_channel_list_fetch */

static void
rc_gui_news_fetch(void)
{
    GByteArray *data = NULL;
    GUITransfer *t;
    const char *m;
    char *url;
    GUITransferError terr;

    t = gui_transfer_new(TRUE, TRUE);
    if (appinfo->splash) {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS (e_splash_get_progress_bar (appinfo->splash)));
    } else {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(GET_WIDGET("single_progress")));
        gui_transfer_set_progress_label(
            t, GTK_LABEL(GET_WIDGET("progress_label")));
    }

    m = gui_prefs_get_mirror_url();
    if (m && 0)
        url = g_strconcat(m, "/red-carpet.rdf", NULL);
    else {
        url = rc_build_url(
            "http", rc_gui_get_host(), NULL, "red-carpet.rdf");
    }

    data = gui_transfer(t, url);
    g_free(url);

    terr = gui_transfer_get_error(t);
    gtk_object_destroy(GTK_OBJECT(t));

    if (terr == GUI_TRANSFER_ERROR_RETRY) {
        if (data)
            g_byte_array_free(data, FALSE);
        return rc_gui_news_fetch();
    }
    else if (terr == GUI_TRANSFER_ERROR_ABORT) {
        /* Oh well. No news (is good news) */
        appinfo->news = NULL;
        return;
    }

    appinfo->news = rc_gui_rdf_parse(data);
    if (!appinfo->news) {
        rc_debug (RC_DEBUG_LEVEL_WARNING, "unable to parse RDF file\n");
    }
    g_byte_array_free(data, TRUE);
} /* rc_gui_news_fetch */

static void
rc_gui_channel_list_clear(void)
{
    GSList *i;

    i = appinfo->channel_list;
    while (i) {
        GUIChannel *channel = i->data;

        if (channel->html_data) {
            g_byte_array_free(channel->html_data, TRUE);
            channel->html_data = NULL;
        }

        if (channel->icon_pb)
            gdk_pixbuf_unref(channel->icon_pb);

        rc_channel_free(RCCHANNEL(channel));

        i = i->next;
    }

    g_slist_free(appinfo->channel_list);
    appinfo->channel_list = NULL;
} /* rc_gui_channel_list_clear */

static void
rc_gui_news_clear(void)
{
    GSList *i;

    if (!appinfo->news)
        return;

    i = appinfo->news->items;
    while (i) {
        RDFItem *item = i->data;

        g_free(item->title);
        g_free(item->location);

        g_free(item);

        i = i->next;
    }
    g_slist_free(appinfo->news->items);

    g_free(appinfo->news->title);
    g_free(appinfo->news->icon);
    g_free(appinfo->news->location);
    g_free(appinfo->news);
    appinfo->news = NULL;
} /* rc_gui_news_clear */

/* CB */
static void
pause_clicked_cb(GtkWidget *button, gpointer data)
{
    gboolean startup = GPOINTER_TO_INT(data);
    GSList *transfers;
    GSList *i;

    if (startup) {
        gtk_widget_hide(GET_WIDGET("startup_pause"));
        gtk_widget_show(GET_WIDGET("startup_resume"));
    }
    else {
        gtk_widget_hide(GET_WIDGET("pause_button"));
        gtk_widget_show(GET_WIDGET("resume_button"));
    }

    transfers = gui_transfer_get_current_transfers();
    i = transfers;
    while (i) {
        GUITransfer *t = i->data;
        GnomeTransfer *gt = t->gt;

        gnome_transfer_pause(gt);

        i = i->next;
    }
} /* pause_clicked_cb */

/* CB */
static void
resume_clicked_cb(GtkWidget *button, gpointer data)
{
    gboolean startup = GPOINTER_TO_INT(data);
    GSList *transfers;
    GSList *i;

    if (startup) {
        gtk_widget_show(GET_WIDGET("startup_pause"));
        gtk_widget_hide(GET_WIDGET("startup_resume"));
    }
    else {
        gtk_widget_show(GET_WIDGET("pause_button"));
        gtk_widget_hide(GET_WIDGET("resume_button"));
    }

    transfers = gui_transfer_get_current_transfers();
    i = transfers;
    while (i) {
        GUITransfer *t = i->data;
        GnomeTransfer *gt = t->gt;

        gnome_transfer_resume(gt);

        i = i->next;
    }
} /* resume_clicked_cb */

/* CB */
static void
cancel_yes_clicked(GtkWidget *button)
{
    GSList *i;
    GUITransfer *t;
    int len;

    i = gui_transfer_get_current_transfers();
    len = g_slist_length(i);
    if (!len) {
        /* The timing was *just right* so that the transfer finished but was
           cancelled. */
        /* FIXME: So, this doesn't actually abort... We need to come up with 
           a way to make this thing bail out. Probably some logic in 
           gui-transfer.c will do it. */
        return;
    }
    if (len > 1) {
        rc_debug (RC_DEBUG_LEVEL_WARNING, "More than one transfer going on.");
    }
    t = (GUITransfer *) i->data;
    if (t->gt) {
        if (t->paused) {
            t->paused = FALSE;
            gtk_widget_show(GET_WIDGET("pause_button"));
            gtk_widget_hide(GET_WIDGET("resume_button"));
        }
        
        gnome_transfer_abort(t->gt);
    }
    
    t->aborted = TRUE;
} /* cancel_yes_clicked */

/* CB */
static void
cancel_no_clicked(GtkWidget *button)
{
    GSList *cpaused = gtk_object_get_data(
        GTK_OBJECT(GET_WIDGET("cancel_dialog")), "cpaused");
    GSList *i;

    i = cpaused;
    while (i) {
        GUITransfer *t = i->data;

        gnome_transfer_resume(t->gt);

        i = i->next;
    }

    g_slist_free(cpaused);
} /* cancel_no_clicked */

/* CB */
static void
cancel_clicked_cb(GtkWidget *button)
{
    GSList *transfers;
    GSList *cpaused = NULL;
    GSList *i;

    transfers = gui_transfer_get_current_transfers();
    i = transfers;
    while (i) {
        GUITransfer *t = i->data;

        if (!t->paused) {
            cpaused = g_slist_append(cpaused, t);
            gnome_transfer_pause(t->gt);
        }

        i = i->next;
    }

    gtk_object_set_data(
        GTK_OBJECT(GET_WIDGET("cancel_dialog")), "cpaused", cpaused);
    gtk_widget_show(GET_WIDGET("cancel_dialog"));
} /* cancel_clicked_cb */

/* Initialize the content area */
static void
content_area_init(void)
{
    GtkWidget *frame;
    GtkWidget *w;
    int i;

    RC_ENTRY;

    w = GET_WIDGET("minor_notebook");
    if (!getenv("RC_DEBUG_TABS"))
        gtk_notebook_set_show_tabs(GTK_NOTEBOOK(w), FALSE);
    gtk_notebook_set_show_border(GTK_NOTEBOOK(w), FALSE);

    w = GET_WIDGET("main_notebook");
    if (!getenv("RC_DEBUG_TABS"))
        gtk_notebook_set_show_tabs(GTK_NOTEBOOK(w), FALSE);
    gtk_notebook_set_show_border(GTK_NOTEBOOK(w), FALSE);

    appinfo->update_html    = gtk_object_get_user_data (
        GTK_OBJECT (GET_WIDGET ("updates_html")));
    appinfo->installed_html = gtk_object_get_user_data (
        GTK_OBJECT (GET_WIDGET ("installed_html")));
    appinfo->available_html = gtk_object_get_user_data (
        GTK_OBJECT (GET_WIDGET ("available_html")));
    appinfo->info_html      = gtk_object_get_user_data (
        GTK_OBJECT (GET_WIDGET ("info_html")));
    appinfo->exec_head      = gtk_object_get_user_data (
        GTK_OBJECT (GET_WIDGET ("exec_head")));
    appinfo->exec_body      = gtk_object_get_user_data (
        GTK_OBJECT (GET_WIDGET ("exec_body")));
    appinfo->action_html    = gtk_object_get_user_data (
        GTK_OBJECT (GET_WIDGET ("action_html")));
    appinfo->package_html   = gtk_object_get_user_data (
        GTK_OBJECT (GET_WIDGET ("package_html")));
    appinfo->download_html  = gtk_object_get_user_data (
        GTK_OBJECT (GET_WIDGET ("download_html")));

    /* Connect the signal to exit the application when the window is closed */
    w = GET_WIDGET("main_window");
    gtk_signal_connect(
        GTK_OBJECT(w), "delete_event", GTK_SIGNAL_FUNC(exit_clicked), NULL);
    w = GET_WIDGET("startup_window");
    gtk_signal_connect(
        GTK_OBJECT(w), "delete_event", GTK_SIGNAL_FUNC(exit_clicked), NULL);

    /* Connect the link clicked signal for the error HTML page and the
       executive summary HTML */
    frame = GET_WIDGET("error_html");
    w = gtk_object_get_user_data(GTK_OBJECT(frame));
    gtk_signal_connect(
        GTK_OBJECT(w), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), NULL);
    frame = GET_WIDGET("exec_body");
    w = gtk_object_get_user_data(GTK_OBJECT(frame));
    gtk_signal_connect(
        GTK_OBJECT(w), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), NULL);
    frame = GET_WIDGET("exec_head");
    w = gtk_object_get_user_data(GTK_OBJECT(frame));
    gtk_signal_connect(
        GTK_OBJECT(w), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), NULL);
    frame = GET_WIDGET("action_html");
    w = gtk_object_get_user_data(GTK_OBJECT(frame));
    gtk_signal_connect(
        GTK_OBJECT(w), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), NULL);
    frame = GET_WIDGET("package_html");
    w = gtk_object_get_user_data(GTK_OBJECT(frame));
    gtk_signal_connect(
        GTK_OBJECT(w), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), NULL);
    frame = GET_WIDGET("download_html");
    w = gtk_object_get_user_data(GTK_OBJECT(frame));
    gtk_signal_connect(
        GTK_OBJECT(w), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), NULL);
    frame = GET_WIDGET("close_html");
    w = gtk_object_get_user_data(GTK_OBJECT(frame));
    gtk_signal_connect(
        GTK_OBJECT(w), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), NULL);
    frame = GET_WIDGET("register_html");
    w = gtk_object_get_user_data(GTK_OBJECT(frame));
    gtk_signal_connect(
        GTK_OBJECT(w), "link_clicked",
        GTK_SIGNAL_FUNC(rc_gui_html_link_clicked), NULL);

    /* Connect the activate signal to the summary refresh menu item */
    w = GET_WIDGET("summary_refresh");
    g_assert(w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(summary_refresh_clicked),
        NULL);

    /* Connect the activate signal to the news refresh menu item */
    w = GET_WIDGET("news_refresh");
    g_assert(w);
    gtk_signal_connect(
        GTK_OBJECT(w), "activate", GTK_SIGNAL_FUNC(news_refresh_clicked),
        NULL);

    /* Connect handlers to the pause, resume, and cancel buttons */
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("pause_button")), "clicked",
        GTK_SIGNAL_FUNC(pause_clicked_cb), GINT_TO_POINTER(FALSE));
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("resume_button")), "clicked",
        GTK_SIGNAL_FUNC(resume_clicked_cb), GINT_TO_POINTER(FALSE));
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("startup_pause")), "clicked",
        GTK_SIGNAL_FUNC(pause_clicked_cb), GINT_TO_POINTER(TRUE));
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("startup_resume")), "clicked",
        GTK_SIGNAL_FUNC(resume_clicked_cb), GINT_TO_POINTER(TRUE));
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("cancel_button")), "clicked",
        GTK_SIGNAL_FUNC(cancel_clicked_cb), NULL);
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("cancel_yes_button")), "clicked",
        GTK_SIGNAL_FUNC(cancel_yes_clicked), NULL);
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("cancel_no_button")), "clicked",
        GTK_SIGNAL_FUNC(cancel_no_clicked), NULL);

    /* Glade seems broked -- This shit is SET in glade but it ignores it */
    gnome_dialog_close_hides(
        GNOME_DIALOG(GET_WIDGET("cancel_dialog")), TRUE);

    /* Ensure that only one of the Select All/None pairs is shown */
    rc_gui_set_select_item(ACTION_UPDATE);
    rc_gui_sensitize_select_items(FALSE);

    /* Set up the log objects */
    for (i = 0; i < LOG_LAST; i++)
        appinfo->log[i] = gui_log_new();

    RC_EXIT;
} /* content_area_init */

/* FIXME: Temporary */
static void
rc_gui_shortcut_selected(EShortcutBar *bar, GdkEvent *event, int group,
                         int item)
{
    char *uri;

    if (event->button.button != 1) {
        /* User right-clicked, which we're not handling now. */
        return;
    }

    e_shortcut_model_get_item_info(
        appinfo->model, group, item, &uri, NULL);
    rc_gui_handle_uri(NULL, uri, NULL);
} /* rc_gui_shortcut_selected */

static void
populate_navigation_bar(void)
{
    GtkWidget *vbox;

    vbox = GET_WIDGET("shortcut_thingie");

    if (!appinfo->bar) {
        gtk_widget_push_visual(gdk_rgb_get_visual());
        gtk_widget_push_colormap(gdk_rgb_get_cmap());

        appinfo->bar = e_shortcut_bar_new();

        gtk_widget_pop_colormap();
        gtk_widget_pop_visual();

        gtk_signal_connect(
            GTK_OBJECT(appinfo->bar), "item_selected",
            GTK_SIGNAL_FUNC(rc_gui_shortcut_selected), NULL);

        appinfo->model = e_shortcut_model_new();
        e_shortcut_bar_set_model(
            E_SHORTCUT_BAR(appinfo->bar), appinfo->model);
        
        e_shortcut_bar_set_icon_callback(
            E_SHORTCUT_BAR(appinfo->bar), sb_icon_cb, NULL);

        e_shortcut_model_add_group(
            appinfo->model, SHORTCUT_NAVIGATION, "Navigation");
#if 0
        /* FIXME: Add later... */
        e_shortcut_model_add_group(
            appinfo->model, SHORTCUT_SHORTCUTS, "Shortcuts");
#endif

        gtk_box_pack_start(GTK_BOX(vbox), appinfo->bar, TRUE, TRUE, 0);
    }
    else {
        int item, items, group, groups;

        groups = e_shortcut_model_get_num_groups(appinfo->model);
        for (group = 0; group < groups; group++) {
            items = e_shortcut_model_get_num_items(appinfo->model, group);
            for (item = items - 1; item >= 0; item--)
                e_shortcut_model_remove_item(appinfo->model, group, item);
        }
    }

    /* Okay, set up the icons for the navigation bar */
    e_shortcut_model_add_item(
        appinfo->model, SHORTCUT_NAVIGATION, -1, "updater:summary",
        "Summary");
    e_shortcut_model_add_item(
        appinfo->model, SHORTCUT_NAVIGATION, -1, "updater:news_page",
        "News");
#if 0
    /* This is disabled for now. */
    e_shortcut_model_add_item(
        appinfo->model, SHORTCUT_NAVIGATION, -1, "updater:featured_page",
        "Featured Channels");
#endif
    e_shortcut_model_add_item(
        appinfo->model, SHORTCUT_NAVIGATION, -1, "updater:subscribed_page",
        "Subscribed Channels");
    e_shortcut_model_add_item(
        appinfo->model, SHORTCUT_NAVIGATION, -1, "updater:unsubscribed_page",
        "Unsubscribed Channels");

    /* FIXME: In the future, pull shortcuts out of a file and add them to the
       shortcuts group */

    gtk_widget_show(appinfo->bar);
} /* populate_navigation_bar */

static void
summary_refresh_clicked(GtkWidget *widget)
{
    GtkWidget *window;

    gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "activate");

    gtk_widget_show(GET_WIDGET("pause_button"));
    gtk_widget_hide(GET_WIDGET("resume_button"));
    gtk_widget_hide(GET_WIDGET("ok_button"));
    gtk_widget_hide(GET_WIDGET("cancel_button"));

    gtk_widget_show(GET_WIDGET("startup_pause"));
    gtk_widget_hide(GET_WIDGET("startup_resume"));

    window = GET_WIDGET("startup_window");
    gtk_label_set_text(
        GTK_LABEL(GET_WIDGET("progress_label")),
        "Downloading channel list...");

    gtk_widget_show(window);
    gtk_window_set_modal(GTK_WINDOW(window), TRUE);

    rc_gui_channel_list_clear();
    rc_gui_channel_list_fetch();

    gtk_label_set_text(
        GTK_LABEL(GET_WIDGET("progress_label")),
        "Downloading news...");

    rc_gui_news_clear();
    rc_gui_news_fetch();

    gtk_label_set_text(
        GTK_LABEL(GET_WIDGET("progress_label")),
        "Downloading channel data...");

    rc_gui_channel_fetch_all_data();

    gtk_label_set_text(
        GTK_LABEL(GET_WIDGET("progress_label")),
        "Downloading channel data...");

    rc_gui_fetch_all_icons();
    rc_gui_fetch_pretty_names();

    gtk_label_set_text(GTK_LABEL(GET_WIDGET("progress_label")), "Finished.");

    gtk_widget_hide(window);
    gtk_window_set_modal(GTK_WINDOW(window), FALSE);

    rc_gui_set_page("main_notebook", PAGE_CHANNELS);
    gtk_widget_set_sensitive(GET_WIDGET("summary_refresh"), TRUE);
    gtk_widget_set_sensitive(GET_WIDGET("search"), TRUE);

#if 0
    populate_channel_list();
#endif
    rc_gui_channel_select(NULL);
    rc_gui_show_executive_summary();
} /* summary_refresh_clicked */

/* CB */
static void
news_refresh_clicked(GtkWidget *item)
{
    GtkWidget *window;

    gtk_widget_show(GET_WIDGET("startup_pause"));
    gtk_widget_hide(GET_WIDGET("startup_resume"));

    window = GET_WIDGET("startup_window");
    gtk_label_set_text(
        GTK_LABEL(GET_WIDGET("progress_label")),
        "Downloading news...");
    gtk_window_set_modal(GTK_WINDOW(window), TRUE);
    gtk_widget_show(window);

    rc_gui_news_clear();
    rc_gui_news_fetch();

    gtk_label_set_text(GTK_LABEL(GET_WIDGET("progress_label")), "Finished.");

    gtk_widget_hide(window);
    gtk_window_set_modal(GTK_WINDOW(window), FALSE);
    
    rc_gui_set_refresh_item(REFRESH_NEWS);

    rc_gui_set_page("minor_notebook", PAGE_SUMMARY);
    rc_gui_set_action_html(NULL, ACTION_NEWS, 0);
    rc_gui_show_news_page();
} /* news_refresh_clicked */

/* CB */
static GdkPixbuf *
sb_icon_cb (EShortcutBar *sb, const gchar *url, gpointer data)
{
    rc_debug(
        RC_DEBUG_LEVEL_INFO, "Shortcut bar asked for icon for: %s\n", url);

    if (strcmp(url, "updater:summary") == 0)
        return gdk_pixbuf_new_from_file(SHAREDIR"/summary.png");
    else if (strcmp(url, "updater:news_page") == 0)
        return gdk_pixbuf_new_from_file(SHAREDIR"/news.png");
    else if (strcmp(url, "updater:featured_page") == 0)
        return gdk_pixbuf_new_from_file(SHAREDIR"/featured.png");
    else if (strcmp(url, "updater:subscribed_page") == 0)
        return gdk_pixbuf_new_from_file(SHAREDIR"/subscribed.png");
    else if (strcmp(url, "updater:unsubscribed_page") == 0)
        return gdk_pixbuf_new_from_file(SHAREDIR"/unsubscribed.png");
    else
        return gdk_pixbuf_new_from_file(SHAREDIR"/summary.png");
} /* sb_icon_cb */

void
rc_gui_fetch_icon(GUIChannel *channel)
{
    RCChannel *rc;
    GUITransfer *t;
    char *url;
    GByteArray *data;
    GdkPixbufLoader *loader;
    GUITransferError terr;
    char *filename;
    char *cachedir;
    int err = 0;

    rc = RCCHANNEL(channel);

    if (!rc->icon_file) {
        rc_debug(
            RC_DEBUG_LEVEL_WARNING, "No icon for \"%s\" channel.", rc->name);
        return;
    }

    g_return_if_fail(rc->icon_file);

    t = gui_transfer_new(TRUE, TRUE);
    if (appinfo->splash) {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS (e_splash_get_progress_bar (appinfo->splash)));
    } else {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(GET_WIDGET("single_progress")));
        gui_transfer_set_progress_label(
            t, GTK_LABEL(GET_WIDGET("progress_label")));
    }

    url = maybe_mirror_channel_url (rc->icon_file, rc);

    data = gui_transfer(t, url);
    filename = g_strdup(t->filename);
    terr = gui_transfer_get_error(t);
    gtk_object_destroy(GTK_OBJECT(t));
    
    if (terr == GUI_TRANSFER_ERROR_RETRY) {
        if (data)
            g_byte_array_free(data, FALSE);
        return rc_gui_fetch_icon(channel);
    }
    else if (terr == GUI_TRANSFER_ERROR_ABORT) {
        channel->icon_pb = NULL;
        return;
    }

    /* Only do this if it's not a file URL and the user is root. */
    if (!appinfo->not_root && g_strncasecmp(url, "file://", 7) != 0) {
        char *file;

        gui_prefs_get_cache_info(&cachedir, NULL, NULL, NULL);

        if (cachedir) {
            file = gui_cache_get_filename(filename);
        }
        else {
            cachedir = g_strdup("/var/cache/redcarpet/icons");
            file = g_strdup_printf("%s/%d", cachedir, rc->id);
        }
        g_free(filename);
        
        if (!g_file_exists(cachedir)) {
            gboolean retry;
            
            do {
                err = rc_mkdir(cachedir, 0755);
                
                if (err) {
                    char *reason;
                    
                    reason = g_strdup_printf(
                        "Unable to create directory \"%s\" for local icon "
                        "storage", cachedir);
                    retry = rc_gui_error_box(reason, RC_ERROR_RETRY);
                    g_free(reason);
                }
            } while (err && retry);
        }
        
        /* No error occurred, let's go ahead and save this thing to disk and
           change the channel's icon URL to reflect the on disk version */
        if (!err) {
            int fd;
            
            fd = open(file, O_WRONLY | O_CREAT | O_TRUNC, 0644);
            rc_write(fd, data->data, data->len);
            rc_close(fd);
            
            /* This is kind of evil. */
            g_free(rc->icon_file);
            rc->icon_file = g_strconcat("file://", file, NULL);
        }
    }
    
    loader = gdk_pixbuf_loader_new();
    gdk_pixbuf_loader_write(loader, data->data, data->len);
    gdk_pixbuf_loader_close(loader);
    channel->icon_pb = gdk_pixbuf_loader_get_pixbuf(loader);
    gdk_pixbuf_ref(channel->icon_pb);
    gtk_object_unref(GTK_OBJECT(loader));
    g_byte_array_free(data, TRUE);
} /* rc_gui_fetch_icon */

void
rc_gui_fetch_all_icons(void)
{
    GSList *i;
    int count = 0;
    GtkProgress *total = NULL;

    gtk_progress_configure(
        GTK_PROGRESS(GET_WIDGET("single_progress")), 0.0, 0.0, 1.0);
    
    if (appinfo->splash) {
        total = GTK_PROGRESS(e_splash_get_progress_bar(appinfo->splash));
    }
    else {
        total = GTK_PROGRESS(GET_WIDGET("total_progress"));
    }

    gtk_progress_configure(
        GTK_PROGRESS (total), 0.0, 0.0, 
        (gfloat) g_slist_length(appinfo->channel_list));

    i = appinfo->channel_list;
    while (i) {
        GUIChannel *c = i->data;
        
        if (c->icon_pb)
            gdk_pixbuf_unref(c->icon_pb);

        rc_gui_fetch_icon(c);

        gtk_progress_set_value(GTK_PROGRESS(total), (gfloat) ++count);

        i = i->next;
    }
} /* rc_gui_fetch_all_icons */

static void
rc_gui_fetch_pretty_names (void)
{
    GByteArray *data = NULL;
    const char *m;
    char *url;

    GUITransfer *t = NULL;
    GUITransferError terr;

    t = gui_transfer_new(TRUE, TRUE);
    if (appinfo->splash) {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS (e_splash_get_progress_bar (appinfo->splash)));
    } else {
        gui_transfer_set_progress_bar(
            t, GTK_PROGRESS(GET_WIDGET("single_progress")));
        gui_transfer_set_progress_label(
            t, GTK_LABEL(GET_WIDGET("progress_label")));
    }

    m = gui_prefs_get_mirror_url();
    if (m && 0)
        url = g_strconcat(m, "/prettynames.xml.gz", NULL);
    else {
        url = rc_build_url(
            "http", rc_gui_get_host (), NULL, "prettynames.xml.gz");
    }
    data = gui_transfer(t, url);
    g_free(url);

    terr = gui_transfer_get_error(t);
    gtk_object_destroy(GTK_OBJECT(t));
    if (terr == GUI_TRANSFER_ERROR_RETRY) {
        if (data)
            g_byte_array_free(data, FALSE);
        rc_gui_fetch_pretty_names ();
        return;
    }
    else if (terr == GUI_TRANSFER_ERROR_ABORT)
        return;

    rc_pretty_name_parse_xml(data->data, data->len);
    g_byte_array_free (data, TRUE);
} /* rc_gui_fetch_pretty_names */

static void
rc_maybe_do_cache_stuff (int unused, void *wannabe_bool)
{
    gboolean cache;
    gboolean do_at_end, do_at_start;
    gboolean at_start = (gboolean) wannabe_bool;

    cache = gui_prefs_get_cache_info (NULL, NULL, &do_at_start, &do_at_end);

    if ((at_start && do_at_start) || (!at_start && do_at_end) || !cache)
    {
        rc_rmdir ("/var/cache/redcarpet/packages");
    }
} /* rc_maybe_do_cache_stuff */

int
main(int argc, char *argv[])
{ 
    FILE *log_file = NULL;

#undef malloc
    malloc (0);

    srand (time (NULL));

    /* GNOME intitializations */
    gnome_init_with_popt_table(
        "red-carpet", VERSION, argc, (char **) argv,
        options, 0, NULL);
    /*gnome_window_icon_set_default_from_file (SHAREDIR"/helix-gnome.png");*/
    glade_gnome_init();
#ifdef GTKHTML_HAVE_GCONF
    gconf_init(argc, argv, NULL);
#endif
    gdk_rgb_init();

    rc_debug_set_display_level (display_level);
    rc_debug_set_log_level (log_level);

    /* FIXME: this is probably a security problem waiting to happen
    if (!g_file_exists ("/var/log/redcarpet")) {
        rc_debug (RC_DEBUG_LEVEL_WARNING,
                  "/var/log/redcarpet does not exist, logging disabled\n");
    } else {
        if ((log_file = fopen ("/var/log/redcarpet/session.log", "w"))) {
            rc_debug_set_log_file (log_file);
        } else {
            rc_debug (RC_DEBUG_LEVEL_WARNING,
                      "Unable to open /var/log/redcarpet/session.log for "
                      "writing, logging disabled\n");
        }
    }
    */

    rc_gui_initialize_subscriptions();
    rc_glade_init();
    rc_setup_e_paned();
    content_area_init();
    rc_gui_init_install_page();
    initialize_packman();
    rc_gui_query_system_packages ();
    check_for_root();
    refresh_mirror_list(GTK_CLIST(GET_WIDGET("mirror_clist")));
    gui_prefs_load();

    rc_maybe_do_cache_stuff(0, (void *) TRUE);
    on_exit (rc_maybe_do_cache_stuff, (void *) FALSE);

    download_channel_info();

#if 0
    if (!appinfo->not_root && !rc_gui_registered())
        rc_gui_register();
#endif

    display_main_window();

    gtk_main();

    return 0;
} /* main */
