/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * gui-install.c: Routines for installing and removing packages in the GUI
 *
 * Copyright (C) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>

#include <glib.h>

#include <gnome.h>
#include <glade/glade.h>

#include <gui.h>
#include <gui-prefs.h>
#include <mirror.h>

#include "gui-download.h"
#include "gui-html.h"
#include "gui-install.h"
#include "gui-log.h"
#include "gui-log-display.h"
#include "gui-transfer.h"
#include "gui-util.h"

#include "rc-deps.h"
#include "rc-deps-util.h"
#include "rc-debug-misc.h"

static char *
make_filename(const char *url)
{
    char *tmp;
    char *pkg_dir;
    char *fp;
    char *filename;

    if (gui_prefs_get_cache_info(&tmp, NULL, NULL, NULL)) {
        pkg_dir = g_strconcat(tmp, "/packages", NULL);
        g_free(tmp);
    }
    else {
        pkg_dir = g_strdup("/var/cache/redcarpet/packages");
    }

    /* First create the directory */
    if (!g_file_exists(pkg_dir)) {
        int err;
        gboolean retry;

        do {
            err = rc_mkdir(pkg_dir, 0755);
            if (err) {
                gchar *error = g_strdup_printf(
                    "Unable to create directory %s for local package storage",
                    pkg_dir);
                retry = rc_gui_error_box(error, RC_ERROR_RETRY);
                g_free(error);
                if (!retry) {
                    g_free(pkg_dir);
                    return NULL;
                }
            }
        } while (err && retry);
    }
    
    fp = strrchr(url, '/');
    fp++;

    filename = g_strconcat(pkg_dir, "/", fp, NULL);
    g_free(pkg_dir);

    return filename;
} /* make_filename */

static void
verify_cancel_cb (GtkWidget *widget, gboolean *data)
{
    if (data)
        *data = FALSE;

    gtk_main_quit ();
} /* verify_cancel_cb */

static void
verify_ok_cb (GtkWidget *widget, gboolean *data)
{
    *data = TRUE;

    gtk_main_quit ();
} /* verify_ok_cb */

static void
dependency_failure(GUIChannel *current_channel, RCDepResolutionResult *rr)
{
    GtkWidget *verify_cancel;

    rc_gui_render_dependency_failure(current_channel, rr);

    gtk_widget_set_sensitive (GET_WIDGET ("verify_ok"), FALSE);
    verify_cancel = GET_WIDGET ("verify_cancel");
    gtk_widget_set_sensitive (verify_cancel, TRUE);
    
    gtk_signal_connect (GTK_OBJECT (verify_cancel), "clicked",
                        GTK_SIGNAL_FUNC (verify_cancel_cb),
                        NULL);
    gtk_widget_show (verify_cancel);
    
    rc_gui_set_page ("main_notebook", PAGE_VERIFY);
    
    gtk_main ();
} /* dependency_failure */

static gboolean
install_dependency_check(RCPackageSList *requested_pkgs,
                         GUIChannel     *current_channel,
                         RCChannelSList *subscribed_channels,
                         RCChannelSList *unsubscribed_channels,
                         GHashTable     **added_hash,
                         RCPackageSList **added_pkgs,
                         RCPackageSList **removed_pkgs)
{
    RCDepResolution  *rcdr;
    RCDepResolutionResult *rcrr;
    RCPackageSList   *piter;
    GSList           *iter;

    rcdr = rc_dep_resolution_new (appinfo->system_packages,
                                  RCCHANNEL (current_channel),
                                  subscribed_channels,
                                  unsubscribed_channels,
                                  current_channel?
                                  current_channel->subchannel_preference:0);

    /* Add all of the requested packages to the black box of dependency
       loving */
    for (piter = requested_pkgs; piter; piter = piter->next) {
        RCPackage *pkg = (RCPackage *)(piter->data);

        rc_dep_resolution_add_package (rcdr, pkg);
    }

    /* Maybe it works the first time? ;) */
    rcrr = rc_dep_resolution_get_result (rcdr);
    debug_rc_dep_print_results (RC_DEBUG_LEVEL_INFO, rcrr);

    if (rcrr->abort) {
        gchar *text = g_strdup_printf(
            "<p>An error has occurred while Red Carpet was processing the "
            "dependencies for this transaction. This may be caused by an "
            "error in the package or on the server. You may want to try the "
            "transaction again with fewer packages. You should report this "
            "problem to <a href=\"distribution@ximian.com\">"
            "distributon@ximian.com</a> as soon as possible.</p>"
            "<p>Information from the dependency system:<br> Recursion too "
            "deep.</p>");
        rc_dep_resolution_free (rcdr);
        rc_gui_error_box(text, RC_ERROR_ABORT);
        g_free(text);
        return FALSE;
    }

    /* Go through the conflicts and get rid of held base packages */

#if 0
/******************************************************************************************/
    for (iter = rr->conflicts; iter; iter = iter->next) {
        PackageConflicts *pc = (PackageConflicts *)(iter->data);
        RCPackage *pkg = pc->the_package;
        RCPackageDepItem *di = pc->the_conflict;

        if (pc->type == CONF_PACKAGE_HELD) {
            deps_resolution_remove_package (ri, pc->base_package);
        } else if (pc->type == CONF_NEED_NEWER || pc->type == CONF_TOO_NEW) {
            deps_resolution_uninstall_named_package (ri, pc->the_conflict->spec.name);
        } else if (pc->type == CONF_CANT_INSTALL) {
            deps_resolution_remove_package (ri, pc->base_package);
        }
    }

    if (g_slist_length (rr->conflicts) != 0) {
        deps_resolution_result_destroy (rr);
        rr = deps_resolution_get_result (ri);
        deps_print_results (rr);
    }

    while ((g_slist_length (rr->conflicts) > 0) &&
           (++level < MAX_DEPTH) &&
           (g_slist_length (rr->packages_not_provided) == 0)) {
        PackageConflicts *pc = (PackageConflicts *)(rr->conflicts->data);
        RCPackageDepItem *di = pc->the_conflict;

        deps_resolution_result_destroy (rr);

        deps_resolution_uninstall_named_package (ri, di->spec.name);

        rr = deps_resolution_get_result (ri);
        deps_print_results (rr);
    }

    if ((g_slist_length (rr->conflicts) > 0)) {
        /* Whoops, looks like we were unable to resolve some conflicts after
           MAX_DEPTH of searching.  In an optimal world, we'd tell the user
           something helpful here, like why we failed.  For now, we'll just
           fail.  It's easier. */
        rc_debug (RC_DEBUG_LEVEL_WARNING, "##### CANNOT RESOLVE CONFLICT ERROR\n");
        deps_print_results (rr);
        dependency_failure(current_channel, rr);
        return (FALSE);
    }
#endif

    if (g_slist_length (rcrr->unresolved_reqs) || 
        g_slist_length (rcrr->packages_conflicting)) {
        /* Whoops, looks like there were some packages we just couldn't find.
           How strange.  I can't imagine any severely broken distributions like
           that.  No, wait, I can. */
        rc_debug (RC_DEBUG_LEVEL_WARNING, "##### DEPS WARNING\n");
        debug_rc_dep_print_results (RC_DEBUG_LEVEL_WARNING, rcrr);
        dependency_failure (current_channel, rcrr);
        rc_dep_resolution_free (rcdr);
        return (FALSE);
    }

#if 0
    /* Fix up the packages_needed bit to make sense */
    g_slist_free (*requested_pkgs);
    *requested_pkgs = g_slist_copy (rr->packages_requested);
#endif

    *added_hash = g_hash_table_new (NULL, NULL);

    /* Looks like we've succeeded in some capacity.  We'll hash all of the
       non-requested-but-required packages by channel id, so that we can get
       them out by channel. */
    for (iter = rcrr->packages_to_install; iter; iter = iter->next) {
        RCPackage      *pkg = (RCPackage *)(iter->data);
        RCPackageSList *pkg_list;

        if (g_slist_find_custom (requested_pkgs,
                                 pkg, rc_package_spec_not_equal)) {
            continue;
        }

        pkg_list = g_hash_table_lookup(
            *added_hash, (gpointer)pkg->subchannel->channel->id);

        pkg_list = g_slist_append (pkg_list, pkg);
        *added_pkgs = g_slist_append(*added_pkgs, pkg);

        g_hash_table_insert(
            *added_hash, (gpointer) pkg->subchannel->channel->id,
            (gpointer) pkg_list);
    }

    for (iter = rcrr->packages_to_remove; iter; iter = iter->next) {
        RCPackage *pkg = (RCPackage *)(iter->data);

        rc_debug(RC_DEBUG_LEVEL_INFO, "Need to remove:  %s\n", pkg->spec.name);

        *removed_pkgs = g_slist_append(*removed_pkgs, pkg);
    }

    rc_dep_resolution_free (rcdr);

    return (TRUE);
} /* install_dependency_check */

static gboolean
dependency_confirm(gboolean       install,
                   RCPackageSList *requested_pkgs,
                   GUIChannel     *current_channel,
                   RCChannelSList *subscribed_channels,
                   RCChannelSList *unsubscribed_channels,
                   RCPackageSList *added_pkgs,
                   RCPackageSList *removed_pkgs)
{
    GtkWidget *verify_cancel;
    GtkWidget *verify_ok;
    gboolean go_on;
    gboolean unsub;
    gboolean can_go_on = TRUE;

    if (!requested_pkgs && !removed_pkgs && !added_pkgs)
        can_go_on = FALSE;

    rc_gui_render_dependencies(
        requested_pkgs, removed_pkgs, added_pkgs, current_channel, !install);

    verify_cancel = GET_WIDGET("verify_cancel");
    verify_ok = GET_WIDGET("verify_ok");
    gtk_widget_show(verify_cancel);
    gtk_widget_show(verify_ok);

    gtk_widget_set_sensitive (verify_ok, can_go_on ? TRUE : FALSE);

    gtk_signal_connect(
        GTK_OBJECT(verify_cancel), "clicked",
        GTK_SIGNAL_FUNC(verify_cancel_cb), &go_on);
    gtk_signal_connect(
        GTK_OBJECT(verify_ok), "clicked",
        GTK_SIGNAL_FUNC(verify_ok_cb), &go_on);

    rc_gui_set_page("main_notebook", PAGE_VERIFY);

    unsub = GTK_WIDGET_IS_SENSITIVE(GET_WIDGET("unsubscribe1"));
    rc_gui_set_menubar_sensitive (FALSE);

    gtk_main ();

    rc_gui_set_menubar_sensitive (TRUE);
    if (unsub)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), TRUE);

    gtk_signal_disconnect_by_func(
        GTK_OBJECT(verify_cancel),
        GTK_SIGNAL_FUNC(verify_cancel_cb), &go_on);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(verify_ok),
        GTK_SIGNAL_FUNC(verify_ok_cb), &go_on);

    if (go_on) {
        int r;

        r = g_slist_length(removed_pkgs);
        if (!install)
            r += g_slist_length(requested_pkgs);

        if (r > 6) {
            int button;
            GtkWidget *d = GET_WIDGET("remove_dialog");
            
            gnome_dialog_close_hides(GNOME_DIALOG(d), TRUE);
            button = gnome_dialog_run_and_close(GNOME_DIALOG(d));

            go_on = button == 0 ? TRUE : FALSE;
        }
    }

    return go_on;
} /* dependency_confirm */

static void
configure_progress_cb(RCPackman *p, int amount, int total)
{
    gtk_progress_set_value(
        GTK_PROGRESS(GET_WIDGET("single_progress")), (gfloat) amount + 1.0);
} /* configure_progress_cb */

static void
configure_step_cb(RCPackman *p, int pos, int tot)
{
    GtkWidget *w;
    char *text;

    rc_debug (RC_DEBUG_LEVEL_INFO, "Configuring %d of %d\n", pos, tot);
    w = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");
    gtk_progress_set_value(GTK_PROGRESS(w), (gfloat) pos);
    rc_gui_set_progress(PROGRESS_CONFIG, tot, pos, NULL);

    if (pos) {
        text = g_strdup_printf("Package %d of %d configured.", pos, tot);
        gui_log_append(appinfo->log[LOG_CONFIG], text);
        g_free(text);
    }
} /* configure_step_cb */

static void
configure_done_cb(RCPackman *p)
{
    rc_debug (RC_DEBUG_LEVEL_INFO, "Configuring done\n");
} /* configure_done_cb */
    
static void
transaction_progress_cb(RCPackman *p, gchar *filename, gint amount, gint total)
{
    gtk_progress_set_value(
        GTK_PROGRESS(GET_WIDGET("single_progress")),
        ((gfloat) amount) / ((gfloat) total) * 100.0);
} /* transaction_progress_cb */

static void
transaction_step_cb(RCPackman *p, gint pos, gint tot)
{
    GtkWidget *w;
    GUIInstallType itype;
    char *text;

    rc_debug (RC_DEBUG_LEVEL_INFO, "Transacting %d of %d\n", pos, tot);
    w = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");
    gtk_progress_set_value(GTK_PROGRESS(w), (gfloat) pos);
    rc_gui_set_progress(PROGRESS_INSTALL, tot, pos, NULL);

    if (pos) {
        itype = rc_gui_get_install_type();
        text = g_strdup_printf(
            "Package %d of %d %s", pos, tot, 
            itype == GUI_INSTALL ? "installed" : "removed");
        gui_log_append(
            appinfo->log[itype == GUI_INSTALL ? LOG_INSTALL : LOG_REMOVE], 
            text);
        g_free(text);
    }
} /* transaction_step_cb */

static void
transaction_done_cb(RCPackman *p)
{
    rc_debug (RC_DEBUG_LEVEL_INFO, "Transacting done\n");
} /* transaction_done_cb */

static void
update_total_progress(GUITransfer *t, gchar *buf, int size, int sizes[])
{
    GtkWidget *w;
    char *s1, *s2, *str;

    w = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");

    s1 = rc_gui_format_size(sizes[0] + t->trans_size);
    s2 = rc_gui_format_size(sizes[1]);
    
    str = g_strdup_printf("%%p%%%% (%s of %s)", s1, s2);
    g_free(s1);
    g_free(s2);
    gtk_progress_set_format_string(GTK_PROGRESS(w), str);
    g_free(str);
    gtk_progress_set_value(
        GTK_PROGRESS(w), (gfloat) (sizes[0] + t->trans_size));
} /* update_total_progress */

static gboolean
download_files(GSList *urls, GSList *filenames, guint64 total_size)
{
    GtkWidget *single_progress, *total_progress;
    gboolean unsub;
    RCPackmanFeatures features;
    int total_count;
    int current_count = 0;
    guint64 current_size = 0;
    GSList *iter, *iter2;

    /* Reset the progress thingys */

    single_progress = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "single_progress");
    total_progress = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");

    gtk_progress_configure(GTK_PROGRESS (single_progress), 0.0, 0.0, 100.0);
    gtk_progress_configure(
        GTK_PROGRESS (total_progress), 0.0, 0.0, (gfloat) total_size);
    gtk_progress_set_format_string(GTK_PROGRESS(total_progress), "%p%%");

    /* Desensitize the OK button during download */
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_set_sensitive (GET_WIDGET("ok_button"), FALSE);

    /* Show/hide the appropriate buttons. This is more paranoid than it needs
       to be, but better safe than sorry. */
    gtk_widget_show(GET_WIDGET("pause_button"));
    gtk_widget_hide(GET_WIDGET("resume_button"));
    gtk_widget_show(GET_WIDGET("cancel_button"));

    features = rc_packman_get_features(appinfo->pman);

    total_count = g_slist_length(filenames);

    rc_gui_set_progress_inverted(features & RC_PACKMAN_FEATURE_POST_CONFIG);
    rc_gui_set_progress(PROGRESS_DOWNLOAD, total_count, 0, NULL);

    unsub = GTK_WIDGET_IS_SENSITIVE(GET_WIDGET("unsubscribe1"));
    rc_gui_set_menubar_sensitive (FALSE);

    rc_gui_set_page ("main_notebook", PAGE_DOWNLOAD);

    iter = urls;
    iter2 = filenames;
    while (iter) {
        gchar *url = (gchar *) iter->data;
        gchar *fn = (gchar *) iter2->data;
        int fd;
        GUITransfer *t;
        GUITransferError terr;
        GByteArray *data = NULL;
        int sizes[2];
        char *s1, *s2, *str;
        char *tmp;

        gtk_progress_set_value(GTK_PROGRESS(single_progress), 0.0);
        gtk_progress_set_format_string(GTK_PROGRESS(single_progress), "%p%%");
        tmp = g_strdup_printf("Downloading %s", strrchr(fn, '/') + 1);
        rc_gui_set_progress(
            PROGRESS_DOWNLOAD, total_count, current_count, tmp);
        g_free(tmp);

        t = gui_transfer_new(TRUE, TRUE);
        gui_transfer_set_progress_bar(t, GTK_PROGRESS(single_progress));
        sizes[0] = current_size;
        sizes[1] = total_size;
        gtk_signal_connect(
            GTK_OBJECT(t), "file_data",
            GTK_SIGNAL_FUNC(update_total_progress), sizes);

        if (t->aborted) {
            rc_gui_set_page("main_notebook", PAGE_CHANNELS);
            goto download_cleanup;
        }
        
        data = gui_transfer(t, url);

        if (t->aborted) {
            rc_gui_set_page("main_notebook", PAGE_CHANNELS);
            goto download_cleanup;
        }

        terr = gui_transfer_get_error(t);
        if (terr) {
            if (data)
                g_byte_array_free(data, FALSE);
            gtk_object_destroy(GTK_OBJECT(t));

            if (terr == GUI_TRANSFER_ERROR_RETRY) {
                /* Don't iterate. Just do this one over again. */
                continue;
            }
            else if (terr == GUI_TRANSFER_ERROR_ABORT) {
                rc_gui_set_page("main_notebook", PAGE_CHANNELS);
                goto download_cleanup;
            }
        }

        if (t->timer == NULL) {
            /* Cache hit */
            str = g_strdup_printf("%s pulled from cache", url);
        }
        else {
            s1 = rc_gui_format_size((int)
                ((float) t->file_size / g_timer_elapsed(t->timer, NULL)));
            str = g_strdup_printf(
                "%s downloaded %d bytes in %d seconds (%s/s)",
                url, t->file_size, (int) g_timer_elapsed(t->timer, NULL), s1);
            g_free(s1);
        }
        gui_log_append(appinfo->log[LOG_DOWNLOAD], str);
        g_free(str);

        current_count++;
        current_size += t->file_size;
        /* This is mainly for cache hits */
        s1 = rc_gui_format_size(current_size);
        s2 = rc_gui_format_size(total_size);
        str = g_strdup_printf("%%p%%%% (%s of %s)", s1, s2);
        g_free(s1);
        g_free(s2);
        gtk_progress_set_format_string(
            GTK_PROGRESS(total_progress), str);
        gtk_progress_set_value(
            GTK_PROGRESS(total_progress), (gfloat) current_size);
        g_free(str);
 
        gtk_object_destroy(GTK_OBJECT(t));

        fd = open (fn, O_WRONLY | O_CREAT | O_TRUNC, 0644);
        rc_write (fd, data->data, data->len);
        close (fd);

        g_byte_array_free(data, TRUE);

        iter = iter->next;
        iter2 = iter2->next;
    }

    rc_gui_set_progress(PROGRESS_DOWNLOAD, total_count, total_count, NULL);

    return TRUE;

download_cleanup:
    rc_gui_set_menubar_sensitive (TRUE);
    if (unsub)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), TRUE);
    return FALSE;
} /* download_files */

static gboolean
verify_packages(RCPackageSList *packages)
{
    GtkWidget *single, *total;
    int total_count;
    int current_count = 0;
    GSList *i;

    gtk_widget_hide(GET_WIDGET("cancel_button"));
    gtk_widget_hide(GET_WIDGET("pause_button"));
    gtk_widget_hide(GET_WIDGET("resume_button"));

    total_count = g_slist_length(packages);

    single = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "single_progress");
    total = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");

    gtk_progress_configure(GTK_PROGRESS(single), 0.0, 0.0, 100.0);
    gtk_progress_configure(
        GTK_PROGRESS(total), 0.0, 0.0, (gfloat) total_count);

    gtk_progress_set_format_string(GTK_PROGRESS(single), "%p%%");
    gtk_progress_set_format_string(GTK_PROGRESS(total), "%p%% (%v of %u)");

    i = packages;
    while (i) {
        RCPackage *package = i->data;
        RCVerificationSList *vers;
        RCVerificationStatus worst_status = RC_VERIFICATION_STATUS_PASS;
        gboolean gpg = FALSE;
        char *gpg_info = NULL;
        GSList *v;
        char *tmp;

        tmp = g_strdup_printf(
            "Verifying %s", rc_gui_get_pretty_name(package->spec.name));
        rc_gui_set_progress(PROGRESS_VERIFY, total_count, current_count, tmp);
        g_free(tmp);
        while (gtk_events_pending())
            gtk_main_iteration();

        vers = rc_packman_verify(appinfo->pman, package);
        v = vers;
        while (v) {
            RCVerification *ver = v->data;

            if (worst_status > ver->status)
                worst_status = ver->status;

            if (ver->type == RC_VERIFICATION_TYPE_GPG) {
                gpg = TRUE;
                gpg_info = ver->info;
            }

            v = v->next;
        }

        if (worst_status == RC_VERIFICATION_STATUS_FAIL) {
            char *text;

            text = g_strdup_printf(
                "Cryptographic verification of %s (%s) has FAILED.",
                rc_gui_get_pretty_name(package->spec.name), 
                package->spec.name);
            gui_log_append(appinfo->log[LOG_VERIFY], text);
            g_free(text);
            
            text = g_strdup_printf(
                "Red Carpet's cryptographic verification of %s (%s) has "
                "<b><font color=\"#ff0000\">failed</font></b>. This means "
                "that the security of this package has somehow been "
                "compromised and Red Carpet will not install it. The "
                "downloaded package may have been corrupted and downloading "
                "the package again may fix the problem. If it doesn't, "
                "however, you should contact "
                "<a href=\"mailto:distribution@ximian.com\">"
                "distribution@ximian.com</a> immediately.",
                rc_gui_get_pretty_name(package->spec.name), 
                package->spec.name);
            rc_gui_error_box(text, RC_ERROR_ABORT);
            g_free(text);
            return FALSE;
        }
        else if (worst_status == RC_VERIFICATION_STATUS_UNDEF) {
            gboolean ask = gui_prefs_get_signature_ask();

            if (ask) {
                char *text;
                GnomeDialog *dialog;
                int res;

                text = g_strdup_printf(
#if 0
                    "CAUTION: Red Carpet is unable to ensure the "
                    "cryptographic integrity of this package. The\n"
                    "package might not be signed, or Red Carpet may not have "
                    "the key needed to verify this\n"
                    "package. You may continue and install this package, but "
                    "Red Carpet cannot ensure that\n"
                    "it is safe or from a trusted source.\n\n"
#endif
                    "INFO:\n"
                    "Name: %s (%s)\n"
                    "Version: %s\n"
                    "Release: %s\n"
                    "Package Filename:\n"
                    "        %s\n",
                    rc_gui_get_pretty_name(package->spec.name), 
                    package->spec.name,
                    package->spec.version, 
                    package->spec.release,
                    package->package_filename);
#if 0
                gtk_label_set_text(
                    GTK_LABEL(GET_WIDGET("caution_label")), text);
#else
                gtk_label_set_text(
                    GTK_LABEL(GET_WIDGET("caution_info")), text);
#endif
                g_free(text);
                dialog = GNOME_DIALOG(GET_WIDGET("signature_dialog"));
                gnome_dialog_close_hides(dialog, TRUE);
                res = gnome_dialog_run_and_close(dialog);

                text = g_strdup_printf(
                    "Verification of %s (%s) is inconclusive; %s",
                    rc_gui_get_pretty_name(package->spec.name), 
                    package->spec.name, res != 0 ? "aborting" : "continuing");
                gui_log_append(appinfo->log[LOG_VERIFY], text);
                g_free(text);

                if (res != 0) /* OK Button */
                    return FALSE;
            }
            else {
                char *text;

                text = g_strdup_printf(
                    "Verification of %s (%s) is inconclusive; continuing",
                    rc_gui_get_pretty_name(
                        package->spec.name), package->spec.name);
                gui_log_append(appinfo->log[LOG_VERIFY], text);
                g_free(text);
            }
        }
        else { /* Passed */
            char *text;

            if (gpg) {
                text = g_strdup_printf(
                    "%s (%s) has been verified as cryptographically secure "
                    "by %s",
                    rc_gui_get_pretty_name(package->spec.name), 
                    package->spec.name, gpg_info);
            }
            else {
                text = g_strdup_printf(
                    "The integrity of %s (%s) has been confirmed.",
                    rc_gui_get_pretty_name(package->spec.name), 
                    package->spec.name);
            }
            gui_log_append(appinfo->log[LOG_VERIFY], text);
            g_free(text);
        }

        current_count++;
        i = i->next;
    }

    rc_gui_set_progress(PROGRESS_VERIFY, total_count, total_count, NULL);

    return TRUE;
} /* verify_packages */    

static gboolean
install_packages(RCPackageSList *add_packages, RCPackageSList *remove_packages)
{
    GtkWidget *single, *total;
    int package_count;
    RCPackmanError err;

    gtk_widget_hide(GET_WIDGET("cancel_button"));
    gtk_widget_hide(GET_WIDGET("pause_button"));
    gtk_widget_hide(GET_WIDGET("resume_button"));

    package_count = g_slist_length(add_packages);
    
    single = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "single_progress");
    total = gtk_object_get_data(
        GTK_OBJECT(appinfo->download_html), "total_progress");
   
    gtk_progress_configure(GTK_PROGRESS(single), 0.0, 0.0, 100.0);
    gtk_progress_configure(
        GTK_PROGRESS(total), 0.0, 0.0, (gfloat) package_count);

    gtk_progress_set_format_string(GTK_PROGRESS(single), "%P%%");
    gtk_progress_set_format_string(GTK_PROGRESS(total), "%P%% (%V of %U)");

    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "configure_progress",
        GTK_SIGNAL_FUNC(configure_progress_cb), NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "configure_step",
        GTK_SIGNAL_FUNC(configure_step_cb), NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "configure_done",
        GTK_SIGNAL_FUNC(configure_done_cb), NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transaction_progress",
        GTK_SIGNAL_FUNC(transaction_progress_cb), NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transaction_step",
        GTK_SIGNAL_FUNC(transaction_step_cb), NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transaction_done",
        GTK_SIGNAL_FUNC(transaction_done_cb), NULL);

    rc_packman_transact(appinfo->pman, add_packages, remove_packages);
    err = rc_packman_get_error(appinfo->pman);
    if (err) {
        char *text;
        
        if (err == RC_PACKMAN_ERROR_FATAL) {
            text = g_strdup_printf(
                "<p>A fatal error has occurred while Red Carpet was "
                "processing this transaction. This is likely caused by an "
                "improperly built package. You should report this problem to "
                "<a href=\"mailto:distribution@ximian.com\">"
                "distribution@ximian.com</a> immediately.</p>"
                "<p>Information from the packaging system:<br>%s</p>",
                rc_packman_get_reason(appinfo->pman));
            rc_gui_error_box(text, RC_ERROR_FATAL);
            g_free(text);
            gtk_exit(1);
        }
        else {
            text = g_strdup_printf(
                "<p>An error has occurred while Red Carpet was processing "
                "this transaction. This may be caused by an improperly built "
                "or corrupted package. You may want to try the transaction "
                "again with fewer packages. You should report this problem to "
                "<a href=\"mailto:distribution@ximian.com\">"
                "distribution@ximian.com</a> as soon as possible.</p>"
                "<p>Information from the packaging system:<br>%s</p>",
                rc_packman_get_reason(appinfo->pman));
            rc_gui_error_box(text, RC_ERROR_ABORT);
            g_free(text);
            return FALSE;
        }
    }

    return TRUE;
} /* install_packages */

static void
rc_gui_channel_update_all(GUIChannel *current)
{
    GSList *i;
	
    i = appinfo->channel_list;
    while (i) {
        GUIChannel *c = i->data;
        
        if (c->subscribed)
            rc_gui_channel_update_package_lists(c);

        if (!current || current == c) {
            RCPackageSList **packages;

            packages = gtk_object_get_data(
                GTK_OBJECT(appinfo->update_html), "packages_to_install");
            if (packages) {
                g_slist_free(*packages);
                *packages = NULL;
            }

            packages = gtk_object_get_data(
                GTK_OBJECT(appinfo->available_html), "packages_to_install");
            if (packages) {
                g_slist_free(*packages);
                *packages = NULL;
            }

            packages = gtk_object_get_data(
                GTK_OBJECT(appinfo->installed_html), "packages_to_remove");
            if (packages) {
                g_slist_free(*packages);
                *packages = NULL;
            }
        }
        
        i = i->next;
    }

    rc_gui_channel_select(current);
    if (!current)
        rc_gui_show_executive_summary();
} /* rc_gui_channel_update_all */

static void
ok_clicked_cb (GtkWidget *button)
{
    rc_debug (RC_DEBUG_LEVEL_INFO, "### OK CLICKED\n");
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(GET_WIDGET("ok_button")), ok_clicked_cb, NULL);
    rc_gui_set_page("main_notebook", PAGE_CHANNELS);
} /* ok_clicked_cb */

void
rc_gui_verify_installed_packages (void)
{
    RCChannelSList *sublist = NULL;
    RCChannelSList *unsublist = NULL;
    GHashTable     *added_hash = NULL;
    RCPackageSList *added_pkgs = NULL;
    RCPackageSList *removed_pkgs = NULL;
    RCPackageSList *packages = NULL;
    GSList         *urls = NULL;
    GSList         *filenames = NULL;
    GSList         *iter;
    guint64        total_size = 0;
    int            count;
    RCDepsFixInfo  *fixinfo;

    /* Split the list of channels into subscribed and unsubscribed */
    rc_gui_split_channel_list (appinfo->channel_list, &sublist, &unsublist);

    fixinfo = rc_deps_fix_system (appinfo->system_packages,
                                  sublist,
                                  unsublist);

    added_pkgs = fixinfo->resolution_result->packages_to_install;
    removed_pkgs = fixinfo->resolution_result->packages_to_remove;

    /* Present it all to the user in HTML-ish magic. */
    
    if (!(dependency_confirm (TRUE, NULL, NULL, sublist, unsublist,
                              added_pkgs, removed_pkgs)))
    {
        /* The user didn't agree */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return;
    }

    /* Run through the list of packages and build a list of URLs from them */
    iter = added_pkgs;
    while (iter) {
        RCPackage *pkg = (RCPackage *)(iter->data);
        RCPackageUpdate *upd = rc_package_get_latest_update(pkg);
        char *url;
        char *fn;

        total_size += upd->package_size;
        packages = g_slist_append(packages, pkg);
        url = maybe_mirror_channel_url(
            upd->package_url, RCCHANNEL(pkg->subchannel->channel));
        fn = make_filename(url);
        if (!fn) {
            /* Abort the operation */
            rc_gui_set_page("main_notebook", PAGE_CHANNELS);
            goto cleanup;
        }
        filenames = g_slist_append(filenames, fn);
        pkg->package_filename = fn;
        urls = g_slist_append(urls, url);

        if (upd->signature_url) {
            total_size += upd->signature_size;
            url = maybe_mirror_channel_url(
                upd->signature_url, RCCHANNEL(pkg->subchannel->channel));
            fn = make_filename(url);
            if (!fn) {
                /* Abort the operation */
                rc_gui_set_page("main_notebook", PAGE_CHANNELS);
                goto cleanup;
            }
            filenames = g_slist_append(filenames, fn);
            pkg->signature_filename = fn;
            urls = g_slist_append(urls, url);
        }

        iter = iter->next;
    }

    for (iter = urls; iter; iter = iter->next) {
        rc_debug (RC_DEBUG_LEVEL_INFO, "URL: %s\n", (char *) iter->data);
    }
    
    rc_gui_prepare_install_page(GUI_INSTALL);
    rc_gui_log_display_reset_all();

    gui_cache_override_cache_name_func(make_filename);
    /* First do the download */
    if (!download_files(urls, filenames, total_size)) {
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        goto cleanup;
    }
    gui_cache_override_cache_name_func(NULL);

    /* Then verify the packages */
    if (!verify_packages(packages)) {
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        goto cleanup;
    }

    /* Install and configure */
    if (!install_packages(packages, removed_pkgs)) {
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        goto install_cleanup;
    }

    count = g_slist_length(packages);
    rc_gui_set_progress(PROGRESS_WAIT_FOR_IT, count, count, NULL);

    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("ok_button")), "clicked",
        GTK_SIGNAL_FUNC(ok_clicked_cb), NULL);

    gtk_widget_set_sensitive(GET_WIDGET("ok_button"), TRUE);
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_grab_default(GET_WIDGET("ok_button"));

    rc_gui_query_system_packages ();

    rc_gui_channel_update_all(NULL);

    rc_gui_set_progress(PROGRESS_FINISHED, count, count, NULL);

install_cleanup:
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), configure_progress_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), configure_step_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), configure_done_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transaction_progress_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transaction_step_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transaction_done_cb, NULL);

cleanup:
    g_slist_free (filenames);

    g_slist_free (urls);

    g_hash_table_destroy (added_hash);
    g_slist_free (removed_pkgs);

    g_slist_free (sublist);
    g_slist_free (unsublist);

    rc_gui_set_menubar_sensitive (TRUE);
}

void
gui_install_packages (GUIChannel *current_channel, RCPackageSList *pkgs)
{
    RCChannelSList *sublist = NULL;
    RCChannelSList *unsublist = NULL;
    GHashTable     *added_hash = NULL;
    RCPackageSList *added_pkgs = NULL;
    RCPackageSList *removed_pkgs = NULL;
    RCPackageSList *packages = NULL;
    GSList         *urls = NULL;
    GSList         *filenames = NULL;
    GSList         *iter;
    guint64        total_size = 0;
    int            count;

//    if (!pkgs) {
//        return;
//    }

    /* Split the list of channels into subscribed and unsubscribed */
    rc_gui_split_channel_list (appinfo->channel_list, &sublist, &unsublist);
    
    /* Run it through the dependency system */
    if (!(install_dependency_check(
              pkgs, current_channel, sublist, unsublist, &added_hash,
              &added_pkgs, &removed_pkgs))) {
        /* This is one of the error cases, dependency_helper should have
           already displayed the error to the user, so we're free to quit at
           this point.  Yay. */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return;
    }

    /* Present it all to the user in HTML-ish magic. */
    
    if (!(dependency_confirm(
              TRUE, pkgs, current_channel, sublist, unsublist,
              added_pkgs, removed_pkgs))) {
        /* The user didn't agree */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return;
    }

    /* Run through the list of packages and build a list of URLs from them */
    iter = added_pkgs;
    while (iter) {
        RCPackage *pkg = (RCPackage *)(iter->data);
        RCPackageUpdate *upd = rc_package_get_latest_update(pkg);
        char *url;
        char *fn;

        total_size += upd->package_size;
        packages = g_slist_append(packages, pkg);
        url = maybe_mirror_channel_url(
            upd->package_url, RCCHANNEL(pkg->subchannel->channel));
        fn = make_filename(url);
        if (!fn) {
            /* Abort the operation */
            rc_gui_set_page("main_notebook", PAGE_CHANNELS);
            goto cleanup;
        }
        filenames = g_slist_append(filenames, fn);
        pkg->package_filename = fn;
        urls = g_slist_append(urls, url);

        if (upd->signature_url) {
            total_size += upd->signature_size;
            url = maybe_mirror_channel_url(
                upd->signature_url, RCCHANNEL(pkg->subchannel->channel));
            fn = make_filename(url);
            if (!fn) {
                /* Abort the operation */
                rc_gui_set_page("main_notebook", PAGE_CHANNELS);
                goto cleanup;
            }
            filenames = g_slist_append(filenames, fn);
            pkg->signature_filename = fn;
            urls = g_slist_append(urls, url);
        }

        iter = iter->next;
    }

    iter = pkgs;
    while (iter) {
        RCPackage *pkg = (RCPackage *)(iter->data);
        RCPackageUpdate *upd = rc_package_get_latest_update(pkg);
        char *url;
        char *fn;

        total_size += upd->package_size;
        packages = g_slist_append(packages, pkg);

        url = maybe_mirror_channel_url(
            upd->package_url, RCCHANNEL(pkg->subchannel->channel));
        fn = make_filename(url);
        if (!fn) {
            /* Abort the operation */
            rc_gui_set_page("main_notebook", PAGE_CHANNELS);
            goto cleanup;
        }
        filenames = g_slist_append(filenames, fn);
        pkg->package_filename = fn;
        urls = g_slist_append(urls, url);

        if (upd->signature_url) {
            total_size += upd->signature_size;
            url = maybe_mirror_channel_url(
                upd->signature_url, RCCHANNEL(pkg->subchannel->channel));
            fn = make_filename(url);
            if (!fn) {
                /* Abort the operation */
                rc_gui_set_page("main_notebook", PAGE_CHANNELS);
                goto cleanup;
            }
            filenames = g_slist_append(filenames, fn);
            pkg->signature_filename = fn;
            urls = g_slist_append(urls, url);
        }

        iter = iter->next;
    }
    
    for (iter = urls; iter; iter = iter->next) {
        rc_debug (RC_DEBUG_LEVEL_INFO, "URL: %s\n", (char *) iter->data);
    }
    
    rc_gui_prepare_install_page(GUI_INSTALL);
    rc_gui_log_display_reset_all();

    gui_cache_override_cache_name_func(make_filename);
    /* First do the download */
    if (!download_files(urls, filenames, total_size)) {
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        goto cleanup;
    }
    gui_cache_override_cache_name_func(NULL);

    /* Then verify the packages */
    if (!verify_packages(packages)) {
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        goto cleanup;
    }

    /* Install and configure */
    if (!install_packages(packages, removed_pkgs)) {
        rc_gui_set_page("main_notebook", PAGE_CHANNELS);
        goto install_cleanup;
    }

    count = g_slist_length(packages);
    rc_gui_set_progress(PROGRESS_WAIT_FOR_IT, count, count, NULL);

    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("ok_button")), "clicked",
        GTK_SIGNAL_FUNC(ok_clicked_cb), NULL);

    gtk_widget_set_sensitive(GET_WIDGET("ok_button"), TRUE);
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_grab_default(GET_WIDGET("ok_button"));

    rc_gui_query_system_packages ();

    rc_gui_channel_update_all(current_channel);

    rc_gui_set_progress(PROGRESS_FINISHED, count, count, NULL);

install_cleanup:
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), configure_progress_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), configure_step_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), configure_done_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transaction_progress_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transaction_step_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transaction_done_cb, NULL);

cleanup:
    g_slist_free (filenames);

    g_slist_free (urls);

    g_hash_table_destroy (added_hash);
    g_slist_free (removed_pkgs);

    g_slist_free (sublist);
    g_slist_free (unsublist);

    rc_gui_set_menubar_sensitive (TRUE);
} /* gui_install_packages */

static gboolean
remove_dependency_check (RCPackageSList *requested_pkgs,
                         GUIChannel     *current_channel,
                         RCChannelSList *subscribed_channels,
                         RCChannelSList *unsubscribed_channels,
                         RCPackageSList **removed_pkgs)
{
    RCDepResolution *rcdr;
    RCDepResolutionResult *rr;
    RCPackageSList *piter;

    rcdr = rc_dep_resolution_new (appinfo->system_packages,
                                  RCCHANNEL (current_channel),
                                  subscribed_channels,
                                  unsubscribed_channels,
                                  current_channel->subchannel_preference);

    for (piter = requested_pkgs; piter; piter = piter->next) {
        RCPackage *pkg = (RCPackage *) piter->data;

        rc_debug(
            RC_DEBUG_LEVEL_DEBUG, "--> Uninstalling %s\n", pkg->spec.name);

        rc_dep_resolution_remove_package (rcdr, pkg);
    }

    rr = rc_dep_resolution_get_result (rcdr);
    debug_rc_dep_print_results (RC_DEBUG_LEVEL_INFO, rr);

    if (rr->abort) {
        gchar *text = g_strdup_printf(
            "<p>An error has occurred while Red Carpet was processing the "
            "dependencies for this transaction. This may be caused by an "
            "error in the package or on the server. You may want to try the "
            "transaction again with fewer packages. You should report this "
            "problem to <a href=\"distribution@ximian.com\">"
            "distributon@ximian.com</a> as soon as possible.</p>"
            "<p>Information from the dependency system:<br> Recursion too "
            "deep.</p>");
        rc_gui_error_box(text, RC_ERROR_ABORT);
        g_free(text);
        rc_dep_resolution_free (rcdr);
        return FALSE;
    }

    for (piter = rr->packages_to_remove; piter; piter = piter->next) {
        RCPackage *pkg = (RCPackage *) piter->data;

        if (g_slist_find_custom (requested_pkgs, pkg,
                                 rc_package_spec_not_equal)) {
            continue;
        }

        *removed_pkgs = g_slist_append (*removed_pkgs, pkg);
    }

    rc_dep_resolution_free (rcdr);

    return (TRUE);
} /* remove_dependency_check */

void
gui_remove_packages (GUIChannel *channel, RCPackageSList *pkgs)
{
    RCChannelSList *subscribed_channels = NULL;
    RCChannelSList *unsubscribed_channels = NULL;
    RCPackageSList *removed_pkgs = NULL;
    RCPackmanFeatures features;
    gboolean unsub;
    GtkWidget *single_progress;
    GtkWidget *total_progress;
    int count;
    RCPackmanError err;

    rc_gui_split_channel_list(
        appinfo->channel_list, &subscribed_channels, &unsubscribed_channels);

    if (!(remove_dependency_check(
              pkgs, channel, subscribed_channels,
              unsubscribed_channels, &removed_pkgs))) {
        /* An error occured, which has already been displayed to the user in
           remove_dependency_check, so we're free to quit now.  Adios. */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return;
    }

    if (!(dependency_confirm(
              FALSE, pkgs, channel, NULL, NULL, NULL, removed_pkgs))) {
        /* The user changed his/her mind.  PC points! */
        rc_gui_set_page ("main_notebook", PAGE_CHANNELS);
        return;
    }

    rc_gui_log_display_reset_all();

    unsub = GTK_WIDGET_IS_SENSITIVE(GET_WIDGET("unsubscribe1"));
    rc_gui_set_menubar_sensitive (FALSE);

    removed_pkgs = g_slist_concat(removed_pkgs, pkgs);

    single_progress = GET_WIDGET("single_progress");
    total_progress = GET_WIDGET("total_progress");
    count = g_slist_length(removed_pkgs);

    gtk_progress_configure(GTK_PROGRESS(single_progress), 0.0, 0.0, 100.0);
    gtk_progress_configure(GTK_PROGRESS(total_progress), 0.0, 0.0, count);
    gtk_progress_set_format_string(
        GTK_PROGRESS(GET_WIDGET("single_progress")), "%P%%");
    gtk_progress_set_format_string(
        GTK_PROGRESS(GET_WIDGET("total_progress")), "%P%% (%V of %U)");

    /* Show the OK button but desensitize it */
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_set_sensitive(GET_WIDGET("ok_button"), FALSE);

    /* Hide the Cancel, Pause, and Resume buttons, as they are N/A */
    gtk_widget_hide(GET_WIDGET("cancel_button"));
    gtk_widget_hide(GET_WIDGET("pause_button"));
    gtk_widget_hide(GET_WIDGET("resume_button"));

    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "configure_progress",
        configure_progress_cb, NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "configure_step",
        configure_step_cb, NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "configure_done",
        configure_done_cb, NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transaction_progress",
        transaction_progress_cb, NULL);
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transaction_step",
        transaction_step_cb, GINT_TO_POINTER(FALSE));
    gtk_signal_connect(
        GTK_OBJECT(appinfo->pman), "transaction_done",
        transaction_done_cb, NULL);

    rc_gui_prepare_install_page(GUI_REMOVE);
    features = rc_packman_get_features(appinfo->pman);
    rc_gui_set_progress_inverted(features & RC_PACKMAN_FEATURE_POST_CONFIG);

    rc_gui_set_page("main_notebook", PAGE_DOWNLOAD);

    rc_packman_transact(appinfo->pman, NULL, removed_pkgs);
    err = rc_packman_get_error(appinfo->pman);
    if (err) {
        char *text;
        
        if (err == RC_PACKMAN_ERROR_FATAL) {
            text = g_strdup_printf(
                "<p>A fatal error has occurred while Red Carpet was "
                "processing this transaction. This is likely caused by an "
                "improperly built package. You should report this problem to "
                "<a href=\"mailto:distribution@ximian.com\">"
                "distribution@ximian.com</a> immediately.</p>"
                "<p>Information from the packaging system:<br>%s</p>",
                rc_packman_get_reason(appinfo->pman));
            rc_gui_error_box(text, RC_ERROR_FATAL);
            g_free(text);
            gtk_exit(1);
        }
        else {
            text = g_strdup_printf(
                "<p>An error has occurred while Red Carpet was processing "
                "this transaction. This may be caused by an improperly built "
                "or corrupted package. You may want to try the transaction "
                "again with fewer packages. You should report this problem to "
                "<a href=\"mailto:distribution@ximian.com\">"
                "distribution@ximian.com</a> as soon as possible.</p>"
                "<p>Information from the packaging system:<br>%s</p>",
                rc_packman_get_reason(appinfo->pman));
            rc_gui_error_box(text, RC_ERROR_ABORT);
            g_free(text);
            goto remove_cleanup;
        }
    }

    rc_gui_set_progress(PROGRESS_WAIT_FOR_IT, count, count, NULL);

    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("ok_button")), "clicked",
        GTK_SIGNAL_FUNC(ok_clicked_cb), NULL);

    gtk_widget_set_sensitive(GET_WIDGET("ok_button"), TRUE);
    gtk_widget_show(GET_WIDGET("ok_button"));
    gtk_widget_grab_default(GET_WIDGET("ok_button"));
    
    rc_gui_query_system_packages ();

    rc_gui_channel_update_all(channel);

    rc_gui_set_progress(PROGRESS_FINISHED, count, count, NULL);

remove_cleanup:
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), configure_progress_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), configure_step_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), configure_done_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transaction_progress_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transaction_step_cb, NULL);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(appinfo->pman), transaction_done_cb, NULL);

    rc_gui_set_menubar_sensitive (TRUE);
    if (unsub)
        gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), TRUE);
} /* gui_remove_packages */
