/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-prefs.c: proxy / mirror settings
 *
 * Copyright 2000, 2001 Ximian, Inc.
  *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gnome.h>
#include "gui.h"
#include "gui-prefs.h"
#include "gui-transfer.h"
#include "mirror.h"

typedef struct {
    char *widget_name;
    char *conf_path;
    enum {
        CONF_TOGGLE,
        CONF_ENTRY,
        CONF_LIST
    } type;
    void *p;
    gboolean priv : 1;
    GtkWidget *w;
} ConfVal;

typedef struct {
    GtkWidget *dialog;
    gchar *proxy_host;
    gchar *proxy_port;
    gchar *proxy_name;
    gchar *proxy_pass;
    int initializing;
    gboolean proxy_enabled;
    gboolean proxy_auth;
} ProxyConfig;

typedef struct {
    gboolean cache_enabled;
    gchar *cache_path;
    gchar *cache_difference;
    gboolean cache_clear_packages_on_startup;
    gboolean cache_clear_packages_on_exit;
} CacheConfig;

typedef struct {
    gchar *url;
    int selected_row;
} MirrorConfig;

typedef struct {
    gboolean sign_ask_enabled;
} SignatureConfig;

typedef struct {
    gboolean show_pretty_names;
} GUIConfig;

ProxyConfig     proxy_config  = { 0 };
CacheConfig     cache_config  = { 0 };
MirrorConfig    mirror_config = { 0 };
SignatureConfig signature_config = { 0 };
GUIConfig       gui_config = { 0 };

static ConfVal conf[] = {
    { "proxy_enable", "Proxy/Enabled",      CONF_TOGGLE, 
      &proxy_config.proxy_enabled },
    { "proxy_host",   "Proxy/Host",         CONF_ENTRY,  
      &proxy_config.proxy_host },
    { "proxy_port",   "Proxy/Port",         CONF_ENTRY, 
      &proxy_config.proxy_port },
    { "proxy_auth",   "Proxy/UseAuth",      CONF_TOGGLE,  
      &proxy_config.proxy_auth },
    { "proxy_name",   "Proxy/Name",         CONF_ENTRY, 
      &proxy_config.proxy_name },
    { "proxy_pass",   "Proxy/Pass",         CONF_ENTRY, 
      &proxy_config.proxy_pass, 1 },
    { "cache_enable", "Cache/Enabled=TRUE", CONF_TOGGLE, 
      &cache_config.cache_enabled },
    { "cache_path",   NULL /* set in on_preferences1_activate */, CONF_ENTRY, 
      &cache_config.cache_path },
    { "cache_difference", "Cache/Difference=3600", CONF_ENTRY, 
      &cache_config.cache_difference },
    { "cache_clear_packages_on_startup", "Cache/ClearOnStartup=TRUE", CONF_TOGGLE,
      &cache_config.cache_clear_packages_on_startup },
    { "cache_clear_packages_on_exit", "Cache/ClearOnExit", CONF_TOGGLE,
      &cache_config.cache_clear_packages_on_exit },
    { "mirror_clist", "Mirror/URL",    CONF_LIST, 
      &mirror_config.url },
    { "sign_ask_enable", "Signatures/Ask=TRUE", CONF_TOGGLE,
      &signature_config.sign_ask_enabled },
    { "gui_show_pretty_names", "GUI/PrettyNames=TRUE", CONF_TOGGLE,
      &gui_config.show_pretty_names },
    { NULL }
};

/* some libglade callbacks */
void on_proxy_enable_toggled (GtkWidget *toggle, gpointer data);
void on_proxy_auth_toggled (GtkWidget *toggle, gpointer data);
void on_proxy_entry_changed (GtkWidget *w, GdkEvent *evt, gpointer data);
void on_cache_enable_toggled(GtkWidget *toggle, gpointer data);
void on_cache_path_changed(GtkWidget *w, GdkEvent *evt, gpointer data);
void on_cache_difference_changed(GtkWidget *w, GdkEvent *evt, gpointer data);
void on_cache_clear_packages_on_startup_toggled (GtkWidget *toggle, gpointer data);
void on_cache_clear_packages_on_exit_toggled (GtkWidget *toggle, gpointer data);
void on_sign_ask_enable_toggled(GtkWidget *toggle, gpointer data);
void on_gui_show_pretty_names_toggled(GtkWidget *toggle, gpointer data);
void on_property_dialog_apply (GtkWidget *dialog, gpointer data);
void on_clist_selected(GtkWidget *clist, int row, int column, GdkEvent *event);
void on_clist_unselected(GtkWidget *clist, int row, int col, GdkEvent *event);
void refresh_button_clicked(GtkWidget *button);
gboolean on_property_dialog_close (GtkWidget *dialog, gpointer data);

static void
apply_proxy_settings (void)
{
    gchar *uri;
    GSList *transfers;
    GSList *i;

    transfers = gui_transfer_get_current_transfers();
    i = transfers;
    while (i) {
        GUITransfer *t = i->data;

        if (!proxy_config.proxy_enabled) {
            gui_transfer_set_proxy_info(t, NULL, NULL, NULL);
            return;
        }

        uri = g_strdup_printf ("http://%s:%s",
                               proxy_config.proxy_host,
                               proxy_config.proxy_port);

        rc_debug (RC_DEBUG_LEVEL_INFO, "Setting proxy to: %s", uri);

        if (proxy_config.proxy_auth) {
            gui_transfer_set_proxy_info(
                t, uri, proxy_config.proxy_name, proxy_config.proxy_pass);
        } else {
            gui_transfer_set_proxy_info(t, uri, NULL, NULL);
        }
        g_free(uri);

        i = i->next;
    }
} /* apply_proxy_settings */

void
gui_prefs_load ()
{
    ConfVal *val;

    if (!proxy_config.dialog)
        proxy_config.dialog = GET_WIDGET ("property_dialog");

    gnome_config_push_prefix ("/red-carpet/");

    proxy_config.initializing = TRUE;

    /* Set the default cache path depending on whether or not this is being
       run as root. */
    conf[7].conf_path = appinfo->not_root ? 
        g_strdup_printf("Cache/Path=%s/.redcarpet", getenv("HOME")) : 
        g_strdup("Cache/Path=/var/cache/redcarpet");

    for (val = conf; val->widget_name; val++) {
        val->w = GET_WIDGET (val->widget_name);
        if (val->type == CONF_TOGGLE) {
            *(gboolean *)val->p = val->priv
                ? gnome_config_private_get_bool (val->conf_path)
                : gnome_config_get_bool (val->conf_path);
            gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (val->w),
                                          *(gboolean *)val->p);
        } else if (val->type == CONF_ENTRY) {
            g_free (*(gchar **)val->p);
            *(gchar **)val->p = val->priv
                ? gnome_config_private_get_string (val->conf_path)
                : gnome_config_get_string (val->conf_path);
            if (*(gchar **)val->p)
                gtk_entry_set_text (GTK_ENTRY (val->w),
                                    *(gchar **)val->p);
        } else if (val->type == CONF_LIST) {
            int i;

            g_free (*(gchar **)val->p);
            *(gchar **)val->p = val->priv
                ? gnome_config_private_get_string(val->conf_path)
                : gnome_config_get_string(val->conf_path);

            if (*(gchar **)val->p) {
                /* First off, if the string is empty (""), then we would prefer
                   if it were NULL. */
                if (!(**(gchar **)val->p))
                    *(gchar **)val->p = NULL;

                for (i = 0; i < GTK_CLIST(val->w)->rows; i++) {
                    char *url = gtk_clist_get_row_data(GTK_CLIST(val->w), i);

                    /* url and val->p are NULL. That's a match */
                    if (!url && !(*(gchar **)val->p))
                        gtk_clist_select_row(GTK_CLIST(val->w), i, 0);
                    else if (url && *(gchar **)val->p) {
                        if (url && g_strcasecmp(*(gchar **)val->p, url) == 0)
                            gtk_clist_select_row(GTK_CLIST(val->w), i, 0);
                    }
                }
            }
        }
    }

    proxy_config.initializing = FALSE;

    gnome_config_pop_prefix ();

    apply_proxy_settings ();

    g_free(conf[7].conf_path);
}

gboolean
gui_prefs_get_proxy_info(char **url, char **name, char **pass)
{
    if (!proxy_config.proxy_enabled) {
        if (url)
            *url = NULL;
        if (name)
            *name = NULL;
        if (pass)
            *pass = NULL;

        return FALSE;
    }

    if (url) 
        *url = g_strdup_printf(
            "http://%s:%s", proxy_config.proxy_host, proxy_config.proxy_port);

    if (name) {
        *name = proxy_config.proxy_auth ? 
            g_strdup(proxy_config.proxy_name) : NULL;
    }

    if (pass) {
        *pass = proxy_config.proxy_auth ? 
            g_strdup(proxy_config.proxy_pass) : NULL;
    }

    return TRUE;
} /* gui_prefs_get_proxy_info */

gboolean
gui_prefs_get_cache_info(char **path, int *difference, gboolean *clear_on_startup, gboolean *clear_on_exit)
{
    if (!cache_config.cache_enabled) {
        if (path)
            *path = NULL;
        if (difference)
            *difference = 0;
        if (clear_on_startup)
            *clear_on_startup = FALSE;
        if (clear_on_exit)
            *clear_on_exit = FALSE;

        return FALSE;
    }

    if (path)
        *path = g_strdup(cache_config.cache_path);

    if (difference)
        *difference = atoi(cache_config.cache_difference);

    if (clear_on_startup)
        *clear_on_startup = cache_config.cache_clear_packages_on_startup;

    if (clear_on_exit)
        *clear_on_exit = cache_config.cache_clear_packages_on_exit;

    return TRUE;
} /* gui_prefs_get_cache_info */

const gchar *
gui_prefs_get_mirror_url(void)
{
    return mirror_config.url;
} /* gui_prefs_get_mirror_url */

gboolean
gui_prefs_get_signature_ask(void)
{
    return signature_config.sign_ask_enabled;
} /* gui_prefs_get_signature_ask */

void
gui_prefs_set_signature_ask(gboolean ask)
{
    GtkWidget *button;

    signature_config.sign_ask_enabled = ask;
    button = GET_WIDGET("sign_ask_enable");
    GTK_TOGGLE_BUTTON(button)->active = ask;
    gtk_widget_queue_draw(button);
    
    /* This probably isn't the right way to do it, but we're gonna save this
       element and just this element now. */
    gnome_config_push_prefix("/red-carpet/");
    gnome_config_set_bool("Signatures/Ask", ask);
    gnome_config_pop_prefix();
    gnome_config_sync();
} /* gui_prefs_set_signature_ask */

gboolean
gui_prefs_get_pretty_names(void)
{
    return gui_config.show_pretty_names;
} /* gui_prefs_get_pretty_names */

static void
proxy_config_save ()
{
    ConfVal *val;

    gnome_config_push_prefix ("/red-carpet/");

    for (val = conf; val->widget_name; val++) {
        if (val->type == CONF_TOGGLE) {
            *(gboolean *)val->p = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (val->w));
            if (val->priv)
                gnome_config_private_set_bool (val->conf_path, *(gboolean *)val->p);
            else
                gnome_config_set_bool (val->conf_path, *(gboolean *)val->p);
        } else if (val->type == CONF_ENTRY) {
            g_free (*(gchar **)val->p);
            *(gchar **)val->p = gtk_editable_get_chars (GTK_EDITABLE (val->w), 0, -1);        
            if (val->priv)
                gnome_config_private_set_string (val->conf_path, *(gchar **)val->p);
            else
                gnome_config_set_string (val->conf_path, *(gchar **)val->p);
        } else if (val->type == CONF_LIST) {
            if (mirror_config.selected_row != -1) {
                g_free (*(gchar **)val->p);
                *(gchar **)val->p = g_strdup(gtk_clist_get_row_data(
                    GTK_CLIST(val->w), mirror_config.selected_row));
                if (val->priv) {
                    gnome_config_private_set_string(
                        val->conf_path, *(gchar **)val->p);
                }
                else
                    gnome_config_set_string(val->conf_path, *(gchar **)val->p);
            }
        }
    }

    gnome_config_pop_prefix ();
    gnome_config_sync ();

    apply_proxy_settings ();
}

static void
set_enabled_from_toggle (GtkWidget *toggle, gchar *widget)
{
    GtkWidget *table = GET_WIDGET (widget);
            
    gtk_widget_set_sensitive (table, GTK_TOGGLE_BUTTON (toggle)->active);

    if (!proxy_config.initializing)
        gnome_property_box_changed (GNOME_PROPERTY_BOX (proxy_config.dialog));
}    

void
on_proxy_auth_toggled (GtkWidget *toggle, gpointer data)
{
    set_enabled_from_toggle (toggle, "auth_table");
}

void
on_proxy_enable_toggled (GtkWidget *toggle, gpointer data)
{
    set_enabled_from_toggle (toggle, "proxy_vbox");
}

void
on_proxy_entry_changed (GtkWidget *w, GdkEvent *evt, gpointer data)
{
    if (!proxy_config.initializing)
        gnome_property_box_changed (GNOME_PROPERTY_BOX (proxy_config.dialog));
}

void
on_sign_ask_enable_toggled(GtkWidget *toggle, gpointer data)
{
    /* Don't call set_enabled_from_toggle because we don't want to
       desensitize anything */

    if (!proxy_config.initializing)
        gnome_property_box_changed (GNOME_PROPERTY_BOX (proxy_config.dialog));
}

void
on_cache_path_changed(GtkWidget *w, GdkEvent *evt, gpointer data)
{
    if (!proxy_config.initializing)
        gnome_property_box_changed(GNOME_PROPERTY_BOX(proxy_config.dialog));
}

void
on_cache_difference_changed(GtkWidget *w, GdkEvent *event, gpointer data)
{
    if (!proxy_config.initializing)
        gnome_property_box_changed(GNOME_PROPERTY_BOX(proxy_config.dialog));
}

void
on_cache_enable_toggled(GtkWidget *toggle, gpointer data)
{
    set_enabled_from_toggle(toggle, "cache_path_hbox");
    set_enabled_from_toggle(toggle, "cache_difference_hbox");
}

void
on_cache_clear_packages_on_exit_toggled (GtkWidget *toggle, gpointer data)
{
    cache_config.cache_clear_packages_on_exit = GTK_TOGGLE_BUTTON (toggle)->active;
    gnome_property_box_changed (GNOME_PROPERTY_BOX (proxy_config.dialog));
}

void
on_cache_clear_packages_on_startup_toggled (GtkWidget *toggle, gpointer data)
{
    cache_config.cache_clear_packages_on_startup = GTK_TOGGLE_BUTTON (toggle)->active;
    gnome_property_box_changed (GNOME_PROPERTY_BOX (proxy_config.dialog));
}

void
on_gui_show_pretty_names_toggled(GtkWidget *toggle, gpointer data)
{
    gui_config.show_pretty_names = GTK_TOGGLE_BUTTON(toggle)->active;
    gnome_property_box_changed(GNOME_PROPERTY_BOX(proxy_config.dialog));
} /* on_gui_show_pretty_name_toggled */

void
on_property_dialog_apply (GtkWidget *dialog, gpointer data)
{
    proxy_config_save ();
}

gboolean
on_property_dialog_close (GtkWidget *dialog, gpointer data)
{
    return FALSE;
}

void
on_clist_selected(GtkWidget *clist, int row, int column, GdkEvent *event)
{
    if (!proxy_config.initializing)
        gnome_property_box_changed(GNOME_PROPERTY_BOX(proxy_config.dialog));

    mirror_config.selected_row = row;
}

void
on_clist_unselected(GtkWidget *clist, int row, int column, GdkEvent *event)
{
    if (!proxy_config.initializing)
        gnome_property_box_changed(GNOME_PROPERTY_BOX(proxy_config.dialog));

    mirror_config.selected_row = -1;
}

void
on_preferences1_activate (GtkWidget *item, gpointer data)
{
    gnome_dialog_close_hides (GNOME_DIALOG (proxy_config.dialog), TRUE);
    gnome_dialog_set_parent(
        GNOME_DIALOG(proxy_config.dialog), 
        GTK_WINDOW(GET_WIDGET("main_window")));
    gtk_widget_show (proxy_config.dialog);
    gdk_window_raise (proxy_config.dialog->window);
}

void
refresh_button_clicked(GtkWidget *button)
{
    save_mirror_list(update_mirrors());
    refresh_mirror_list(GTK_CLIST(GET_WIDGET("mirror_clist")));
}
