/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * gui-transfer.c: Wrapper around GnomeTransfer that does blocking transfers
 * and handles caching
 *
 * Copyright (C) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>

#include <gnome.h>
#include "gui.h"
#include "gui-transfer.h"
#include "gui-prefs.h"
#include "cache.h"
#include "gui-util.h"

#undef DEBUG

static void file_info_cb(GnomeTransfer *gt, gchar *name, gint size, 
                         GUITransfer *t);
static void file_data_cb(GnomeTransfer *gt, gchar *buf, gint size, 
                         GUITransfer *t);
static void file_done_cb(GnomeTransfer *gt, GnomeTransferStatus status, 
                         GUITransfer *t);
static void file_paused_cb(GnomeTransfer *gt, int bytes, GUITransfer *t);
static void auth_required_cb(GnomeTransfer *gt, char *realm, GUITransfer *t);

GSList *current_transfers = NULL;

static GtkObjectClass *gui_transfer_parent;

enum SIGNALS {
    FILE_DATA,
    FILE_DONE,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

static void
gui_transfer_destroy(GtkObject *obj)
{
    GUITransfer *t = GUI_TRANSFER(obj);
    
    if (t->gt) {
        gtk_object_destroy(GTK_OBJECT(t->gt));
        t->gt = NULL;
    }

    if (t->data) {
        g_byte_array_free(t->data, TRUE);
        t->data = NULL;
    }

    if (t->filename) {
        g_free(t->filename);
        t->filename = NULL;
    }

    if (GTK_OBJECT_CLASS(gui_transfer_parent)->destroy)
        (* GTK_OBJECT_CLASS(gui_transfer_parent)->destroy)(obj);
} /* gui_transfer_destroy */

static void
gui_transfer_class_init(GUITransferClass *klass)
{
    GtkObjectClass *object_class = (GtkObjectClass *) klass;

    object_class->destroy = gui_transfer_destroy;

    gui_transfer_parent = gtk_type_class(gtk_object_get_type());

    signals[FILE_DATA] = 
        gtk_signal_new(
            "file_data",
            GTK_RUN_LAST,
            object_class->type,
            GTK_SIGNAL_OFFSET(GUITransferClass, file_data),
            gtk_marshal_NONE__POINTER_INT,
            GTK_TYPE_NONE, 2,
            GTK_TYPE_POINTER,
            GTK_TYPE_INT);

    signals[FILE_DONE] =
        gtk_signal_new(
            "file_done",
            GTK_RUN_LAST,
            object_class->type,
            GTK_SIGNAL_OFFSET(GUITransferClass, file_done),
            gtk_marshal_NONE__NONE,
            GTK_TYPE_NONE, 0);

    gtk_object_class_add_signals(object_class, signals, LAST_SIGNAL);
} /* gui_transfer_class_init */

void
gui_transfer_set_proxy_info(GUITransfer *t, const char *url, const char *name, 
                            const char *pass)
{
    gnome_transfer_proxy(t->gt, (char *) url);

    if (name && pass)
        gnome_transfer_proxy_password(t->gt, (char *) name, (char *) pass);
} /* gui_transfer_set_proxy_info */

static void
gui_transfer_init(GUITransfer *t)
{
    gboolean proxy;
    char *purl, *pname, *ppass;

    t->gt = gnome_transfer_new();
    proxy = gui_prefs_get_proxy_info(&purl, &pname, &ppass);
    if (proxy) {
        gui_transfer_set_proxy_info(t, purl, pname, ppass);

        g_free(purl);
        g_free(pname);
        g_free(ppass);
    }
    t->data = g_byte_array_new();

    t->error = GUI_TRANSFER_ERROR_NONE;

    gtk_signal_connect(
        GTK_OBJECT(t->gt), "file_info",
        GTK_SIGNAL_FUNC(file_info_cb), t);
    gtk_signal_connect(
        GTK_OBJECT(t->gt), "file_data",
        GTK_SIGNAL_FUNC(file_data_cb), t);
    gtk_signal_connect(
        GTK_OBJECT(t->gt), "file_done",
        GTK_SIGNAL_FUNC(file_done_cb), t);
    gtk_signal_connect(
        GTK_OBJECT(t->gt), "file_paused",
        GTK_SIGNAL_FUNC(file_paused_cb), t);
    gtk_signal_connect(
        GTK_OBJECT(t->gt), "auth_required",
        GTK_SIGNAL_FUNC(auth_required_cb), t);
} /* gui_transfer_init */

guint
gui_transfer_get_type(void)
{
    static guint type = 0;

    if (!type) {
        GtkTypeInfo type_info = {
            "GUITransfer",
            sizeof(GUITransfer),
            sizeof(GUITransferClass),
            (GtkClassInitFunc) gui_transfer_class_init,
            (GtkObjectInitFunc) gui_transfer_init,
            (GtkArgSetFunc) NULL,
            (GtkArgGetFunc) NULL
        };

        type = gtk_type_unique(gtk_object_get_type(), &type_info);
    }

    return type;
} /* gui_transfer_get_type */

GUITransfer *
gui_transfer_new(gboolean blocking, gboolean error_notify)
{
    GUITransfer *new = GUI_TRANSFER(gtk_type_new(gui_transfer_get_type()));

    new->block = blocking;
    new->error_notify = error_notify;

    return new;
} /* gui_transfer_new */

GByteArray *
gui_transfer(GUITransfer *t, char *url)
{
    int rc;

    g_return_val_if_fail(t, NULL);

    t->error = GUI_TRANSFER_ERROR_NONE;

    t->url = g_strdup(url);
    if (g_strncasecmp(url, "http", 4) == 0)
        rc = gnome_transfer_headers(t->gt, url);
    else {
        rc = gnome_transfer_begin(t->gt, url);
    }
    current_transfers = g_slist_append(current_transfers, t);

    if (rc) {
        if (t->error_notify) {
            char *err;
            gboolean retry;

            err = g_strdup_printf(
                "Unable to begin transfer to %s: %s", url, 
                url[0] == 'f' ? "File not found" : "Host name unresolved");
            retry = rc_gui_error_box(err, RC_ERROR_RETRY);
            g_free(err);
            if (retry)
                t->error = GUI_TRANSFER_ERROR_RETRY;
            else
                t->error = GUI_TRANSFER_ERROR_ABORT;
        }
        else
            t->error = GUI_TRANSFER_ERROR_ABORT;
        current_transfers = g_slist_remove(current_transfers, t);
        return NULL;
    }

    if (t->block) {
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "Blocking for transfer\n");
        gtk_main();

        if (t->data) {
            GByteArray *ret = t->data;
            t->data = g_byte_array_new();
            return ret;
        }
    }

    return NULL;
} /* gui_transfer */

void
gui_transfer_set_progress_bar(GUITransfer *t, GtkProgress *progress_bar)
{
    g_return_if_fail(t);
    g_return_if_fail(progress_bar);
    g_return_if_fail(GTK_IS_PROGRESS(progress_bar));

    t->progress_bar = GTK_WIDGET(progress_bar);
} /* gui_transfer_set_progess_bar */

void
gui_transfer_set_progress_label(GUITransfer *t, GtkLabel *label)
{
    g_return_if_fail(t);
    g_return_if_fail(label);
    g_return_if_fail(GTK_IS_LABEL(label));

    t->label = GTK_WIDGET(label);
} /* gui_transfer_set_progress_label */

GSList *
gui_transfer_get_current_transfers(void)
{
    return current_transfers;
} /* gui_transfer_get_current_transfers */

GUITransferError
gui_transfer_get_error(GUITransfer *t)
{
    return t->error;
} /* gui_transfer_get_error */

void
gui_transfer_set_cache(GUITransfer *t, gboolean enabled)
{
    t->nocache = !enabled;
} /* gui_transfer_set_cache */

static void
file_info_cb(GnomeTransfer *gt, gchar *name, gint size, GUITransfer *t)
{
    t->filename = g_strdup(name);
    t->file_size = size;

    /* We just got headers, we don't want to do the rest of this stuff */
    if (gt->type == GNOME_TRANSFER_TYPE_HEADERS)
        return;

    /* Set up the timer. */
    if (t->timer)
        g_timer_destroy(t->timer);
    t->timer = g_timer_new();
    g_timer_start(t->timer);

    /* Set the text of the label, if there is one */
    if (t->label) {
        gtk_label_set_text(GTK_LABEL(t->label), "Downloading data...");
    }
    
    /* Set the value and limits of the progress bar, if there is one */
    if (t->progress_bar) {
        gtk_progress_configure(
            GTK_PROGRESS(t->progress_bar), 0.0, 0.0, t->file_size);
    }
} /* file_info_cb */

static void
file_data_cb (GnomeTransfer *gt, gchar *buf, gint size, GUITransfer *t)
{
    t->trans_size += size;

    if (t->timer && t->paused) {
        g_timer_start(t->timer);
        t->paused = FALSE;
    }

//    rc_debug (RC_DEBUG_LEVEL_DEBUG, "Got %d bytes of data\n", size);

    /* Sanity check */
    if (!t->data && t->block) {
        rc_debug(
            RC_DEBUG_LEVEL_WARNING, 
            "Byte array is NULL, but we are supposed to block!");
    }

    if (t->data) {
        t->data = g_byte_array_append(t->data, buf, size);
    }

    gtk_signal_emit(GTK_OBJECT(t), signals[FILE_DATA], buf, size);

    if (t->progress_bar) {
        char *s1, *s2;
        gdouble et;
        char *et_str = NULL;
        char *str;
     
        if (t->timer) {
            et = g_timer_elapsed(t->timer, NULL);
            if (et > 2) {
                int bytes;

                bytes = gnome_transfer_bytes(gt) - t->offset;

                et_str = rc_gui_format_size((unsigned int) (bytes/et));
            }
        }    
        s1 = rc_gui_format_size(gnome_transfer_bytes(gt));
        s2 = rc_gui_format_size(gnome_transfer_total(gt));
        if (et_str) {
            str = g_strdup_printf(
                "%%p%%%% (%s of %s) - %s/s", s1, s2, et_str);
            g_free(et_str);
        }
        else
            str = g_strdup_printf("%%p%%%% (%s of %s)", s1, s2);
        g_free(s1);
        g_free(s2);
        gtk_progress_set_format_string(GTK_PROGRESS(t->progress_bar), str);
        g_free(str);
        gtk_progress_set_value(
            GTK_PROGRESS(t->progress_bar),
            (gfloat) gnome_transfer_bytes(gt));
    }
} /* file_data_cb */

static void
file_done_cb (GnomeTransfer *gt, GnomeTransferStatus status, GUITransfer *t)
{
    current_transfers = g_slist_remove(current_transfers, t);

    /* We just asked for headers, so don't do any of this other stuff. */
    if (gt->type == GNOME_TRANSFER_TYPE_HEADERS) {
        if (g_strcasecmp(gt->url->protocol, "file") != 0 && !t->nocache &&
            gui_prefs_get_cache_info(NULL, NULL, NULL, NULL) &&
            gui_cache_fresh(gt->url->path, gt->last_modified, t->file_size)) {
            GByteArray *data;

            rc_debug(
                RC_DEBUG_LEVEL_DEBUG, 
                "=====> FETCHING FROM CACHE: %s <=====\n", gt->url->path);

            data = gui_cache_get(gt->url->path);
            if (data) {
                gtk_signal_emit(
                    GTK_OBJECT(t), signals[FILE_DATA], data->data, data->len);
                gtk_signal_emit(
                    GTK_OBJECT(t), signals[FILE_DONE]);
                if (t->data && t->data != data)
                    g_byte_array_free(t->data, FALSE);
                t->data = data;

                if (t->block) {
                    gtk_main_quit();
                }
            }
        }
        else {
            if (gnome_transfer_begin(gt, t->url)) {
                if (t->error_notify) {
                    char *err;
                    gboolean retry;
                    
                    err = g_strdup_printf(
                        "Unable to begin transfer to %s: %s", t->url,
                        t->url[0] == 'f' ? "File not found" : 
                        "Host name unresolved");
                    retry = rc_gui_error_box(err, RC_ERROR_RETRY);
                    g_free(err);
                    if (retry)
                        t->error = GUI_TRANSFER_ERROR_RETRY;
                    else
                        t->error = GUI_TRANSFER_ERROR_ABORT;
                }
                else
                    t->error = GUI_TRANSFER_ERROR_ABORT;
                return;
            }
            current_transfers = g_slist_append(current_transfers, t);
        }
        return;
    }

    if (status == GNOME_TRANSFER_ABORTED) {
        t->error = GUI_TRANSFER_ERROR_ABORT;
    }

    if (status != GNOME_TRANSFER_PAUSED && t->timer)
        g_timer_stop(t->timer);

    if (status == GNOME_TRANSFER_FAILED) {
        if (t->error_notify) {
            char *s;
            char *err;
            gboolean retry;

            s = g_url_to_string(gt->url, FALSE);
            err = g_strdup_printf(
                "Transfer of %s did not complete: %s", s,
                gnome_transfer_reason(gt));
            retry = rc_gui_error_box(err, RC_ERROR_RETRY);
            g_free(err);
            g_free(s);
            if (retry)
                t->error = GUI_TRANSFER_ERROR_RETRY;
            else
                t->error = GUI_TRANSFER_ERROR_ABORT;
        }
        else
            t->error = GUI_TRANSFER_ERROR_ABORT;
        if (t->data) {
            g_byte_array_free(t->data, FALSE);
            t->data = NULL;
        }
    }
    else if (status == GNOME_TRANSFER_COMPLETE) {
        if (g_strcasecmp(gt->url->protocol, "file") != 0 && t->filename && 
            !t->nocache && gui_prefs_get_cache_info(NULL, NULL, NULL, NULL)) {
            gui_cache_add(t->filename, t->data);
        }

        gtk_signal_emit(
            GTK_OBJECT(t), signals[FILE_DONE]);
    }

    if (t->block) {
        gtk_main_quit();
    }
        
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "File done!\n");
} /* file_done_cb */

static void
file_paused_cb(GnomeTransfer *gt, int bytes, GUITransfer *t)
{
    char *s1, *s2, *str;

    if (gnome_transfer_total(gt) == -1)
        return;

    if (t->timer)
        g_timer_stop(t->timer);
    t->paused = TRUE;
    t->offset = bytes;

    s1 = rc_gui_format_size(gnome_transfer_bytes(gt));
    s2 = rc_gui_format_size(gnome_transfer_total(gt));
    str = g_strdup_printf("%%p%%%% (%s of %s) - Paused", s1, s2);
    g_free(s1);
    g_free(s2);
    gtk_progress_set_format_string(GTK_PROGRESS(t->progress_bar), str);
    g_free(str);
} /* file_paused_cb */

static void
auth_required_cb(GnomeTransfer *gt, char *realm, GUITransfer *t)
{
    if (t->error_notify) {
        char *err;
        
        err = g_strdup(
            "Proxy authorization required. This usually means "
            "that your proxy requires a name and password and that you are "
            "either not providing that info or that it is incorrect.</p>"
            "<p>You can set your proxy settings in the "
            "<a href=\"updater:prefs_proxy\">proxy preferences</a>."
            "</a>");
        rc_gui_error_box(err, RC_ERROR_ABORT);
        g_free(err);
    }

    if (t->data) {
        g_byte_array_free(t->data, FALSE);
        t->data = NULL;
    }
    
    if (t->block)
        gtk_main_quit();
} /* auth_required_cb */

