/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * gui-util.c: GUI Utility functions
 *
 * Copyright 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <fcntl.h>
#include <crypt.h>
#include <time.h>

#include <gnome.h>
#include <glade/glade.h>
#include <gtkhtml/gtkhtml.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <libgnomeui/gnome-window-icon.h>

#include <gtkhtml/htmlimage.h>
#include <gtkhtml/htmlengine-search.h>

#include "gui.h"
#include "gui-prefs.h"
#include "cache.h"
#include "gui-install.h"
#include "mirror.h"
#include "subscription.h"

#include "gui-util.h"

typedef struct _RCErrorData RCErrorData;

struct _RCErrorData {
    GtkHTML *html;
    RCError error;
    gboolean returncode;
};

/* FIXME: This should not be a #define */
#define DEFAULT_HOST_ROOT "red-carpet.ximian.com:8000"
const char *
rc_gui_get_host (void)
{
    char *host = getenv("RC_HOST_ROOT");
    return host ? host : DEFAULT_HOST_ROOT;
} /* rc_gui_get_host */

void
rc_gui_query_system_packages (void)
{
    RCPackmanError err;
    RCPackageSList *system_packages;

    do {

        system_packages = rc_packman_query_all (appinfo->pman);
        err = rc_packman_get_error (appinfo->pman);

        if (err) {
            gchar *reason;

            reason = g_strdup_printf (
                "Red Carpet encountered an error querying your system:<br>"
                "%s", rc_packman_get_reason (appinfo->pman));
            rc_gui_error_box (
                reason, err == RC_PACKMAN_ERROR_FATAL ? RC_ERROR_FATAL :
                RC_ERROR_RETRY);

            g_free (reason);

            if (err == RC_PACKMAN_ERROR_FATAL) {
                gtk_exit (1);
            }
        }
    } while (err);

    rc_package_slist_free (appinfo->system_packages);
    appinfo->system_packages = system_packages;
}

void
rc_gui_set_page(char *notebook, int page)
{
    gtk_notebook_set_page(GTK_NOTEBOOK(GET_WIDGET(notebook)), page);
} /* rc_gui_set_page */

gboolean
rc_gui_url_is_qualified(const char *url)
{
    /* We only care about the URI types that GnomeTransfer can handle. Notably,
       http, ftp, and file. */

    if (g_strncasecmp(url, "http", 4) == 0 ||
        g_strncasecmp(url, "ftp", 3) == 0 ||
        g_strncasecmp(url, "file", 4) == 0)
        return TRUE;
    else
        return FALSE;
} /* rc_gui_url_is_qualified */

static char *
format_size(unsigned int size, gboolean nbsp)
{
    char *output;

    if (size > 1024*1024)
        output = g_strdup_printf(
            "%.2f%sMB", (float) size / ((float) 1024*1024),
            nbsp ? "&nbsp;" : " ");
    else if (size > 1024)
        output = g_strdup_printf("%d%sKB", size / 1024, nbsp ? "&nbsp;" : " ");
    else
        output = g_strdup_printf("%d%sbytes", size, nbsp ? "&nbsp;" : " ");

    return output;
} /* format_size */

char *
rc_gui_format_size(unsigned int size)
{
    char *output;

    output = format_size(size, FALSE);

    return output;
} /* rc_gui_format_size */

char *
rc_gui_format_size_with_nbsp(unsigned int size)
{
    char *output;

    output = format_size(size, TRUE);

    return output;
} /* rc_gui_format_size_with_nbsp */

char *
rc_gui_number_to_word(int number)
{
    static char *ntable[] = {
        "zero", "one", "two", "three", "four", "five", "six", "seven", "eight",
        "nine" };
    char *n;

    if (number > 9 || number < 0) {
        n = g_strdup_printf("%d", number);
        return n;
    }

    n = g_strdup(ntable[number]);
    return n;
} /* rc_gui_number_to_word */

const gchar *
rc_gui_section_icon_name (RCPackageSection section)
{
    static const gchar *section_names[RC_SECTION_LAST] = {
        "file://"SHAREDIR"/sections/section-office.png",
        "file://"SHAREDIR"/sections/section-imaging.png",
        "file://"SHAREDIR"/sections/section-pim.png",
        "file://"SHAREDIR"/sections/section-game.png",
        "file://"SHAREDIR"/sections/section-multimedia.png",
        "file://"SHAREDIR"/sections/section-internet.png",
        "file://"SHAREDIR"/sections/section-util.png",
        "file://"SHAREDIR"/sections/section-system.png",
        "file://"SHAREDIR"/sections/section-doc.png",
        "file://"SHAREDIR"/sections/section-devel.png",
        "file://"SHAREDIR"/sections/section-develutil.png",
        "file://"SHAREDIR"/sections/section-library.png",
        "file://"SHAREDIR"/sections/section-xapp.png",
        "file://"SHAREDIR"/sections/section-misc.png"
    };

    return (section_names[section]);
}

const char *
rc_gui_get_pretty_name(const char *package_name)
{
    gboolean pretty = gui_prefs_get_pretty_names();
    const char *name;

    g_return_val_if_fail(package_name, NULL);

    if (pretty)
        name = rc_pretty_name_lookup(package_name);
    else
        name = package_name;

    return name;
} /* rc_gui_get_package_name */

/* CB */
static void
error_ok_clicked(GtkWidget *button, RCErrorData *errordata)
{
    int page = GPOINTER_TO_INT(gtk_object_get_data(
        GTK_OBJECT(errordata->html), "prev_page"));

    if (errordata->error == RC_ERROR_FATAL) {
        /* FIXME: Perhaps exit with a more meaningful error code? */
        gtk_exit(1);
    }
    else if (errordata->error == RC_ERROR_RETRY) {
        gtk_notebook_set_page(GTK_NOTEBOOK(GET_WIDGET("main_notebook")), page);
        gtk_object_set_data(
            GTK_OBJECT(errordata->html), "prev_page", GINT_TO_POINTER(-1));
        errordata->returncode = TRUE;
    }
    else if (errordata->error == RC_ERROR_ABORT) {
        gtk_notebook_set_page(GTK_NOTEBOOK(GET_WIDGET("main_notebook")), page);
        gtk_object_set_data(
            GTK_OBJECT(errordata->html), "prev_page", GINT_TO_POINTER(-1));
        errordata->returncode = FALSE;
    }
    gtk_main_quit();
} /* error_ok_clicked */

/* CB */
static void
error_cancel_clicked(GtkWidget *button, RCErrorData *errordata)
{
    int page;

    g_assert(errordata->error == RC_ERROR_RETRY);

    page = GPOINTER_TO_INT(gtk_object_get_data(
        GTK_OBJECT(errordata->html), "prev_page"));
    gtk_notebook_set_page(GTK_NOTEBOOK(GET_WIDGET("main_notebook")), page);
    gtk_object_set_data(
        GTK_OBJECT(errordata->html), "prev_page", GINT_TO_POINTER(-1));
    errordata->returncode = FALSE;
    gtk_main_quit();
} /* error_cancel_clicked */

/* Display an HTML error box. The message passed in should be HTML. */
gboolean
rc_gui_error_box(const char *message, RCError err)
{
    GtkWidget *frame;
    GtkWidget *html;
    GtkWidget *notebook;
    GtkHTMLStream *stream;
    int page;
    gchar *hstr;
    RCErrorData errordata;
    gboolean startup, mainw;

    startup = GTK_WIDGET_VISIBLE(GET_WIDGET("startup_window"));
    mainw = GTK_WIDGET_VISIBLE(GET_WIDGET("main_window"));

    if (startup)
        gtk_widget_hide(GET_WIDGET("startup_window"));

    if (!mainw)
        gtk_widget_show(GET_WIDGET("main_window"));

    notebook = GET_WIDGET("main_notebook");

    page = gtk_notebook_current_page(GTK_NOTEBOOK(notebook));
    rc_gui_set_page ("main_notebook", PAGE_ERROR);

    gtk_widget_hide(GET_WIDGET("cancel_dialog"));

    frame = GET_WIDGET("error_html");
    g_assert(frame);
    html = gtk_object_get_user_data(GTK_OBJECT(frame));
    g_assert(html);
    gtk_html_allow_selection(GTK_HTML(html), TRUE);

    gtk_object_set_data(GTK_OBJECT(html), "prev_page", GINT_TO_POINTER(page));

    stream = gtk_html_begin(GTK_HTML(html));
    hstr = g_strdup_printf(
        "<html><body marginwidth=0 marginheight=0>"
        "<table cellspacing=0 cellpadding=1 border=0 bgcolor=\"#ffffff\""
        "    width=\"100%%\">"
        "<tr bgcolor=\"#000000\"><td align=left width=36>"
        "<img width=36 height=36 src=\"file://" SHAREDIR "/warning.png\"></td>"
        "<td width=\"100%%\" align=left>"
        "<font color=\"#ffffff\" size=+2><b>%s</b></font>"
        "</td></tr>"
        "</table>"
        "<table cellspacing=0 cellpadding=1 border=0 bgcolor=\"#ffffff\""
        "    width=\"100%%\">"
        "<tr bgcolor=\"#e5e3e3\"><td colspan=2>"
        "<p>Red Carpet has encountered an error while trying to process your "
        "request:</p>"
        "<p>%s</p>"
        "</td></tr>"
        "<tr background=\"file://"SHAREDIR"/warning-bar.png\">"
        "<td colspan=2 height=15>&nbsp;"
        "</td></tr>"
        "<tr><td colspan=2><p>Possible actions:<ul>", 
        err == RC_ERROR_FATAL ? "Fatal Error" : "Error", message);
    gtk_html_write(GTK_HTML(html), stream, hstr, strlen(hstr));
    g_free (hstr);

    if (err == RC_ERROR_FATAL) {
        hstr = "<li>Click <b>OK</b> to quit.";
        gtk_widget_hide(GET_WIDGET("error_cancel"));
    }
    else if (err == RC_ERROR_RETRY) {
        hstr = 
            "<li>Click <b>OK</b> to retry."
            "<li>Click <b>Cancel</b> to cancel.";
        gtk_widget_show(GET_WIDGET("error_ok"));
    }
    else if (err == RC_ERROR_ABORT) {
        hstr = "<li>Click <b>OK</b> to cancel.";
        gtk_widget_hide(GET_WIDGET("error_cancel"));
    }

    gtk_html_write(GTK_HTML(html), stream, hstr, strlen(hstr));

    hstr = "</ul></td></tr></table></body></html>";
    gtk_html_write(GTK_HTML(html), stream, hstr, strlen(hstr));

    gtk_html_end(GTK_HTML(html), stream, GTK_HTML_STREAM_OK);

    errordata.html = GTK_HTML(html);
    errordata.error = err;
    /* This will determine how to return from this function and is set in the
       callback. We are just initializing it here. */
    errordata.returncode = FALSE;

    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("error_ok")), "clicked",
        GTK_SIGNAL_FUNC(error_ok_clicked), &errordata);
    gtk_signal_connect(
        GTK_OBJECT(GET_WIDGET("error_cancel")), "clicked",
        GTK_SIGNAL_FUNC(error_cancel_clicked), &errordata);

    gtk_main();

    gtk_signal_disconnect_by_func(
        GTK_OBJECT(GET_WIDGET("error_ok")), 
        GTK_SIGNAL_FUNC(error_ok_clicked), &errordata);
    gtk_signal_disconnect_by_func(
        GTK_OBJECT(GET_WIDGET("error_cancel")), 
        GTK_SIGNAL_FUNC(error_cancel_clicked), &errordata);

    if (startup)
        gtk_widget_show(GET_WIDGET("startup_window"));

    if (!mainw)
        gtk_widget_hide(GET_WIDGET("main_window"));
    
    return errordata.returncode;
} /* rc_gui_error_box */

void
rc_gui_split_channel_list (GUIChannelSList  *channels,
                           RCChannelSList  **subscribed_channels,
                           RCChannelSList  **unsubscribed_channels)
{
    GUIChannelSList *citer;

    for (citer = channels; citer; citer = citer->next) {
        GUIChannel *channel = (GUIChannel *)(citer->data);

        if (channel->subscribed) {
            *subscribed_channels = g_slist_append (
                *subscribed_channels, RCCHANNEL (channel));
        } else {
            *unsubscribed_channels = g_slist_append (
                *unsubscribed_channels, RCCHANNEL (channel));
        }
    }
}

int
rc_gui_split_by_package_category(RCPackageSList *packages, 
                                 RCPackageSList ***split_packages)
{
    GSList *i;
    int count = 0;

    g_return_val_if_fail(split_packages, 0);

    /* RC_SECTION_LAST is just an index for the number of categories we have */
    *split_packages = g_new0(RCPackageSList *, RC_SECTION_LAST);
    i = packages;
    while (i) {
        RCPackage *package = i->data;

        (*split_packages)[package->section] = g_slist_append(
            (*split_packages)[package->section], package);
        count++;
        
        i = i->next;
    }

    return count;
} /* rc_gui_split_by_package_category */

/* FIXME: this came from rc-package.c, but it doesn't fit well there,
 * because it needs to have a reference to a packman object to compare
 * RCPackageSpec's well... this probably belongs somewhere more
 * general in the end, but it had to come out of rc-package.c ... */

static RCPackageImportance
rc_package_get_highest_importance (RCPackage *package)
{
    RCPackage *system_package;
    RCPackageUpdateSList *iter;
    RCPackageImportance highest = RC_IMPORTANCE_INVALID;

    system_package = rc_package_new ();

    system_package->spec.name = g_strdup (package->spec.name);

    system_package = rc_packman_query (appinfo->pman, system_package);

    for (iter = package->history; iter; iter = iter->next) {
        RCPackageUpdate *update = (RCPackageUpdate *)(iter->data);

        if (rc_packman_version_compare (
                appinfo->pman, RC_PACKAGE_SPEC (update),
                RC_PACKAGE_SPEC (system_package)) <= 0) {
            break;
        }

        if (update->importance < highest || highest == RC_IMPORTANCE_INVALID) {
            highest = update->importance;
        }
    }

    /* The XML is broken: the importance is invalid. In this case,
       throw a g_warning and set the importance to RC_IMPORTANCE_MINOR */
    if (highest == RC_IMPORTANCE_INVALID) {
        g_warning(
            "Update \"%s\" has invalid importance. Setting to minor.",
            system_package->spec.name);
        highest = RC_IMPORTANCE_MINOR;
    }

    rc_package_free (system_package);
    return (highest);
}

int
rc_gui_split_by_package_importance(RCPackageSList *available_updates,
                                   RCPackageSList *system_packages,
                                   RCPackageSList ***split_packages,
                                   int *total_size)
{
    GSList *i;
    int count = 0;
    int tot = 0;

    g_return_val_if_fail(split_packages, 0);

    *split_packages = g_new0(RCPackageSList *, RC_IMPORTANCE_LAST);
    i = available_updates;
    while (i) {
        RCPackage *package = i->data;
        RCPackageUpdate *update = package->history->data;
        RCPackageImportance high;

        high = rc_package_get_highest_importance(package);

        /* We can add this to a list of available updates */
        (*split_packages)[high] = g_slist_append(
            (*split_packages)[high], package);
        tot += update->package_size;
        count++;
        
        i = i->next;
    }

    if (total_size)
        *total_size = tot;

    return count;
} /* split_by_package_importance */

void
rc_gui_set_refresh_item(RefreshType type)
{
    GtkWidget *w;

    w = GET_WIDGET("channel_refresh");
    if (type == REFRESH_CHANNEL)
        gtk_widget_show(w);
    else
        gtk_widget_hide(w);

    w = GET_WIDGET("summary_refresh");
    if (type == REFRESH_SUMMARY)
        gtk_widget_show(w);
    else
        gtk_widget_hide(w);

    w = GET_WIDGET("news_refresh");
    if (type == REFRESH_NEWS)
        gtk_widget_show(w);
    else
        gtk_widget_hide(w);
} /* rc_gui_set_refresh_item */
    
void
rc_gui_set_select_item(ActionType action)
{
    GtkWidget *w;

    w = GET_WIDGET("select_all_update");
    if (action == ACTION_UPDATE)
        gtk_widget_show(w);
    else
        gtk_widget_hide(w);

    w = GET_WIDGET("select_none_update");
    if (action == ACTION_UPDATE)
        gtk_widget_show(w);
    else 
        gtk_widget_hide(w);

    w = GET_WIDGET("select_all_install");
    if (action == ACTION_INSTALL)
        gtk_widget_show(w);
    else
        gtk_widget_hide(w);

    w = GET_WIDGET("select_none_install");
    if (action == ACTION_INSTALL)
        gtk_widget_show(w);
    else
        gtk_widget_hide(w);

    w = GET_WIDGET("select_all_remove");
    if (action == ACTION_REMOVE)
        gtk_widget_show(w);
    else
        gtk_widget_hide(w);
    gtk_widget_set_sensitive(w, FALSE);

    w = GET_WIDGET("select_none_remove");
    if (action == ACTION_REMOVE)
        gtk_widget_show(w);
    else
        gtk_widget_hide(w);
} /* rc_gui_set_select_item */

void
rc_gui_sensitize_select_items(gboolean sensitize)
{
    GtkWidget *w;

    w = GET_WIDGET("select_all_update");
    gtk_widget_set_sensitive(w, sensitize);
    w = GET_WIDGET("select_none_update");
    gtk_widget_set_sensitive(w, sensitize);
    w = GET_WIDGET("select_all_install");
    gtk_widget_set_sensitive(w, sensitize);
    w = GET_WIDGET("select_none_install");
    gtk_widget_set_sensitive(w, sensitize);
    w = GET_WIDGET("select_all_remove");
    gtk_widget_set_sensitive(w, FALSE);
    w = GET_WIDGET("select_none_remove");
    gtk_widget_set_sensitive(w, sensitize);
} /* rc_gui_sensitize_select_items */


static void
install_local_file_do (GtkWidget *widget, gpointer user_data)
{
    RCPackage *pkg;
    RCPackageUpdate *pkg_up;
    RCPackageSList *pkgs = NULL;
    gchar *file_name;

    file_name = g_strdup (gtk_file_selection_get_filename (GTK_FILE_SELECTION(user_data)));

    pkg = rc_packman_query_file (appinfo->pman, file_name);

    if (!pkg) {
        g_free (file_name);
        return;
    }

    /* make an 'update' for this package */
    pkg_up = rc_package_update_new ();
    pkg_up->package_url = g_strconcat ("file:", file_name, NULL);
    pkg_up->importance = RC_IMPORTANCE_SUGGESTED;
    pkg_up->description = g_strdup ("No information available.");

    pkg->history = g_slist_append (pkg->history, pkg_up);

    pkgs = g_slist_append (pkgs, pkg);

    gui_install_packages (NULL, pkgs);

    g_free (file_name);
    rc_package_free (pkg);
    g_slist_free (pkgs);

    gtk_widget_hide (GTK_WIDGET (user_data));
}

void
rc_gui_install_local_file (void)
{
    GtkFileSelection *fs;

    fs = (GtkFileSelection *) gtk_file_selection_new ("Choose a package...");
    /* FIXME -- this should prepend arch */

    /* We can't just use IS_RC_RPMMAN/etc. since those files aren't included anywhere
     * but distman
     */

//    if (IS_RC_RPMMAN (appinfo->pman)) {
//        gtk_file_selection_complete (fs, "*.deb");
//    } else if (IS_RC_DEBMAN (appinfo->pman)) {
//        gtk_file_selection_complete (fs, "*.rpm");
//    }
#ifdef HAVE_RPM_RPMLIB_H
    gtk_file_selection_complete (fs, "*.rpm");
#else
    gtk_file_selection_complete (fs, "*.deb");
#endif

    gtk_signal_connect (GTK_OBJECT (fs->ok_button),
                        "clicked", GTK_SIGNAL_FUNC (install_local_file_do),
                        (gpointer) fs);

    gtk_signal_connect_object (GTK_OBJECT (fs->ok_button),
                               "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
                               (gpointer) fs);

    gtk_signal_connect_object (GTK_OBJECT (fs->cancel_button),
                               "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy),
                               (gpointer) fs);
    gtk_widget_show (GTK_WIDGET (fs));
} /* rc_gui_install_local_file */

void
rc_gui_set_menubar_sensitive (gboolean sense)
{
    gtk_widget_set_sensitive(GET_WIDGET("summary_refresh"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("channel_refresh"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("search"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("unsubscribe1"), sense);

    gtk_widget_set_sensitive(GET_WIDGET("verify_installed_packages"), sense);
//    gtk_widget_set_sensitive(GET_WIDGET("install_local_file1"), sense);
    gtk_widget_set_sensitive(GET_WIDGET("about1"), sense);
}
