/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * mirror.c: Mirroring
 *
 * Copyright (c) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gnome.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "mirror.h"

#include "rc-util.h"
#include "gui-transfer.h"
#include "gui-util.h"
#include "gui-prefs.h"

typedef struct _RCMirror RCMirror;

struct _RCMirror {
    char *name;
    char *location;
    char *url;
};

static GSList *
parse_mirrors_xml(void)
{
    xmlDoc *doc;
    xmlNode *root, *node;
    GSList *mirrors = NULL;
    RCMirror *m;

    m = g_new0(RCMirror, 1);
    m->name = g_strdup("No mirror (always use default host)");
    m->location = g_strdup("N/A");
    m->url = NULL;
    mirrors = g_slist_append(mirrors, m);

    if (!rc_file_exists("/var/lib/redcarpet/mirrors.xml")) {
        /* No mirrors.xml file on the local system yet. No big deal. */
        return mirrors;
    }

    doc = xmlParseFile("/var/lib/redcarpet/mirrors.xml");
    if (!doc) {
        GtkWidget *message;

        message = gnome_app_error(
            GNOME_APP(GET_WIDGET("main_window")), 
            "Unable to parse mirror list.");
        gnome_dialog_run(GNOME_DIALOG(message));
        return NULL;
    }
#if LIBXML_VERSION < 20000
    root = doc->root;
#else
    root = xmlDocGetRootElement(doc);
#endif
    if (!root) {
        /* Uh. Bad. */
        rc_debug (RC_DEBUG_LEVEL_WARNING, "mirrors.xml document has no root\n");
        xmlFreeDoc(doc);
        return NULL;
    }

#if LIBXML_VERSION < 20000
    node = root->childs;
#else
    node = root->children;
#endif
    while (node) {
        char *tmp;
        RCMirror *mirror;

        tmp = xml_get_value(node, "internal");

        if (tmp && !getenv("RC_INTERNAL_MIRRORS")) {
            g_free(tmp);
            node = node->next;
            continue;
        }

        mirror = g_new0(RCMirror, 1);
        mirror->name = xml_get_value(node, "name");
        mirror->location = xml_get_value(node, "location");
        mirror->url = xml_get_value(node, "url");

        mirrors = g_slist_append(mirrors, mirror);

        node = node->next;
    }

    xmlFreeDoc(doc);

    return mirrors;
} /* parse_mirrors_xml */

GByteArray *
update_mirrors(void)
{
    GUITransfer *t;
    char *url;
    GByteArray *data = NULL;

    t = gui_transfer_new(TRUE, TRUE);
    url = rc_build_url("http", rc_gui_get_host(), NULL, "mirrors.xml");
    data = gui_transfer(t, url);
    gtk_object_destroy(GTK_OBJECT(t));
    g_free(url);

    if (data && data->data)
        data = g_byte_array_append(data, "\0", 1);

    return data;
} /* update_mirrors */

void
save_mirror_list(GByteArray *data)
{
    int fd;

    fd = open("/var/lib/redcarpet/mirrors.xml", 
              O_CREAT | O_TRUNC | O_WRONLY, 0644);
    if (fd < 0) {
        GtkWidget *message;

        message = gnome_app_error(
            GNOME_APP(GET_WIDGET("main_window")),
            "Unable to save mirror list.");
        gnome_dialog_run(GNOME_DIALOG(message));
        return;
    }
    write(fd, data->data, data->len);
    close(fd);
    g_byte_array_free(data, TRUE);
} /* save_mirror_list */

void
refresh_mirror_list(GtkCList *clist)
{
    GSList *mirror_list = NULL;
    GSList *i;

    g_return_if_fail(clist);
    g_return_if_fail(GTK_IS_CLIST(clist));

    mirror_list = parse_mirrors_xml();

    if (!mirror_list)
        return;

    gtk_clist_clear(clist);
    i = mirror_list;
    while (i) {
        RCMirror *mirror = i->data;
        char *stuff[] = { mirror->name, mirror->location, NULL};
        int row;

        row = gtk_clist_append(clist, stuff);
        gtk_clist_set_row_data(clist, row, mirror->url);

        i = i->next;
    }
} /* refresh_mirror_list */

char *
merge_mirror_url_and_path(const char *mirror, const char *path)
{
    Gurl *purl;
    char *merge;

    g_return_val_if_fail(mirror, NULL);

    if (!path)
        return g_strdup(mirror);

    if (!rc_url_is_absolute(path)) {
        merge = g_strconcat(mirror, *path == '/' ? (path+1) : path, NULL);
        return merge;
    }

    purl = g_url_new(path);
    merge = g_strconcat(
        mirror, purl->path ? 
        (*purl->path == '/' ? (purl->path + 1) : purl->path) : "", NULL);
    g_url_free(purl);
    return merge;
} /* merge_mirror_url_and_path */

char *
maybe_mirror_channel_url (const char *path, RCChannel *rc)
{
    gchar *url = NULL;

    if (rc_gui_url_is_qualified (path)) {
        return g_strdup (path);
    } else {
        const char *m = gui_prefs_get_mirror_url();

        if (m && rc->mirrored) {
            char *p = g_strconcat(
                rc->path, *path == '/' ? path + 1 :
                path, NULL);

            url = merge_mirror_url_and_path(m, p);
            g_free(p);
        }
        else {
            url = rc_build_url(
                "http", rc_gui_get_host(), rc->path,
                *path == '/' ? path + 1 : path);
        }
    }

    return url;
} /* maybe_mirror_channel_url */


