/* -*- Mode: C; tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/*
 * subscription.c: Code for managing subscriptions
 *
 * Copyright (c) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <stdlib.h>
#include <libgnome/libgnome.h>

#include "subscription.h"

#include "libredcarpet.h"

typedef struct _Subscription Subscription;

struct _Subscription {
    int id;
    time_t last_updated;
};

#define DEAD_BEEF ((gpointer)0xdeadbeef)

GSList *subs = DEAD_BEEF;

void
rc_gui_initialize_subscriptions(void)
{
    xmlDoc *doc;
    xmlNode *root, *node;
    
    RC_ENTRY;

    if (!g_file_exists("/var/lib/redcarpet/subscriptions.xml")) {
        subs = NULL;
        return;
    }

    doc = xmlParseFile("/var/lib/redcarpet/subscriptions.xml");
    if (!doc) {
        rc_debug (RC_DEBUG_LEVEL_WARNING, "Unable to open subscription file\n");
        RC_EXIT;
        return;
    }
    
    root = doc->root;
    if (!root) {
        /* No root? Bad(tm). */
        rc_debug (RC_DEBUG_LEVEL_WARNING, "subscription.xml document has no root");
        xmlFreeDoc(doc);
        RC_EXIT;
        return;
    }

    subs = NULL;
    
    node = root->childs;
    while (node) {
        Subscription *sub;
        char *tmp;
        
        sub = g_new0(Subscription, 1);

        tmp = xml_get_prop(node, "channel_id");
        if (tmp)
            sub->id = atoi(tmp);
        g_free(tmp);
        
        tmp = xml_get_prop(node, "last_updated");
        if (tmp)
            sub->last_updated = atol(tmp);
        g_free(tmp);
        
        subs = g_slist_append(subs, sub);
        
        node = node->next;
    }
    
    xmlFreeDoc(doc);
    
    RC_EXIT;
} /* rc_gui_initialize_subscriptions */

void
rc_gui_free_subscriptions(void)
{
    GSList *i;

    RC_ENTRY;

    g_return_if_fail(subs != DEAD_BEEF);

    i = subs;
    while (i) {
        /* Free the subscription struct. It doesn't contain any dynamic data */
        g_free(i->data);
        
        i = i->next;
    }
    
    g_slist_free(subs);
    subs = DEAD_BEEF;
    
    RC_EXIT;
} /* free_subscriptions */

gboolean
rc_gui_subscribed(RCChannel *channel)
{
    GSList *i;

    RC_ENTRY;

    g_return_val_if_fail(subs != DEAD_BEEF, FALSE);

    i = subs;
    while (i) {
        Subscription *sub = i->data;

        if (sub->id == channel->id) {
            RC_EXIT;
            return TRUE;
        }

        i = i->next;
    }

    RC_EXIT;
    return FALSE;
} /* rc_gui_subscribed */

static void
save_subscriptions(void)
{
    xmlDoc *doc;
    GSList *i;

    g_return_if_fail(subs != DEAD_BEEF);

    if (!g_file_exists("/var/lib/redcarpet")) {
        if (rc_mkdir("/var/lib/redcarpet", 0755)) {
            rc_debug (RC_DEBUG_LEVEL_WARNING, "Unable to create library directory /var/lib/redcarpet\n");
            return;
        }
    }

    doc = xmlNewDoc("1.0");
    doc->root = xmlNewDocNode(doc, NULL, "subscriptions", NULL);
    
    i = subs;
    while (i) {
        Subscription *sub = i->data;
        xmlNode *node;
        char *tmp;

        node = xmlNewChild(doc->root, NULL, "channel", NULL);

        tmp = g_strdup_printf("%d", sub->id);
        xmlNewProp(node, "channel_id", tmp);
        g_free(tmp);

        tmp = g_strdup_printf("%ld", sub->last_updated);
        xmlNewProp(node, "last_updated", tmp);
        g_free(tmp);

        i = i->next;
    }
    xmlSaveFile("/var/lib/redcarpet/subscriptions.xml", doc);
    xmlFreeDoc(doc);
} /* save_subscriptions */

void
rc_gui_subscribe(RCChannel *channel)
{
    /* I've decided against checking to see if the channel is already
       subscribed to in this function because the gui should really either
       know or check using rc_gui_subscribed to see if it's subscribed. */

    Subscription *sub;

    RC_ENTRY;

    g_return_if_fail(subs != DEAD_BEEF);

    sub = g_new0(Subscription, 1);
    sub->id = channel->id;
    sub->last_updated = channel->last_update;

    subs = g_slist_append(subs, sub);
    save_subscriptions();

    RC_EXIT;
} /* rc_gui_subscribe */

void
rc_gui_unsubscribe(RCChannel *channel)
{
    GSList *i;

    RC_ENTRY;

    g_return_if_fail(subs != DEAD_BEEF);

    i = subs;
    while (i) {
        Subscription *sub = i->data;

        if (sub->id == channel->id) {
            subs = g_slist_remove_link(subs, i);
            g_free(sub);
            break;
        }
        
        i = i->next;
    }

    save_subscriptions();

    RC_EXIT;
} /* rc_gui_unsubscribe */

int
rc_gui_subscription_length(void)
{
    g_return_val_if_fail(subs != DEAD_BEEF, -1);

    return g_slist_length(subs);
} /* rc_gui_subscription_length */

time_t
rc_gui_subscription_last_updated(RCChannel *channel)
{
    GSList *i;

    RC_ENTRY;

    g_return_val_if_fail(subs != DEAD_BEEF, 0);

    i = subs;
    while (i) {
        Subscription *sub = i->data;
        
        if (sub->id == channel->id) {
            RC_EXIT;
            return sub->last_updated;
        }
        
        i = i->next;
    }

    return 0;
} /* rc_gui_subscription_last_updated */

void
rc_gui_subscription_update(RCChannel *channel, time_t time)
{
    GSList *i;

    RC_ENTRY;

    g_return_if_fail(subs != DEAD_BEEF);

    i = subs;
    while (i) {
        Subscription *sub = i->data;

        if (sub->id == channel->id) {
            sub->last_updated = time;
            RC_EXIT;
            return;
        }

        i = i->next;
    }
} /* rc_gui_subscription_update */
